#ifndef _AR_KDOR_ALGORITHM_H_
#define _AR_KDOR_ALGORITHM_H_

#include "ar_mgr_types.h"
#include "plft_based_ar_algorithm.h"

#define INVALID_DIM_INDEX 0xFF
#define MAX_KDOR_PLFT_NUMBER 4

typedef uint8_t port_index_t;
typedef int8_t dim_sign_t;
typedef uint8_t dim_index_t;
typedef uint16_t dim_size_t;

class KdorSwData;
class ArKdorAlgorithm;
struct KdorConnection;
struct KdorRouteInfo;

typedef list <KdorSwData*> KdorSwDataList;
typedef list <osm_physp_t*> PhysPortsList;
typedef PhysPortsList::iterator PhysPortsListIter;
typedef vector <KdorConnection*> ConnectionVec;
typedef vector <KdorRouteInfo> KdorRouteInfoVec;


typedef map <osm_switch_t*, KdorConnection> SwToConnectionMap;
typedef SwToConnectionMap::iterator SwToConnectionMapIter;
typedef pair <SwToConnectionMapIter, bool> SwToConnectionMapInsertRes;

struct KdorGroupData {
    GroupData* group_data_array[MAX_KDOR_PLFT_NUMBER - 1];
};

typedef map <u_int16_t, KdorGroupData> LidToKdorGroupDataMap;
typedef LidToKdorGroupDataMap::iterator LidToKdorGroupDataMapIter;
typedef LidToKdorGroupDataMap::const_iterator LidToKdorGroupDataMapConstIter;
typedef pair <LidToKdorGroupDataMapIter, bool> LidToKdorGroupDataMapInsertRes;

enum kdor_turn_t {
    KDOR_TURN_TYPE_0 = 0,   //  +-Low   -> +-High
    KDOR_TURN_TYPE_1,       //  +-High  ->  +Low
    KDOR_TURN_TYPE_2,        //  +-High  ->  -Low
    KDOR_TURN_TYPE_LAST
};

enum bfs_status_t {
    BFS_INIT,
    BFS_QUEUED,
    BFS_DONE
};

struct KdorConnection {
    //if m_remote_switch == NULL this is a connection to all non sw ports
    osm_switch_t   *m_remote_switch;
    PhysPortsList   m_ports; // Elements of type, osm_physp_t*
    PortsBitset     m_ports_bitset;
    PhysPortsListIter  m_next_port_for_routing; // Will be used for load-balancing DLIDs among ports
    dim_index_t     m_dim_idx;
    dim_sign_t      m_dim_sign;
    bool            m_dateline_crossing; //for torus

    KdorConnection() :
        m_remote_switch(NULL), m_next_port_for_routing(0),
        m_dim_idx(0), m_dim_sign(0),
        m_dateline_crossing(false) {}

    ARSWDataBaseEntry *GetRemoteSwitchDbEntry() {
        return ((ARSWDataBaseEntry *)(m_remote_switch->priv));
    }

    string ToString() const;

};

struct KdorRouteInfo {
    KdorConnection     *m_connection;
    uint8_t             m_vl_inc; //kdor plane number
    kdor_turn_t         m_turn_type; //first turn type that is used in the path

private:
    static uint8_t turn_type_wait[KDOR_TURN_TYPE_LAST];
public:

    KdorRouteInfo() : m_connection(NULL), m_vl_inc(0),
        m_turn_type(KDOR_TURN_TYPE_0) {}

    void Init() {
        m_connection = NULL;
        m_vl_inc = 0;
        m_turn_type = KDOR_TURN_TYPE_0;
    }

    bool operator<(const KdorRouteInfo& __rhs) const;
};

//TODO: Create base class AlgorithmData
//struct AlgorithmData {
//  GroupBitmaskToDataMap m_groups_map;
//  GroupsList            m_assign_groups;
//};
struct KdorAlgorithmData : public TreeAlgorithmData {

    LidToKdorGroupDataMap m_lid_to_kdor_group_map;

    //Add lid to non empty group_data in ginen kdor_group_data
    //And update lid_to_kdor_group_map
    void AddLidToKdorGroupData(uint16_t lid_num,
                               uint16_t sw_lid_num,
                               KdorGroupData &kdor_group_data,
                               bool is_new_group,
                               osm_log_t *p_osm_log);
private:

    void AddLidToKdorGroupData(uint16_t lid_num,
                               KdorGroupData &kdor_group_data);

};


class KdorSwData {
    CalculatePortGroupsTask  m_calculate_port_groups_task_;

    void               *m_saved_priv_;
    ARSWDataBaseEntry  &m_db_entry_;
    uint32_t            m_sw_idx_;

    KdorRouteInfoVec    m_route_info_; //sw_idx_ to route info;
    SwToConnectionMap   m_connections_; // Inter-switch connections
    ConnectionVec       m_port_num_to_connection_tbl_;  // Will be used for fast lookup in VL2VL

    // Meta-data for BFS traversing during routing
    bfs_status_t        m_route_status_;
    uint16_t            m_route_distance_;

    //max lid number is restricted by plft table size
    uint16_t            m_max_lid_number_;

public:
    /**Data that should moved to OSM sw object**/
    PLFTMads            m_plft[MAX_KDOR_PLFT_NUMBER];
    u_int8_t            m_plft_number_configured; //configured on sw
    SlToVlPtrVecVec     m_vl2vl;   //key is from port to port
    BoolVecVec          m_is_to_set_vl2vl;  //key is from port to port
    //before reading plft_info ModeCap=Active_Mode=0
    //after read, set m_required_plft_mode_ be the desired mode
    //after set, set  Active_Mode = m_required_plft_mode_
    SMP_PrivateLFTInfo  m_plft_info;
    u_int8_t            m_required_plft_mode;
    SMP_PrivateLFTDef   m_plft_def;
    bool                m_set_plft_def;
    SMP_PortSLToPrivateLFTMap *m_plft_map[MAX_SUPPORTED_PORTS /
                                          PLFT_MAP_PORT_BLOCK_SIZE];
    bool                m_set_plft_map;
    bool                m_get_ext_sw_info;
    bool                m_set_ext_sw_info;

public:
    KdorSwData(ARSWDataBaseEntry &db_entry,
               ParallelPortGroupsCalculator *p_port_groups_calculator) :
        m_calculate_port_groups_task_(p_port_groups_calculator),
        m_saved_priv_(NULL), m_db_entry_(db_entry), m_sw_idx_(0),
        m_route_status_(BFS_INIT), m_route_distance_(0),
        m_max_lid_number_(0),
        m_plft_number_configured(0), m_required_plft_mode(0),
        m_set_plft_def(true), m_set_plft_map(true),
        m_get_ext_sw_info(true), m_set_ext_sw_info(true){}

    void Resize(uint32_t ports_num);

    void InitRouteInfo() {
        m_route_status_ = BFS_INIT;
        m_route_distance_ = 0;
    }

    KdorRouteInfo &GetRouteInfo(KdorSwData *p_dst_switch_data) {
        return m_route_info_[p_dst_switch_data->m_sw_idx_];
    }

    int Init(ArKdorAlgorithm &algorithm,
             ARSWDataBaseEntry &db_entry,
             uint32_t sw_idx,
             uint32_t switches_number);

    void SetRouteStatus(bfs_status_t status) { m_route_status_ = status; }
    bfs_status_t GetRouteStatus() const { return m_route_status_;}

    void SetRouteDistance(uint16_t dist) { m_route_distance_ = dist; }
    uint16_t GetRouteDistance() const { return m_route_distance_;}

    void SetMaxLidNumber(u_int8_t max_lid_number) {
        m_max_lid_number_ = max_lid_number;
    }
    uint16_t GetMaxLidNumber() { return m_max_lid_number_; }

    ThreadPoolTask *GetCalculatePortGroupsTask() {
        return &m_calculate_port_groups_task_;
    }

    SwToConnectionMap &GetConnections() { return m_connections_; }

    void *GetSavedPriv() { return m_saved_priv_; }

};

class ArKdorAlgorithm : public PlftBasedArAlgorithm {

protected:
    //dim_index_t     m_num_dims_;
    //dim_size_t     *m_dims_;
    uint8_t         m_max_vlinc_; //max static routing vl increase
    //uint8_t         m_max_cordinate_;

    KdorSwData     *m_sw_lid_to_kdor_data_[IB_LID_UCAST_END_HO + 1];

    SMP_SLToVLMappingTable  m_turn_1_vl2vl_per_op_vls_[MAX_OP_VL_CODE+1];
    SMP_SLToVLMappingTable  m_turn_2_vl2vl_per_op_vls_[MAX_OP_VL_CODE+1];
    SMP_SLToVLMappingTable *m_turn_type_to_vl2vl_per_op_vls_[KDOR_TURN_TYPE_LAST];

    SMP_PortSLToPrivateLFTMap m_plft_map_;

public:
    ArKdorAlgorithm(osm_log_t *p_osm_log,
                    GuidToSWDataBaseEntry &sw_map,
                    AdaptiveRoutingManager &ar_mgr,
                    supported_feature_t algorithm_feature,
                    uint8_t planes_number,
                    uint8_t min_vl_number) :
        PlftBasedArAlgorithm(p_osm_log, sw_map, ar_mgr,
                             algorithm_feature, planes_number, min_vl_number),
        m_max_vlinc_(0) {
            BuildKdorVl2VlPerOpVl();
            BuildKdorPlftMap();
        }

    virtual ~ArKdorAlgorithm(){}

    virtual int Preprocess();
    virtual int RunCycle();

    virtual void Clear() {}
    virtual int Init();

    virtual ThreadPoolTask *GetCalculatePortGroupsTask(
        ARSWDataBaseEntry &db_entry);

    virtual void CalculateSwitchPortGroups(ARSWDataBaseEntry &sw_db_entry,
                                           LidMapping *p_lid_mapping);

    void CalculateRouteInfo(KdorConnection &connection,
                            KdorSwData *p_remote_switch_data,
                            KdorSwData *p_dst_switch_data,
                            KdorRouteInfo &route_info) {
        CalculateRouteInfo(connection,
                           p_remote_switch_data->GetRouteInfo(p_dst_switch_data),
                           route_info);
    }

    void CalculateRouteInfo(KdorConnection &connection,
                            KdorRouteInfo &remote_route_info,
                            KdorRouteInfo &route_info);

    virtual int SetDirection(osm_switch_t* src,
                             osm_switch_t* dst,
                             KdorConnection &connection) = 0;

protected:
    //TODO: create SubnetArInfo class, get and set virtual methods
    virtual bool *GetIsToSetExtSwInfoPtr(ARSWDataBaseEntry &sw_db_entry);
    virtual SlToVlPtrVecVec &GetVl2VlMapping(ARSWDataBaseEntry &sw_db_entry);
    virtual BoolVecVec &GetIsToSetVl2VlMapping(ARSWDataBaseEntry &sw_db_entry);
    virtual SMP_PrivateLFTInfo &GetPlftInfo(ARSWDataBaseEntry &sw_db_entry);
    virtual uint8_t GetRequiredPlftMode(ARSWDataBaseEntry &sw_db_entry);
    virtual bool IsToSetPlftDef(ARSWDataBaseEntry &sw_db_entry);
    virtual SMP_PrivateLFTDef *GetPlftDef(ARSWDataBaseEntry &sw_db_entry);
    virtual SMP_PortSLToPrivateLFTMap **GetPlftMap(ARSWDataBaseEntry &sw_db_entry);
    virtual bool IsToSetPlftMap(ARSWDataBaseEntry &sw_db_entry);
    virtual void ToSetPlftMap(ARSWDataBaseEntry &sw_db_entry, bool is_to_set);
    virtual uint8_t *GetPlftNumberConfiguredPtr(ARSWDataBaseEntry &sw_db_entry);
    virtual PLFTMads *GetPlft(ARSWDataBaseEntry &sw_db_entry);

    virtual void RestorePriv(ARSWDataBaseEntry &sw_db_entry);
    virtual void ClearSwitchAlgotithmData(ARSWDataBaseEntry &sw_db_entry);

private:
    bool SetCapable();
    void SetPlftConfiguration();
    int SetPlftConfiguration(ARSWDataBaseEntry &db_entry);

    // Buils static unicast route info
    int BuildStaticRouteInfo();

    int BuildStaticRouteInfoToSwitch(SwDbEntryPrtList &bfs_q,
                                     ARSWDataBaseEntry &dst_db_entry);

    int ProcessNeighborsBfs(SwDbEntryPrtList &bfs_q,
                            ARSWDataBaseEntry *p_sw_db_entry,
                            ARSWDataBaseEntry *p_dst_db_entry);

    // Calculate Groups and pLFTs
    int CalculatePortGroups();

    void CalculateArGroups(ARSWDataBaseEntry &sw_db_entry,
                           u_int16_t *lid_to_sw_lid_mapping,
                           KdorAlgorithmData &algorithm_data,
                           SMP_ARGroupTable calculated_ar_group_table[]);

    void CalculateArPlfts(ARSWDataBaseEntry &sw_db_entry,
                          LidMapping *p_lid_mapping,
                          KdorAlgorithmData &algorithm_data,
                          SMP_AR_LFT calculated_ar_plft[]);

    void CalculateArPlft(ARSWDataBaseEntry &sw_db_entry,
                         KdorAlgorithmData &algorithm_data,
                         u_int8_t ucast_lft_port,
                         unsigned dest_lid,
                         int plft,
                         LidPortMappingPtrSX *p_lft_entry_ptr,
                         LidToKdorGroupDataMapConstIter &ar_kdor_groups_iter);

    uint8_t GetStaticUcastLftPort(ARSWDataBaseEntry &sw_db_entry,
                                  LidMapping *p_lid_mapping,
                                  KdorConnection *dst_sw_lid_to_conection[],
                                  uint8_t base_lid_to_port[],
                                  uint16_t dest_lid);

    kdor_turn_t GetTurnType(KdorConnection &from_connection,
                            KdorConnection &to_connection);
    //get static routing from given sw
    void BuildDstSwLidToConnection(ARSWDataBaseEntry &sw_db_entry,
                                   KdorConnection *dst_sw_lid_to_conection[]);
    void BuildKdorVl2VlPerOpVl();
    void BuildKdorPlftMap();

    void CalculateVl2VlMappingnOnSwitches();
    void CalculateVl2VlMappingnOnSwitch(ARSWDataBaseEntry &sw_db_entry);
    void CalculateVl2VlMappingnOnSwitch(ARSWDataBaseEntry &sw_db_entry,
                                        KdorConnection &from_connection,
                                        KdorConnection &to_connection);

    void SetVl2VlMappingn(ARSWDataBaseEntry &sw_db_entry,
                          uint8_t in_port,
                          uint8_t out_port,
                          SMP_SLToVLMappingTable *vl2vlmapping);
    void SetSl2VlMappingnOnSwitchEnhancedPort0(ARSWDataBaseEntry &sw_db_entry);

    void SetPlftMap(ARSWDataBaseEntry &sw_db_entry);
};

#define HC_PLANES_NUMBER 4

class ArHcAlgorithm : public ArKdorAlgorithm {

public:

    ArHcAlgorithm(osm_log_t *p_osm_log,
                  GuidToSWDataBaseEntry &sw_map,
                  AdaptiveRoutingManager &ar_mgr) :
        ArKdorAlgorithm(p_osm_log, sw_map, ar_mgr,
                        SUPPORT_HC,
                        HC_PLANES_NUMBER, HC_PLANES_NUMBER) {}

    virtual ~ArHcAlgorithm(){}

    virtual int SetDirection(osm_switch_t* src,
                             osm_switch_t* dst,
                             KdorConnection &connection);

    virtual ar_algorithm_t GetAlgorithm() { return AR_ALGORITHM_KDOR_HC; }

};

#endif //_AR_KDOR_ALGORITHM_H_
