/*
 * Copyright (c) 2018 Mellanox Technologies LTD. ALL RIGHTS RESERVED.
 * See file LICENSE for terms.
 */

#include <sys/sysinfo.h> // get_nprocs()
#include <errno.h>
#include <stdexcept>

#include "thread_pool.h"

void* thread_run(void* arg)
{
    ThreadPool *p_thread_pool = (ThreadPool*) arg;
    p_thread_pool->ThreadRun();
    return NULL;
}

int
ThreadPool::Init(uint16_t num_threads)
{
    if (m_init_) {
        return 0;
    }

    int rc;

    rc = pthread_mutex_init(&m_queue_lock_, NULL);
    if (rc) {
        OSM_AR_LOG(m_osm_log_, OSM_LOG_ERROR,
                   "ThreadPool failed to init mutex: %s\n",
                   strerror(errno));
        return rc;
    }

    rc = pthread_cond_init(&m_queue_cond_, NULL);
    if (rc) {
        pthread_mutex_destroy(&m_queue_lock_);
        OSM_AR_LOG(m_osm_log_, OSM_LOG_ERROR,
                   "ThreadPool failed to init condition variable: %s\n",
                   strerror(errno));
        return rc;
    }

    m_init_ = true;

    if (num_threads == 0) {
        num_threads = get_nprocs();

        if (num_threads == 0) {
            OSM_AR_LOG(m_osm_log_, OSM_LOG_INFO,
                       "Failed to get number of available processors. "
                       "Using %u threads.\n", DEFAULT_THREADPOOL_THREADS_NUMBER);
            num_threads = DEFAULT_THREADPOOL_THREADS_NUMBER;
        }
    }

    for (int i = 0; i < num_threads; ++i) {
        pthread_t worker_thread;
        rc = pthread_create(&worker_thread, NULL, thread_run, (void*) this);

        if (rc != 0) {
            OSM_AR_LOG(m_osm_log_, OSM_LOG_ERROR,
                       "Failed to create thread rc: %d\n", rc);
            return -1;
        }

        m_threads_.push_back(worker_thread);
    }

    OSM_AR_LOG(m_osm_log_, OSM_LOG_INFO,
               "ThreadPool init with %u threads\n",
               (uint16_t)m_threads_.size());

    return 0;
}

ThreadPool::~ThreadPool()
{
    if (!m_init_) {
        return;
    }

    pthread_mutex_lock(&m_queue_lock_);
    m_stop_ = true;
    pthread_mutex_unlock(&m_queue_lock_);

    //signall waiting threads so they could exit
    pthread_cond_broadcast(&m_queue_cond_);

    for (ListThreads::iterator iter = m_threads_.begin();
         iter!= m_threads_.end(); ++iter) {

        void *result;
        int rc = pthread_join(*iter, &result);
        if (rc) {
            OSM_AR_LOG(m_osm_log_, OSM_LOG_ERROR,
                       "ThreadPool pthread_join() failed: %s\n",
                       strerror(errno));
        }
    }

    pthread_mutex_destroy(&m_queue_lock_);
    pthread_cond_destroy(&m_queue_cond_);
    OSM_AR_LOG(m_osm_log_, OSM_LOG_INFO,"ThreadPool destroyed\n");
}

void
ThreadPool::AddTask(ThreadPoolTask *p_task)
{
    pthread_mutex_lock(&m_queue_lock_);

    m_task_queue_.push(p_task);
    pthread_cond_signal(&m_queue_cond_);

    pthread_mutex_unlock(&m_queue_lock_);
}

void
ThreadPool::ThreadRun()
{
    ThreadPoolTask* p_task = NULL;

    OSM_AR_LOG(m_osm_log_, OSM_LOG_DEBUG,"Start handle ThreadPool tasks\n");

    while (true) {

        pthread_mutex_lock(&m_queue_lock_);

        while (!m_stop_ && m_task_queue_.empty()) {
            pthread_cond_wait(&m_queue_cond_, &m_queue_lock_);
        }

        if (m_stop_) {
            pthread_mutex_unlock(&m_queue_lock_);
            break;
        }

        p_task = m_task_queue_.front();
        m_task_queue_.pop();
        pthread_mutex_unlock(&m_queue_lock_);

        p_task->Run();
    }

    OSM_AR_LOG(m_osm_log_, OSM_LOG_DEBUG,"Stop handle ThreadPool tasks\n");
}

void
ThreadPoolTasksCollection::Init()
{
    if (m_is_init_) {
        return;
    }

    int rc = pthread_mutex_init(&m_tasks_lock_, NULL);
    if (rc) {
        OSM_AR_LOG(m_osm_log_, OSM_LOG_ERROR,
                   "ThreadPoolTasksCollection failed to init mutex: %s\n",
                   strerror(errno));
        throw std::runtime_error("Failed to init mutex");
    }

    rc = pthread_cond_init(&m_tasks_cond_, NULL);
    if (rc) {
        OSM_AR_LOG(m_osm_log_, OSM_LOG_ERROR,
                   "ThreadPoolTasksCollection failed to init condition variable: %s\n",
                   strerror(errno));
        throw std::runtime_error("Failed to init condition variable");
    }
    m_is_init_ = true;
}

ThreadPoolTasksCollection::~ThreadPoolTasksCollection()
{
    if (m_is_init_) {
        pthread_mutex_destroy(&m_tasks_lock_);
        pthread_cond_destroy(&m_tasks_cond_);
    }
}

void
ThreadPoolTasksCollection::AddTaskToThreadPool(ThreadPool *p_thread_pool,
                                               ThreadPoolTask *p_task)
{
    pthread_mutex_lock(&m_tasks_lock_);
    p_task->SetCheckResult(true);
    m_num_tasks_in_progress_++;
    pthread_mutex_unlock(&m_tasks_lock_);

    p_thread_pool->AddTask(p_task);
}

void
ThreadPoolTasksCollection::WaitForTasks()
{
    if (!m_is_init_) {
        return;
    }

    pthread_mutex_lock(&m_tasks_lock_);

    while (m_num_tasks_in_progress_) {
        pthread_cond_wait(&m_tasks_cond_, &m_tasks_lock_);
    }

    pthread_mutex_unlock(&m_tasks_lock_);
}

void
ThreadPoolTasksCollection::OnTaskEnd()
{
    pthread_mutex_lock(&m_tasks_lock_);

    if (m_num_tasks_in_progress_) {
        m_num_tasks_in_progress_--;
    } else {
        OSM_AR_LOG(m_osm_log_, OSM_LOG_ERROR,
                   "-A- Unexpected zero num works in progress.\n");
    }

    if (m_num_tasks_in_progress_ == 0) {
        pthread_cond_signal(&m_tasks_cond_);
    }

    pthread_mutex_unlock(&m_tasks_lock_);
}
