/*                  - Mellanox Confidential and Proprietary -
 *
 *  Copyright (C) 2010-2019, Mellanox Technologies Ltd.  ALL RIGHTS RESERVED.
 *
 *  Except as specifically permitted herein, no portion of the information,
 *  including but not limited to object code and source code, may be reproduced,
 *  modified, distributed, republished or otherwise exploited in any form or by
 *  any means for any purpose without the prior written permission of Mellanox
 *  Technologies Ltd. Use of software subject to the terms and conditions
 *  detailed in the file "COPYING".
 *
 */

/* osm_config.h needs to be in all cpp files before all other includes */
#include <opensm/osm_config.h>

#include "ar_algorithm.h"
#include "ar_mgr.h"

int AdaptiveRoutingAlgorithm::BuildLidMapping(
    LidMapping &lid_mapping)
{
    OSM_AR_LOG_ENTER(m_p_osm_log);
    OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG, "BuildLidMapping\n");

    int rc = 0;
    osm_node_t *p_osm_node;
    cl_qmap_t *p_node_guid_tbl = &m_ar_mgr_.m_p_osm_subn->node_guid_tbl;
    u_int16_t base_lid;
    // Go over all nodes in osm database
    for (p_osm_node =  (osm_node_t *) cl_qmap_head(p_node_guid_tbl);
         p_osm_node != (osm_node_t *) cl_qmap_end(p_node_guid_tbl);
         p_osm_node =  (osm_node_t *) cl_qmap_next(&p_osm_node->map_item)) {

        if (p_osm_node->node_info.node_type == IB_NODE_TYPE_SWITCH) {
            base_lid = ADB2C_BE16_TO_CPU(osm_node_get_base_lid(p_osm_node, 0));
            OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
                       "Map Switch LID: %u\n", base_lid);

            lid_mapping.m_lid_to_sw_lid_mapping[base_lid] = base_lid;
            lid_mapping.m_lid_to_base_lid_mapping[base_lid] = base_lid;
            continue;
        }

        for( u_int8_t port_num = 1 ; port_num <= p_osm_node->node_info.num_ports ; port_num++) {
            osm_physp_t *p_physp = osm_node_get_physp_ptr(p_osm_node, port_num);
            osm_node_t *p_remote_node = osm_node_get_remote_node(p_osm_node, port_num, NULL);

            if (!p_physp || !osm_link_is_healthy(p_physp))
                continue;
            osm_physp_t *p_remote_physp = osm_physp_get_remote(p_physp);
            if (!p_remote_physp)
                continue;

            rc |= SetHcaLidMapping(p_physp, p_remote_node, lid_mapping);
        }
    }

    OSM_AR_LOG_RETURN(m_p_osm_log, rc);
}

int AdaptiveRoutingAlgorithm::SetHcaLidMapping(osm_physp_t *p_hca_physp,
                                               osm_node_t *p_remote_sw_node,
                                               LidMapping &lid_mapping)
{
    OSM_AR_LOG_ENTER(m_p_osm_log);

    u_int16_t node_lid = ADB2C_BE16_TO_CPU(osm_physp_get_base_lid(p_hca_physp));
    if (p_remote_sw_node->node_info.node_type != IB_NODE_TYPE_SWITCH ||
        node_lid == 0 ) {
        OSM_AR_LOG_RETURN(m_p_osm_log, -1);
    }

    uint8_t lmc = osm_physp_get_lmc(p_hca_physp);

    u_int16_t remote_node_lid = ADB2C_BE16_TO_CPU(osm_node_get_base_lid(p_remote_sw_node, 0));

    u_int16_t numLidsPerPort = (u_int16_t)((u_int16_t)1 << lmc);
    u_int16_t lid_limit = node_lid+numLidsPerPort;

    for (u_int16_t l = node_lid; l < lid_limit; l++) {
        OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
                   "Map (HCA/Router) LID: %u base LID: %u lmc:%u to SW LID: %u\n",
                   l, node_lid, lmc, remote_node_lid);

        lid_mapping.m_lid_to_sw_lid_mapping[l] = remote_node_lid;
        lid_mapping.m_lid_to_base_lid_mapping[l] = node_lid;
    }

    //Handle vports
    LidsList vlids_list;
    m_ar_mgr_.GetVlidsList(p_hca_physp, vlids_list);
    if (!vlids_list.empty()) {
        for (LidsList::const_iterator iter = vlids_list.begin();
             iter != vlids_list.end(); ++iter) {

            OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
                       "Map VLID %u to SW LID %u.\n",
                       *iter, remote_node_lid);
            lid_mapping.m_lid_to_sw_lid_mapping[*iter] = remote_node_lid;
            lid_mapping.m_lid_to_base_lid_mapping[*iter] = node_lid;
        }
    }

    OSM_AR_LOG_RETURN(m_p_osm_log, 0);
}

//TODO remove doplicated code: inline ar_mgr UpdateCalculatedArGroupTable
//!!! consider using AdaptiveRoutingManager::UpdateCalculatedArGroupTable instead
void AdaptiveRoutingAlgorithm::UpdateCalculatedArGroupTable(
    SMP_ARGroupTable calculated_ar_group_table[AR_GROUP_TABLE_NUM_BLOCKS],
    u_int16_t group_number,
    PortsBitset &group_bitmask,
    unsigned &calculated_max_pg_id)
{
    PortGroup_Block_Element *p_element =
        &calculated_ar_group_table[group_number / AR_GROUP_TABLE_BLOCK_SIZE].
            Group[group_number % AR_GROUP_TABLE_BLOCK_SIZE];

    group_bitmask.get_data(p_element->SubGroup_0,
                           p_element->SubGroup_1,
                           p_element->SubGroup_2,
                           p_element->SubGroup_3);

    if (group_number > calculated_max_pg_id)
        calculated_max_pg_id = group_number;
}

//TODO remove doplicated code: AdaptiveRoutingManager::ARGroupTableProcessDF()
int AdaptiveRoutingAlgorithm::GroupTableProcess()
{
    OSM_AR_LOG_ENTER(this->m_p_osm_log);
    OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_VERBOSE, "GroupTableProcess.\n");
    int rc = 0;

    map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it = m_sw_map_.begin();
    for ( ; sw_node_it != m_sw_map_.end(); ++sw_node_it) {

        if (sw_node_it->second.in_temporary_error)
            continue;

        if(!IsFeatureActive(sw_node_it->second, m_algorithm_feature_)) {
            OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG, "Switch GUID " U64H_FMT ", LID: %u " \
                    "Algorithm %u not supported or not enabled, GroupTableProcess skipped.\n",
                    sw_node_it->second.m_general_sw_info.m_guid,
                    sw_node_it->second.m_general_sw_info.m_lid,
                    m_algorithm_feature_);
            continue;
        }

        m_ar_mgr_.ARGroupTableProcess(
            sw_node_it->second,
            //sw_node_it->second.m_total_groups_number,
            sw_node_it->second.m_group_top,
            false, //reset_direction
            sw_node_it->second.m_ar_group_table);
    }

    //Receive all Set ARGroupTable reply
    m_ar_mgr_.m_ibis_obj.MadRecAll();

    if (m_ar_mgr_.m_ar_clbck.m_errcnt[AR_CLBCK_SET_AR_GROUP_TABLE]) {
        OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
                "Set AR Groups error count: %u.\n",
                m_ar_mgr_.m_ar_clbck.m_errcnt[AR_CLBCK_SET_AR_GROUP_TABLE]);
        m_ar_mgr_.m_is_temporary_error = true; //Need to recalculate AR groups and LFTS
    }

    OSM_AR_LOG_RETURN(this->m_p_osm_log, rc);
}

