/*                  - Mellanox Confidential and Proprietary -
 *
 *  Copyright (C) 2010-2013, Mellanox Technologies Ltd.  ALL RIGHTS RESERVED.
 *
 *  Except as specifically permitted herein, no portion of the information,
 *  including but not limited to object code and source code, may be reproduced,
 *  modified, distributed, republished or otherwise exploited in any form or by
 *  any means for any purpose without the prior written permission of Mellanox
 *  Technologies Ltd. Use of software subject to the terms and conditions
 *  detailed in the file "COPYING".
 *
 */

/* osm_config.h needs to be before all other includes */
#include <opensm/osm_config.h>
#include <errno.h>
#include <ctype.h>
#include <stdlib.h>
#include <sys/time.h>

#include <opensm/osm_sm.h>
#include <opensm/osm_opensm.h>
#include <opensm/osm_base.h>
#include <complib/cl_qmap.h>

#include <algorithm>

#include <iostream>
#include <sstream>

#include "ar_mgr.h"

#define IS_TEMP_ERR(rec_status)                     \
    ((rec_status == IBIS_MAD_STATUS_TIMEOUT)     || \
     (rec_status == IBIS_MAD_STATUS_BUSY)        || \
     (rec_status == IBIS_MAD_STATUS_GENERAL_ERR) || \
     (rec_status == IBIS_MAD_STATUS_SEND_FAILED))


#define ERR_DESC(rec_status, str) \
    (IS_TEMP_ERR(rec_status) ? "Temporary error" : str)

void AdaptiveRoutingClbck::HandleError(
    int rec_status,
    calback_errcnt_t err_type,
    supported_feature_t feature,
    ARSWDataBaseEntry *p_sw_entry) {


    if (IS_TEMP_ERR(rec_status)) {

        m_is_temporary_error = true;
        p_sw_entry->in_temporary_error = true;

    } else {

        m_errcnt[err_type]++;

        p_sw_entry->MarkSWNotSupport(feature, FABRIC_ERR);
    }

}



void GetARInfoClbckDlg(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    reinterpret_cast<AdaptiveRoutingClbck*>(clbck_data.m_p_obj)->
        GetARInfoClbck(clbck_data, rec_status, p_attribute_data);
}


void AdaptiveRoutingClbck::GetARInfoClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    OSM_AR_LOG_ENTER(this->m_p_osm_log);

    rec_status &= 0x00ff; //bits [8-15] are class specific - don't need them here

    ARSWDataBaseEntry *p_sw_entry = (ARSWDataBaseEntry *)clbck_data.m_data1;
    if (rec_status) {

        OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
                "%s Error getting ARInfo from Switch GUID " U64H_FMT ", LID %u, "  \
                "status=%u - %s\n",
                ARMGR_ERR_PREFIX6,
                p_sw_entry->m_general_sw_info.m_guid,
                p_sw_entry->m_general_sw_info.m_lid,
                rec_status,
                ERR_DESC(rec_status, "assuming no AR support"));

        HandleError(rec_status, AR_CLBCK_GET_AR_INFO, SUPPORT_AR, p_sw_entry);

        if (IS_TEMP_ERR(rec_status)) {
            p_sw_entry->temporary_error_num++;

            if (p_sw_entry->temporary_error_num >
                *m_p_max_timeout_num) {
                m_errcnt[AR_CLBCK_GET_AR_INFO]++;
                p_sw_entry->MarkSWNotSupport(SUPPORT_AR, FABRIC_ERR);

                OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
                        "Max ARInfo timeout (%u) exceeded."\
                        " Assuming no AR support\n",
                        *m_p_max_timeout_num);
            }
        } else
            p_sw_entry->temporary_error_num = 0;

    } else {
        struct adaptive_routing_info *p_smp_ar_info =
            (struct adaptive_routing_info *)p_attribute_data;

        p_sw_entry->m_ar_info = *p_smp_ar_info;
        p_sw_entry->temporary_error_num = 0;

        if (p_smp_ar_info->is_arn_sup ||
            p_smp_ar_info->is_frn_sup) {

            memset(&p_sw_entry->m_sub_group_direction, 0xFF,
                   sizeof(p_sw_entry->m_sub_group_direction));

            memset(&p_sw_entry->m_rn_xmit_port_mask, 0,
                   sizeof(p_sw_entry->m_rn_xmit_port_mask));

            if (p_smp_ar_info->rn_xmit_enabled) //mask is unknown
                p_sw_entry->m_to_set_rn_xmit_port_mask = true;
        }

        p_sw_entry->MarkSWSupport(SUPPORT_AR);
    }
    OSM_AR_LOG_RETURN_VOID(this->m_p_osm_log);
}


void GetRequiredARInfoCapClbckDlg(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    reinterpret_cast<AdaptiveRoutingClbck*>(clbck_data.m_p_obj)->
        GetRequiredARInfoCapClbck(clbck_data, rec_status, p_attribute_data);
}


void AdaptiveRoutingClbck::GetRequiredARInfoCapClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    OSM_AR_LOG_ENTER(this->m_p_osm_log);

    rec_status &= 0x00ff; //bits [8-15] are class specific - don't need them here

    ARSWDataBaseEntry *p_sw_entry = (ARSWDataBaseEntry *)clbck_data.m_data1;
    if (rec_status) {

        OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
                "%s Error getting ARInfo cap from Switch GUID " U64H_FMT ", LID %u, "  \
                "status=%u - %s\n",
                ARMGR_ERR_PREFIX6,
                p_sw_entry->m_general_sw_info.m_guid,
                p_sw_entry->m_general_sw_info.m_lid,
                rec_status,
                ERR_DESC(rec_status, "assuming no AR support"));

        HandleError(rec_status, AR_CLBCK_GET_AR_INFO_CAP, SUPPORT_AR, p_sw_entry);

        if (IS_TEMP_ERR(rec_status)) {
            p_sw_entry->temporary_error_num++;

            if (p_sw_entry->temporary_error_num >
                *m_p_max_timeout_num) {
                m_errcnt[AR_CLBCK_GET_AR_INFO_CAP]++;
                p_sw_entry->MarkSWNotSupport(SUPPORT_AR, FABRIC_ERR);

                OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
                        "Max ARInfo timeout (%u) exceeded."\
                        " Assuming no AR support\n",
                        *m_p_max_timeout_num);
            }
        } else
            p_sw_entry->temporary_error_num = 0;

    } else {
        struct adaptive_routing_info *p_smp_ar_info =
            (struct adaptive_routing_info *)p_attribute_data;
        p_sw_entry->temporary_error_num = 0;

        if (!AdaptiveRoutingManager::IsEqualSMPARInfo(p_smp_ar_info,
                                    &p_sw_entry->m_required_ar_info,
                                    true, true)) {
            OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
                    "Switch GUID " U64H_FMT ", LID %u " \
                    "Invalid GetARInfoCap response, no AR support\n",
                    p_sw_entry->m_general_sw_info.m_guid,
                    p_sw_entry->m_general_sw_info.m_lid);

            adaptive_routing_info *p_ar_info1 = p_smp_ar_info;
            adaptive_routing_info *p_ar_info2 =
                &p_sw_entry->m_required_ar_info;

            OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,"\n"
               "sub_grps_active       %u  %u\n"
               "no_fallback           %u  %u\n"
               "ageing_time_value     %u  %u\n"
               "by_transport_disable  %u  %u\n"
               "fr_enabled            %u  %u\n"
               "rn_xmit_enabled       %u  %u\n"
               "BySLEn                %u  %u\n"
               "enable_by_sl_mask     %u  %u\n",
               p_ar_info1->sub_grps_active ,p_ar_info2->sub_grps_active,
               p_ar_info1->no_fallback ,p_ar_info2->no_fallback,
               p_ar_info1->ageing_time_value ,p_ar_info2->ageing_time_value,
               p_ar_info1->by_transport_disable ,p_ar_info2->by_transport_disable,
               p_ar_info1->fr_enabled ,p_ar_info2->fr_enabled,
               p_ar_info1->rn_xmit_enabled ,p_ar_info2->rn_xmit_enabled,
               p_ar_info1->by_sl_en ,p_ar_info2->by_sl_en,
               p_ar_info1->enable_by_sl_mask , p_ar_info2->enable_by_sl_mask);

            p_sw_entry->MarkSWNotSupport(SUPPORT_AR, AR_GROUP_CAP_ZERO);
            m_errcnt[AR_CLBCK_GET_AR_INFO_CAP]++;

        } else if (p_smp_ar_info->group_cap <
                   p_sw_entry->m_general_sw_info.m_num_ports) {
            // Also we check that group cap is >= number of switch ports.
            OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
                    "Switch GUID " U64H_FMT ", LID %u, "    \
                    "group capability is %u but number of ports is %u. "
                    "This is not allowed, no AR support\n",
                    p_sw_entry->m_general_sw_info.m_guid,
                    p_sw_entry->m_general_sw_info.m_lid,
                    p_smp_ar_info->group_cap,
                    p_sw_entry->m_general_sw_info.m_num_ports);
            p_sw_entry->MarkSWNotSupport(SUPPORT_AR, AR_GROUP_LESS_THAN_PORTS);
            m_errcnt[AR_CLBCK_GET_AR_INFO_CAP]++;

        } else {

            u_int16_t ds_group_cap = AR_GROUP_TABLE_NUM_BLOCKS *
                    AR_GROUP_TABLE_BLOCK_SIZE /
                    (p_smp_ar_info->sub_grps_active+1);

            if (p_smp_ar_info->group_cap > ds_group_cap)
                // Also we check that group cap is >= number of switch ports.
                OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_VERBOSE,
                        "Switch GUID " U64H_FMT ", LID %u, "    \
                        "group capability is %u. only %u groups can be used.\n",
                        p_sw_entry->m_general_sw_info.m_guid,
                        p_sw_entry->m_general_sw_info.m_lid,
                        p_smp_ar_info->group_cap, ds_group_cap);

            p_sw_entry->m_required_ar_info.group_cap =
                p_smp_ar_info->group_cap;
        }

    }
    OSM_AR_LOG_RETURN_VOID(this->m_p_osm_log);
}

void SetARInfoClbckDlg(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    reinterpret_cast<AdaptiveRoutingClbck*>(clbck_data.m_p_obj)->
        SetARInfoClbck(clbck_data, rec_status, p_attribute_data);
}


void AdaptiveRoutingClbck::SetARInfoClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    OSM_AR_LOG_ENTER(this->m_p_osm_log);

    rec_status &= 0x00ff; //bits [8-15] are class specific - don't need them here

    ARSWDataBaseEntry *p_sw_entry = (ARSWDataBaseEntry *)clbck_data.m_data1;

    if (rec_status) {
        OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
                "%s Error setting ARInfo to "  \
                "Switch GUID " U64H_FMT ", LID %u, "    \
                "status=%u - %s\n",
                ARMGR_ERR_PREFIX7,
                p_sw_entry->m_general_sw_info.m_guid,
                p_sw_entry->m_general_sw_info.m_lid,
                rec_status,
                ERR_DESC(rec_status, "assuming no AR support"));

        HandleError(rec_status, AR_CLBCK_SET_AR_INFO, SUPPORT_AR, p_sw_entry);

    } else {
        //if this switch was not configured on prev cycle
        //we set here ar_configured = 1 in order to enforce SM configure this SW
        //in case we fail to completly configure this sw
        p_sw_entry->m_general_sw_info.m_p_osm_sw->ar_configured = 1;

        struct adaptive_routing_info *p_smp_ar_info =
            (struct adaptive_routing_info *)p_attribute_data;

        u_int16_t enSLMask = (unsigned)(intptr_t)clbck_data.m_data2;

        if(enSLMask != ALL_SL_VL_AR_ENABLED &&
           (p_smp_ar_info->by_sl_en != 1 ||
            p_smp_ar_info->enable_by_sl_mask != enSLMask)) {
            OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
                    "Failed to set  ARInfo by_sl_en %u to "  \
                    "Switch GUID " U64H_FMT ", LID %u, \n",
                    enSLMask,
                    p_sw_entry->m_general_sw_info.m_guid,
                    p_sw_entry->m_general_sw_info.m_lid);
            m_errcnt[AR_CLBCK_SET_AR_INFO]++;
            OSM_AR_LOG_RETURN_VOID(this->m_p_osm_log);
        }

        // save the new data
        p_sw_entry->m_ar_info = *p_smp_ar_info;
        string str = (p_sw_entry->m_ar_info.e) ? "enabled" : "disabled";
        OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
                "Switch GUID " U64H_FMT ", LID %u: AR is %s,"
                " SubGrps:%d by_sl_en:%u enable_by_sl_mask:0x%x"
                " by_transport_disable:0x%x.\n",
                p_sw_entry->m_general_sw_info.m_guid,
                p_sw_entry->m_general_sw_info.m_lid,
                str.c_str(),
                p_smp_ar_info->sub_grps_active+1,
                p_smp_ar_info->by_sl_en,
                p_smp_ar_info->enable_by_sl_mask,
                p_smp_ar_info->by_transport_disable);

    }
    OSM_AR_LOG_RETURN_VOID(this->m_p_osm_log);
}

void SetARGroupTableClbckDlg(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    reinterpret_cast<AdaptiveRoutingClbck*>(clbck_data.m_p_obj)->
        SetARGroupTableClbck(clbck_data, rec_status, p_attribute_data);
}


void AdaptiveRoutingClbck::SetARGroupTableClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void */*p_attribute_data*/)
{
    OSM_AR_LOG_ENTER(this->m_p_osm_log);

    rec_status &= 0x00ff; //bits [8-15] are class specific - don't need them here

    ARSWDataBaseEntry *p_sw_entry = (ARSWDataBaseEntry *)clbck_data.m_data1;
    unsigned group_block = (unsigned)(intptr_t)clbck_data.m_data2;
    u_int8_t reset_attr = (u_int8_t)(intptr_t)clbck_data.m_data3;

    if (rec_status) {
        OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
                "%s Error setting ARGroupTable (block=%u) to "  \
                "Switch GUID " U64H_FMT ", LID %u " \
                "(Status=%u) - %s\n",
                ARMGR_ERR_PREFIX10, group_block,
                p_sw_entry->m_general_sw_info.m_guid,
                p_sw_entry->m_general_sw_info.m_lid,
                rec_status,
                ERR_DESC(rec_status, "assuming no AR support"));

        HandleError(rec_status, AR_CLBCK_SET_AR_GROUP_TABLE,
                    SUPPORT_AR, p_sw_entry);

    } else {
        // Mark this block was set with success and therefore no need to set in next time
        p_sw_entry->m_to_set_group_table[group_block] = false;
        p_sw_entry->m_is_group_table_valid[group_block] = true;

        if (reset_attr) {
            for (u_int8_t subGroupOffset = 0;
                 subGroupOffset < AR_GROUP_TABLE_BLOCK_SIZE;
                 ++subGroupOffset) {

                if (!(reset_attr & (u_int8_t)(0x1<<subGroupOffset)))
                    continue;

                u_int8_t block_idx =
                    (u_int8_t)((group_block * AR_GROUP_TABLE_BLOCK_SIZE +
                                subGroupOffset) /
                               RN_SUB_GROUP_DIRECTION_BLOCK_SIZE);

                if (block_idx >= RN_SUB_GROUP_DIRECTION_NUM_BLOCKS) {
                    OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
                               "Error setting ARGroupTable on "
                               "Switch GUID " U64H_FMT ", LID %u "
                               "invalid sub group direction block_idx=%u \n",
                               p_sw_entry->m_general_sw_info.m_guid,
                               p_sw_entry->m_general_sw_info.m_lid,
                               block_idx);
                    break;
                }

                u_int8_t element_idx =
                    (u_int8_t)((group_block * AR_GROUP_TABLE_BLOCK_SIZE +
                                subGroupOffset) %
                               RN_SUB_GROUP_DIRECTION_BLOCK_SIZE);

                p_sw_entry->m_sub_group_direction[block_idx].
                    direction_of_sub_group[element_idx] = 0;
            }
        }

    }
    OSM_AR_LOG_RETURN_VOID(this->m_p_osm_log);
}

void ARCopyGroupTableClbckDlg(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    reinterpret_cast<AdaptiveRoutingClbck*>(clbck_data.m_p_obj)->
        ARCopyGroupTableClbck(clbck_data, rec_status, p_attribute_data);
}


void AdaptiveRoutingClbck::ARCopyGroupTableClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    OSM_AR_LOG_ENTER(this->m_p_osm_log);

    rec_status &= 0x00ff; //bits [8-15] are class specific - don't need them here

    ARSWDataBaseEntry *p_sw_entry = (ARSWDataBaseEntry *)clbck_data.m_data1;
    u_int16_t copy_group = (u_int16_t)(intptr_t)clbck_data.m_data2;
    bool copy_direction = (bool)(intptr_t)clbck_data.m_data2;

    if (rec_status) {
        OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
                "Error setting ar_group_table_copy "
                "(copy_group=%u) to "
                "Switch GUID " U64H_FMT ", LID %u "
                "(Status=%u) - %s\n",
                copy_group,
                p_sw_entry->m_general_sw_info.m_guid,
                p_sw_entry->m_general_sw_info.m_lid,
                rec_status,
                ERR_DESC(rec_status, "assuming no AR support"));

        HandleError(rec_status, AR_CLBCK_SET_AR_GROUP_TABLE,
                    SUPPORT_AR, p_sw_entry);

    } else {
        // Mark this block was set with success and therefore no need to set in next time
        struct adaptive_routing_group_table_copy *p_ar_group_table_copy =
            (struct adaptive_routing_group_table_copy *)p_attribute_data;

        u_int8_t *p_copy_from_direction =
            p_sw_entry->GetSubGroupDirectionPtr(copy_group,0);
        PortGroup_Block_Element *p_copy_from =
            p_sw_entry->GetSubGroupPtr(copy_group,0);

        u_int8_t *p_copy_to_direction = NULL;
        PortGroup_Block_Element *p_copy_to = NULL;

        if (p_copy_from_direction == NULL || p_copy_from == NULL) {
            OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
                       "Error setting ar_group_table_copy on "
                       "Switch GUID " U64H_FMT ", LID %u "
                       "invalid copy_from_group=%u \n",
                       p_sw_entry->m_general_sw_info.m_guid,
                       p_sw_entry->m_general_sw_info.m_lid,
                       copy_group);
            OSM_AR_LOG_RETURN_VOID(this->m_p_osm_log);
        }

        for (int i=0 ; i < AR_COPY_GROUP_BLOCK_SIZE ; i++) {
            adaptive_routing_group_table_copy_element *p_element =
                &p_ar_group_table_copy->element[i];

            if (p_element->copy_from_index == 0)
                break;

            for (u_int16_t j=p_element->copy_from_index ;
                 j <= p_element->copy_to_index ; j++) {

                p_copy_to_direction = p_sw_entry->GetSubGroupDirectionPtr(j,0);
                p_copy_to = p_sw_entry->GetSubGroupPtr(j,0);

                if (p_copy_to_direction == NULL || p_copy_to == NULL) {
                    OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
                               "Error setting ar_group_table_copy on "
                               "Switch GUID " U64H_FMT ", LID %u "
                               "invalid copy_to_group=%u \n",
                               p_sw_entry->m_general_sw_info.m_guid,
                               p_sw_entry->m_general_sw_info.m_lid,
                               j);
                    break;
                }

                memcpy(p_copy_to, p_copy_from,
                       (p_sw_entry->m_ar_info.sub_grps_active + 1)*
                       sizeof(PortGroup_Block_Element));

                if (copy_direction)
                    memcpy(p_copy_to_direction, p_copy_from_direction,
                           (p_sw_entry->m_ar_info.sub_grps_active + 1));
            }
        }
    }
    OSM_AR_LOG_RETURN_VOID(this->m_p_osm_log);
}

void SetARLFTTableClbckDlg(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    reinterpret_cast<AdaptiveRoutingClbck*>(clbck_data.m_p_obj)->
        SetARLFTTableClbck(clbck_data, rec_status, p_attribute_data);
}


void AdaptiveRoutingClbck::SetARLFTTableClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void */*p_attribute_data*/)
{
    OSM_AR_LOG_ENTER(this->m_p_osm_log);

    rec_status &= 0x00ff; //bits [8-15] are class specific - don't need them here

    ARSWDataBaseEntry *p_sw_entry = (ARSWDataBaseEntry *)clbck_data.m_data1;
    unsigned lft_block = (unsigned)(intptr_t)clbck_data.m_data2;
    bool *to_set_lft_table = (bool*)clbck_data.m_data3;

    if (rec_status) {
        OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
                "%s Error setting ARLFTTable (block=%u) to "  \
                "Switch GUID " U64H_FMT ", LID %u " \
                "(Status=%u) - %s\n",
                ARMGR_ERR_PREFIX11, lft_block,
                p_sw_entry->m_general_sw_info.m_guid,
                p_sw_entry->m_general_sw_info.m_lid,
                rec_status,
                ERR_DESC(rec_status, "assuming no AR support"));

        AR_MGR_LOG(TT_LOG_LEVEL_ERROR,
                "%s Error setting ARLFTTable (block=%u) to "  \
                "Switch GUID " U64H_FMT ", LID %u " \
                "(Status=%u) - %s\n",
                ARMGR_ERR_PREFIX11, lft_block,
                p_sw_entry->m_general_sw_info.m_guid,
                p_sw_entry->m_general_sw_info.m_lid,
                rec_status,
                ERR_DESC(rec_status, "assuming no AR support"));

        HandleError(rec_status, AR_CLBCK_SET_AR_LFT_TABLE, SUPPORT_AR, p_sw_entry);

    } else {
        // Mark this block was set with success and therefore no need to set in next time
        to_set_lft_table[lft_block] = false;
        //p_sw_entry->temporary_error_num = 0;
    }
    OSM_AR_LOG_RETURN_VOID(this->m_p_osm_log);
}

void SetRNGenStringClbckDlg(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    reinterpret_cast<AdaptiveRoutingClbck*>(clbck_data.m_p_obj)->
        SetRNGenStringClbck(clbck_data, rec_status, p_attribute_data);
}

void AdaptiveRoutingClbck::SetRNGenStringClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    OSM_AR_LOG_ENTER(this->m_p_osm_log);

    rec_status &= 0x00ff; //bits [8-15] are class specific - don't need them here

    ARSWDataBaseEntry *p_sw_entry = (ARSWDataBaseEntry *)clbck_data.m_data1;
    //unsigned lft_block = (unsigned)(intptr_t)clbck_data.m_data2;
    //bool *to_set_lft_table = (bool*)clbck_data.m_data3;

    if (rec_status) {
        OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
                "Error setting rn_gen_string_tbl to "  \
                "Switch GUID " U64H_FMT ", LID %u " \
                "(Status=%u) - %s\n",
                p_sw_entry->m_general_sw_info.m_guid,
                p_sw_entry->m_general_sw_info.m_lid,
                rec_status,
                ERR_DESC(rec_status, "assuming no RN support"));

        AR_MGR_LOG(TT_LOG_LEVEL_ERROR,
                "Error setting rn_gen_string_tbl to "  \
                "Switch GUID " U64H_FMT ", LID %u " \
                "(Status=%u) - %s\n",
                p_sw_entry->m_general_sw_info.m_guid,
                p_sw_entry->m_general_sw_info.m_lid,
                rec_status,
                ERR_DESC(rec_status, "assuming no RN support"));

        HandleError(rec_status, AR_CLBCK_SET_RN_GEN_STRING,
                    SUPPORT_RN, p_sw_entry);

    } else {
        struct rn_gen_string_tbl *p_rn_gen_string_table =
            (struct rn_gen_string_tbl *)p_attribute_data;
        p_sw_entry->m_rn_gen_string =
            p_rn_gen_string_table->element[0].rn_gen_string;

    }
    OSM_AR_LOG_RETURN_VOID(this->m_p_osm_log);
}

void SetRNGenBySubGroupPriorityClbckDlg(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    reinterpret_cast<AdaptiveRoutingClbck*>(clbck_data.m_p_obj)->
        SetRNGenBySubGroupPriorityClbck(clbck_data, rec_status, p_attribute_data);
}

void AdaptiveRoutingClbck::SetRNGenBySubGroupPriorityClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    OSM_AR_LOG_ENTER(this->m_p_osm_log);

    rec_status &= 0x00ff; //bits [8-15] are class specific - don't need them here

    ARSWDataBaseEntry *p_sw_entry = (ARSWDataBaseEntry *)clbck_data.m_data1;

    if (rec_status) {
        OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
                "Error setting RNGenBySubGroupPriority to "  \
                "Switch GUID " U64H_FMT ", LID %u " \
                "(Status=%u) - %s\n",
                p_sw_entry->m_general_sw_info.m_guid,
                p_sw_entry->m_general_sw_info.m_lid,
                rec_status,
                ERR_DESC(rec_status, "assuming no RN support"));

        AR_MGR_LOG(TT_LOG_LEVEL_ERROR,
                "Error setting RNGenBySubGroupPriority to "  \
                "Switch GUID " U64H_FMT ", LID %u " \
                "(Status=%u) - %s\n",
                p_sw_entry->m_general_sw_info.m_guid,
                p_sw_entry->m_general_sw_info.m_lid,
                rec_status,
                ERR_DESC(rec_status, "assuming no RN support"));

        HandleError(rec_status, AR_CLBCK_SET_RN_GEN_BY_SG_PRIORITY,
                    SUPPORT_RN, p_sw_entry);

    } else {
        struct rn_gen_by_sub_group_prio *p_gen_by_sg_priority =
            (struct rn_gen_by_sub_group_prio *)p_attribute_data;
        memcpy(&p_sw_entry->m_rn_gen_by_sub_group_priority,
               p_gen_by_sg_priority,
               sizeof(rn_gen_by_sub_group_prio));
    }
    OSM_AR_LOG_RETURN_VOID(this->m_p_osm_log);
}

void SetRNRcvStringClbckDlg(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    reinterpret_cast<AdaptiveRoutingClbck*>(clbck_data.m_p_obj)->
        SetRNRcvStringClbck(clbck_data, rec_status, p_attribute_data);
}

void AdaptiveRoutingClbck::SetRNRcvStringClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void */*p_attribute_data*/)
{
    OSM_AR_LOG_ENTER(this->m_p_osm_log);

    rec_status &= 0x00ff; //bits [8-15] are class specific - don't need them here

    ARSWDataBaseEntry *p_sw_entry = (ARSWDataBaseEntry *)clbck_data.m_data1;
    //unsigned lft_block = (unsigned)(intptr_t)clbck_data.m_data2;
    //bool *to_set_lft_table = (bool*)clbck_data.m_data3;

    if (rec_status) {
        OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
                "Error setting RNRcvString to "  \
                "Switch GUID " U64H_FMT ", LID %u " \
                "(Status=%u) - %s\n",
                p_sw_entry->m_general_sw_info.m_guid,
                p_sw_entry->m_general_sw_info.m_lid,
                rec_status,
                ERR_DESC(rec_status, "assuming no RN support"));

        AR_MGR_LOG(TT_LOG_LEVEL_ERROR,
                "Error setting RNRcvString to "  \
                "Switch GUID " U64H_FMT ", LID %u " \
                "(Status=%u) - %s\n",
                p_sw_entry->m_general_sw_info.m_guid,
                p_sw_entry->m_general_sw_info.m_lid,
                rec_status,
                ERR_DESC(rec_status, "assuming no RN support"));

        HandleError(rec_status, AR_CLBCK_SET_RN_RCV_STRING,
                    SUPPORT_RN, p_sw_entry);

    } else
        p_sw_entry->m_to_set_rn_rcv_string = false;

    OSM_AR_LOG_RETURN_VOID(this->m_p_osm_log);
}

void SetRNXmitPortMaskClbckDlg(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    reinterpret_cast<AdaptiveRoutingClbck*>(clbck_data.m_p_obj)->
        SetRNXmitPortMaskClbck(clbck_data, rec_status, p_attribute_data);
}

void AdaptiveRoutingClbck::SetRNXmitPortMaskClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void */*p_attribute_data*/)
{
    OSM_AR_LOG_ENTER(this->m_p_osm_log);

    rec_status &= 0x00ff; //bits [8-15] are class specific - don't need them here

    ARSWDataBaseEntry *p_sw_entry = (ARSWDataBaseEntry *)clbck_data.m_data1;
    //unsigned lft_block = (unsigned)(intptr_t)clbck_data.m_data2;
    //bool *to_set_lft_table = (bool*)clbck_data.m_data3;

    if (rec_status) {
        OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
                "Error setting RNXmitPortMask to "  \
                "Switch GUID " U64H_FMT ", LID %u " \
                "(Status=%u) - %s\n",
                p_sw_entry->m_general_sw_info.m_guid,
                p_sw_entry->m_general_sw_info.m_lid,
                rec_status,
                ERR_DESC(rec_status, "assuming no RN support"));

        AR_MGR_LOG(TT_LOG_LEVEL_ERROR,
                "Error setting RNXmitPortMask to "  \
                "Switch GUID " U64H_FMT ", LID %u " \
                "(Status=%u) - %s\n",
                p_sw_entry->m_general_sw_info.m_guid,
                p_sw_entry->m_general_sw_info.m_lid,
                rec_status,
                ERR_DESC(rec_status, "assuming no RN support"));

        HandleError(rec_status, AR_CLBCK_SET_RN_GEN_PORT_MASK,
                    SUPPORT_RN, p_sw_entry);

    } else
        p_sw_entry->m_to_set_rn_xmit_port_mask = false;

    OSM_AR_LOG_RETURN_VOID(this->m_p_osm_log);
}

void GetPrivateLFTInfoClbckDlg(
                          const clbck_data_t &clbck_data,
                          int rec_status,
                          void *p_attribute_data)
{
    reinterpret_cast<AdaptiveRoutingClbck*>(clbck_data.m_p_obj)->
        GetPrivateLFTInfoClbck(clbck_data, rec_status, p_attribute_data);
}

void AdaptiveRoutingClbck::GetPrivateLFTInfoClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    OSM_AR_LOG_ENTER(this->m_p_osm_log);

    rec_status &= 0x00ff; //bits [8-15] are class specific - don't need them here

    ARSWDataBaseEntry *p_sw_entry = (ARSWDataBaseEntry *)clbck_data.m_data1;

    Description_Mode_Block_Element *p_block_element = NULL;
    u_int8_t Mode = 0;

    if (rec_status) {
        OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
                "Error getting PrivateLFTInfo from Switch GUID " U64H_FMT ", LID %u, "  \
                "status=%u\n",
                p_sw_entry->m_general_sw_info.m_guid,
                p_sw_entry->m_general_sw_info.m_lid,
                rec_status);

        HandleError(rec_status, AR_CLBCK_GET_PLFT_INFO, SUPPORT_AR, p_sw_entry);


    } else {
        struct SMP_PrivateLFTInfo *p_plft_info =
            (struct SMP_PrivateLFTInfo *)p_attribute_data;

        u_int8_t ModeCap = p_plft_info->ModeCap;

        if (ModeCap == 0) {
            OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
                    "Switch GUID " U64H_FMT ", LID %u, "  \
                    "Do not support PrivateLFT\n",
                    p_sw_entry->m_general_sw_info.m_guid,
                    p_sw_entry->m_general_sw_info.m_lid);

            p_sw_entry->MarkSWNotSupport(m_algorithm_feature, DF_PLFT_CAP_ZERO);
            m_errcnt[AR_CLBCK_GET_PLFT_INFO]++;

            OSM_AR_LOG_RETURN_VOID(this->m_p_osm_log);
        }

        u_int8_t NumOfLFTables = 0;

        for (int i = 0 ; i < ModeCap ; i++) {
            p_block_element =
                &p_plft_info->Description_Mode[i];

            if ( p_block_element->NumOfLFTables < MIN_DF_PLFT_NUMBER ||
                 p_block_element->LFTableSize < MIN_PLFT_TABLE_SIZE )
                continue;

            if (p_block_element->NumOfLFTables > NumOfLFTables) {
                NumOfLFTables = p_block_element->NumOfLFTables;
                Mode = i;
                if ( NumOfLFTables >= MAX_DF_PLFT_NUMBER) //!!! TAMIR add to OSM modified from MAX_PLFT_NUMBER
                    break;
            }
        }

        if (NumOfLFTables == 0) {
            OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
                    "Switch GUID " U64H_FMT ", LID %u, "  \
                    "Do not have sufficient PrivateLFT cap: (num tables , size). \n",
                    p_sw_entry->m_general_sw_info.m_guid,
                    p_sw_entry->m_general_sw_info.m_lid);

            p_sw_entry->MarkSWNotSupport(m_algorithm_feature, DF_PLFT_CAP_LESS_THAN_MIN);
            m_errcnt[AR_CLBCK_GET_PLFT_INFO]++;

            OSM_AR_LOG_RETURN_VOID(this->m_p_osm_log);

        } else {
            //before reading plft_info ModeCap=Active_Mode=0
            //after read, set Active_Mode to be the desired mode
            //after set, set  ModeCap!=0
            p_sw_entry->m_p_df_data->m_plft_info.Active_Mode = Mode+1;

            if (p_plft_info->Active_Mode ==
                p_sw_entry->m_p_df_data->m_plft_info.Active_Mode)
                // already set - no need to set
                p_sw_entry->m_p_df_data->m_plft_info.ModeCap =
                    p_plft_info->ModeCap;


            OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
                "PLFT Info Get Switch GUID " U64H_FMT ", LID %u:"
                " desired mode:%d Active_Mode:%d.\n",
                p_sw_entry->m_general_sw_info.m_guid,
                p_sw_entry->m_general_sw_info.m_lid,
                p_sw_entry->m_p_df_data->m_plft_info.Active_Mode,
                p_plft_info->Active_Mode);
        }
    }
    OSM_AR_LOG_RETURN_VOID(this->m_p_osm_log);
}

void SetPrivateLFTInfoClbckDlg(
                          const clbck_data_t &clbck_data,
                          int rec_status,
                          void *p_attribute_data)
{
    reinterpret_cast<AdaptiveRoutingClbck*>(clbck_data.m_p_obj)->
        SetPrivateLFTInfoClbck(clbck_data, rec_status, p_attribute_data);
}

//!!! TAMIR add to OSM
//////////////////////////////////////////////////////////////////////
///
/// clbck_data.m_data1: (ARCADataBaseEntry *)
/// clbck_data.m_data2: (SMP_PrivateLFTInfo *)
///
void AdaptiveRoutingClbck::SetPrivateLFTInfoClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    OSM_AR_LOG_ENTER(this->m_p_osm_log);

    rec_status &= 0x00ff; //bits [8-15] are class specific - don't need them here

    ARSWDataBaseEntry *p_sw_entry = (ARSWDataBaseEntry *)clbck_data.m_data1;

    if (rec_status) {
        OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
                "Error getting PrivateLFTInfo from Switch GUID " U64H_FMT ", LID %u, "  \
                "status=%u\n",
                p_sw_entry->m_general_sw_info.m_guid,
                p_sw_entry->m_general_sw_info.m_lid,
                rec_status);

        HandleError(rec_status, AR_CLBCK_SET_PLFT_INFO,
                    m_algorithm_feature, p_sw_entry);

    } else {
        struct SMP_PrivateLFTInfo *p_plft_info =
            (struct SMP_PrivateLFTInfo *)p_attribute_data;

        SMP_PrivateLFTInfo *p_sw_plft_info = (SMP_PrivateLFTInfo *)clbck_data.m_data2;
        *p_sw_plft_info = *p_plft_info;

        OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
                "PLFT Info Set Switch GUID " U64H_FMT ", LID %u:"
                " Active_Mode:%d.\n",
                p_sw_entry->m_general_sw_info.m_guid,
                p_sw_entry->m_general_sw_info.m_lid,
                p_sw_plft_info->Active_Mode);
    }
    OSM_AR_LOG_RETURN_VOID(this->m_p_osm_log);
}

void SetPrivateLFTDefClbckDlg(
                          const clbck_data_t &clbck_data,
                          int rec_status,
                          void *p_attribute_data)
{
    reinterpret_cast<AdaptiveRoutingClbck*>(clbck_data.m_p_obj)->
        SetPrivateLFTDefClbck(clbck_data, rec_status, p_attribute_data);
}

//m_data1 - ARSWDataBaseEntry *
//m_data2 - uint8_t numper of plfts value to set
//m_data3 - *uint8_t numper of plfts ptr to set

void AdaptiveRoutingClbck::SetPrivateLFTDefClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void */*p_attribute_data*/)
{
    OSM_AR_LOG_ENTER(this->m_p_osm_log);

    rec_status &= 0x00ff; //bits [8-15] are class specific - don't need them here

    ARSWDataBaseEntry *p_sw_entry = (ARSWDataBaseEntry *)clbck_data.m_data1;

    if (rec_status) {
        OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
                "Error setting PrivateLFTDef from Switch GUID " U64H_FMT ", LID %u, "  \
                "status=%u\n",
                p_sw_entry->m_general_sw_info.m_guid,
                p_sw_entry->m_general_sw_info.m_lid,
                rec_status);

        HandleError(rec_status, AR_CLBCK_SET_PLFT_DEF,
                    m_algorithm_feature, p_sw_entry);

    } else {
        //!!! TAMIR add to OSM
        uint8_t *p_plft_number = (uint8_t *)clbck_data.m_data3;
        *p_plft_number = (u_int8_t)(intptr_t)clbck_data.m_data2;
    }
    OSM_AR_LOG_RETURN_VOID(this->m_p_osm_log);
}

void SetPrivateLFTMapClbckDlg(
                          const clbck_data_t &clbck_data,
                          int rec_status,
                          void *p_attribute_data)
{
    reinterpret_cast<AdaptiveRoutingClbck*>(clbck_data.m_p_obj)->
        SetPrivateLFTMapClbck(clbck_data, rec_status, p_attribute_data);
}

//!!! TAMIR add to OSM
//////////////////////////////////////////////////////////////////////
///
/// clbck_data.m_data1: (ARCADataBaseEntry *)
/// clbck_data.m_data2: u_int8_t pLFTID
/// clbck_data.m_data3: (PLFTMads *)
///
void AdaptiveRoutingClbck::SetPrivateLFTMapClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    OSM_AR_LOG_ENTER(this->m_p_osm_log);

    rec_status &= 0x00ff; //bits [8-15] are class specific - don't need them here

    ARSWDataBaseEntry *p_sw_entry = (ARSWDataBaseEntry *)clbck_data.m_data1;
    u_int8_t pLFTID = (u_int8_t)(intptr_t)clbck_data.m_data2;

    if (rec_status) {
        OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
                "Error setting PrivateLFTMap from Switch GUID "\
                U64H_FMT ", LID %u, pLFTID:%u status=%u\n",
                p_sw_entry->m_general_sw_info.m_guid,
                p_sw_entry->m_general_sw_info.m_lid,
                pLFTID,
                rec_status);

        HandleError(rec_status, AR_CLBCK_SET_PLFT_MAP,
                    m_algorithm_feature, p_sw_entry);

    } else {
        PLFTMads *p_plft_mads = (PLFTMads *)clbck_data.m_data3;
        struct SMP_PrivateLFTMap * p_plft_map =
            (struct SMP_PrivateLFTMap *)p_attribute_data;

        if (p_plft_map->LFT_TopEn != 1 ||
            p_plft_map->LFT_Top !=
            p_plft_mads->m_max_lid) {

            OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
                    "PrivateLFTMap Set failed for Switch GUID " U64H_FMT ", LID %u, "  \
                    "pLFTID:%u TopEn=%u Mad_LFT_Top:%u Req_LFT_Top:%u\n",
                    p_sw_entry->m_general_sw_info.m_guid,
                    p_sw_entry->m_general_sw_info.m_lid,
                    pLFTID,
                    p_plft_map->LFT_TopEn,
                    p_plft_map->LFT_Top,
                    p_plft_mads->m_max_lid);

            p_sw_entry->MarkSWNotSupport(m_algorithm_feature, DF_INVALID_PLFT_TOP);
            m_errcnt[AR_CLBCK_SET_PLFT_MAP]++;

        } else
            p_plft_mads->m_set_lft_top = false;
    }
    OSM_AR_LOG_RETURN_VOID(this->m_p_osm_log);
}

void SetPortSlToPLFTMapClbckDlg(
                          const clbck_data_t &clbck_data,
                          int rec_status,
                          void *p_attribute_data)
{
    reinterpret_cast<AdaptiveRoutingClbck*>(clbck_data.m_p_obj)->
        SetPortSlToPLFTMapClbck(clbck_data, rec_status, p_attribute_data);
}

void AdaptiveRoutingClbck::SetPortSlToPLFTMapClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void */*p_attribute_data*/)
{
    OSM_AR_LOG_ENTER(this->m_p_osm_log);

    rec_status &= 0x00ff; //bits [8-15] are class specific - don't need them here

    ARSWDataBaseEntry *p_sw_entry = (ARSWDataBaseEntry *)clbck_data.m_data1;
    u_int8_t port_block = (u_int8_t)(intptr_t)clbck_data.m_data2;

    if (rec_status) {
        OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
                "Error setting PortSlToPLFTMap for Switch GUID " U64H_FMT ", LID %u, "  \
                "port_block= %u status=%u\n",
                p_sw_entry->m_general_sw_info.m_guid,
                p_sw_entry->m_general_sw_info.m_lid,
                port_block, rec_status);

        HandleError(rec_status, AR_CLBCK_SET_PORT_SL_TO_PLFT_MAP,
                    m_algorithm_feature, p_sw_entry);

    } else {
        AR_MGR_LOG(TT_LOG_LEVEL_MAD,
                   "PortSlToPLFTMap for Switch GUID " U64H_FMT ", LID %u, "  \
                   "port_block= %u Was set\n",
                   p_sw_entry->m_general_sw_info.m_guid,
                   p_sw_entry->m_general_sw_info.m_lid,
                   port_block);
    }


    OSM_AR_LOG_RETURN_VOID(this->m_p_osm_log);
}

void SetVL2VLMapClbckDlg(const clbck_data_t &clbck_data,
                         int rec_status,
                         void *p_attribute_data)
{
    reinterpret_cast<AdaptiveRoutingClbck*>(clbck_data.m_p_obj)->
        SetVL2VLMapClbck(clbck_data, rec_status, p_attribute_data);
}

void AdaptiveRoutingClbck::SetVL2VLMapClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void */*p_attribute_data*/)
{
    OSM_AR_LOG_ENTER(this->m_p_osm_log);

    rec_status &= 0x00ff; //bits [8-15] are class specific - don't need them here

    ARSWDataBaseEntry *p_sw_entry = (ARSWDataBaseEntry *)clbck_data.m_data1;
    u_int8_t out_port_number = (u_int8_t)(intptr_t)clbck_data.m_data2;
    u_int8_t in_port_number = (u_int8_t)(intptr_t)clbck_data.m_data3;

    if (rec_status) {
        OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
                "Error setting VL2VL for Switch GUID " U64H_FMT ", LID %u, " \
                "in_port_number= %u out_port_number= %u status=%u\n",
                p_sw_entry->m_general_sw_info.m_guid,
                p_sw_entry->m_general_sw_info.m_lid,
                in_port_number, out_port_number, rec_status);

        HandleError(rec_status, AR_CLBCK_SET_VL_TO_VL_MAP,
                    m_algorithm_feature, p_sw_entry);
    }

    OSM_AR_LOG_RETURN_VOID(this->m_p_osm_log);
}

//////////////////////////////////////////////////////////////////////
///
/// clbck_data.m_data1: (ARCADataBaseEntry *)
/// clbck_data.m_data2: guid
/// clbck_data.m_data3: lid

void SetSL2VLMapOnHostsClbckDlg(const clbck_data_t &clbck_data,
                                int rec_status,
                                void *p_attribute_data)
{
    reinterpret_cast<AdaptiveRoutingClbck*>(clbck_data.m_p_obj)->
        SetSL2VLMapOnHostsClbck(clbck_data, rec_status, p_attribute_data);
}

void AdaptiveRoutingClbck::SetSL2VLMapOnHostsClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    OSM_AR_LOG_ENTER(m_p_osm_log);

    rec_status &= 0x00ff; //bits [8-15] are class specific - don't need them here

    ARCADataBaseEntry *p_ca_entry = (ARCADataBaseEntry *)clbck_data.m_data1;
    //osm_physp_t *p_physp = (osm_physp_t *)clbck_data.m_data2;
    uint64_t guid = (uint64_t)(intptr_t)clbck_data.m_data2;
    uint16_t lid = (uint16_t)(intptr_t)clbck_data.m_data3;

    if (rec_status) {
        OSM_AR_LOG(m_p_osm_log, OSM_LOG_VERBOSE,
                "Error setting SL2VL for CA GUID " U64H_FMT ", LID %u, " \
                "status=%u\n", guid, lid, rec_status);

        return;
    } else {

        struct SMP_SLToVLMappingTable * p_slvl_mapping =
            (struct SMP_SLToVLMappingTable *)p_attribute_data;

        if( memcmp(p_slvl_mapping, &p_ca_entry->m_slvl_mapping,
                   sizeof(SMP_SLToVLMappingTable))) {
            OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
                       "Unexpected result on setting SL2VL for CA GUID "
                       U64H_FMT ", LID %u req: %s != res %s\n",
                       guid, lid,
                       AdaptiveRoutingManager::ConvertSLToVLMappingToStr(
                           p_slvl_mapping).c_str(),
                       AdaptiveRoutingManager::ConvertSLToVLMappingToStr(
                           &p_ca_entry->m_slvl_mapping).c_str());
        } else {
            if (OSM_LOG_DEBUG & m_p_osm_log->level) {
                OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
                           "SL2VL for CA port GUID " U64H_FMT
                           ", LID %u, was successfuly set. %s\n", guid, lid,
                           AdaptiveRoutingManager::ConvertSLToVLMappingToStr(
                               &p_ca_entry->m_slvl_mapping).c_str());
            }

            p_ca_entry->m_to_set_slvl_mapping = false;
        }
    }

    OSM_AR_LOG_RETURN_VOID(m_p_osm_log);
}

//////////////////////////////////////////////////////////////////////
///
/// clbck_data.m_data1: (ARCADataBaseEntry *)
/// clbck_data.m_data2: bool* get_ext_sw_info
/// clbck_data.m_data3: bool* set_ext_sw_info
void GetExtendedSwitchInfoClbckDlg(
                          const clbck_data_t &clbck_data,
                          int rec_status,
                          void *p_attribute_data)
{
    reinterpret_cast<AdaptiveRoutingClbck*>(clbck_data.m_p_obj)->
        GetExtendedSwitchInfoClbck(clbck_data, rec_status, p_attribute_data);
}


void AdaptiveRoutingClbck::GetExtendedSwitchInfoClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    OSM_AR_LOG_ENTER(this->m_p_osm_log);

    rec_status &= 0x00ff; //bits [8-15] are class specific - don't need them here

    ARSWDataBaseEntry *p_sw_entry = (ARSWDataBaseEntry *)clbck_data.m_data1;

    if (rec_status) {
        OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
                "Error getting ExtendedSwitchInfo from Switch GUID "\
                U64H_FMT ", LID %u, status=%u\n",
                p_sw_entry->m_general_sw_info.m_guid,
                p_sw_entry->m_general_sw_info.m_lid,
                rec_status);

        p_sw_entry->MarkSWNotSupport(m_algorithm_feature, FABRIC_ERR);
        m_errcnt[AR_CLBCK_GET_EXT_SW_INFO]++;
    } else {
        SMP_ExtSWInfo * p_ext_sw_info =
            (SMP_ExtSWInfo *)p_attribute_data;

        if (!(p_ext_sw_info->SL2VL_Cap & 0x1)) {
            OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
                    "Switch GUID " U64H_FMT ", LID %u, "  \
                    "Do not support VL2VL\n",
                    p_sw_entry->m_general_sw_info.m_guid,
                    p_sw_entry->m_general_sw_info.m_lid);

            HandleError(rec_status, AR_CLBCK_GET_EXT_SW_INFO,
                        m_algorithm_feature, p_sw_entry);

            OSM_AR_LOG_RETURN_VOID(this->m_p_osm_log);
        }
        bool *get_ext_sw_info = (bool *)clbck_data.m_data2;
        bool *set_ext_sw_info = (bool *)clbck_data.m_data3;

        *get_ext_sw_info = false;

        if ( p_ext_sw_info->SL2VL_Act != 1 )  //VL2VLOnly
            *set_ext_sw_info = true;
        else
            *set_ext_sw_info = false;


    }
    OSM_AR_LOG_RETURN_VOID(this->m_p_osm_log);
}

//////////////////////////////////////////////////////////////////////
///
/// clbck_data.m_data1: (ARCADataBaseEntry *)
/// clbck_data.m_data2: bool* set_ext_sw_info
void SetExtendedSwitchInfoClbckDlg(
                          const clbck_data_t &clbck_data,
                          int rec_status,
                          void *p_attribute_data)
{
    reinterpret_cast<AdaptiveRoutingClbck*>(clbck_data.m_p_obj)->
        SetExtendedSwitchInfoClbck(clbck_data, rec_status, p_attribute_data);
}

void AdaptiveRoutingClbck::SetExtendedSwitchInfoClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    OSM_AR_LOG_ENTER(this->m_p_osm_log);

    rec_status &= 0x00ff; //bits [8-15] are class specific - don't need them here

    ARSWDataBaseEntry *p_sw_entry = (ARSWDataBaseEntry *)clbck_data.m_data1;
    if (rec_status) {
        OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
                "Error setting ExtendedSwitchInfo from Switch GUID " \
                U64H_FMT ", LID %u, status=%u\n",
                p_sw_entry->m_general_sw_info.m_guid,
                p_sw_entry->m_general_sw_info.m_lid,
                rec_status);

        HandleError(rec_status, AR_CLBCK_SET_EXT_SW_INFO,
                    m_algorithm_feature, p_sw_entry);

    } else {
        SMP_ExtSWInfo * p_ext_sw_info =
            (SMP_ExtSWInfo *)p_attribute_data;
        bool *set_ext_sw_info = (bool *)clbck_data.m_data2;

        //!!! TAMIR add to OSM
        if (p_ext_sw_info->SL2VL_Act == 1)
            *set_ext_sw_info = false;
        else
            *set_ext_sw_info = true;;
    }
    OSM_AR_LOG_RETURN_VOID(this->m_p_osm_log);
}

//!!! TAMIR add to OSM
void GetPrivateLFTInfoKdorClbckDlg(
                          const clbck_data_t &clbck_data,
                          int rec_status,
                          void *p_attribute_data)
{
    reinterpret_cast<AdaptiveRoutingClbck*>(clbck_data.m_p_obj)->
        GetPrivateLFTInfoKdorClbck(clbck_data, rec_status, p_attribute_data);
}

//!!! TAMIR add to OSM
//////////////////////////////////////////////////////////////////////
///
/// clbck_data.m_data1: (ARCADataBaseEntry *)
/// clbck_data.m_data2: (SMP_PrivateLFTInfo *)
///
void AdaptiveRoutingClbck::GetPrivateLFTInfoKdorClbck(
        const clbck_data_t &clbck_data,
        int rec_status,
        void *p_attribute_data)
{
    OSM_AR_LOG_ENTER(this->m_p_osm_log);

    rec_status &= 0x00ff; //bits [8-15] are class specific - don't need them here

    ARSWDataBaseEntry *p_sw_entry = (ARSWDataBaseEntry *)clbck_data.m_data1;

    if (rec_status) {
        OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
                "Error getting PrivateLFTInfo from Switch GUID " U64H_FMT ", LID %u, "  \
                "status=%u\n",
                p_sw_entry->m_general_sw_info.m_guid,
                p_sw_entry->m_general_sw_info.m_lid,
                rec_status);

        HandleError(rec_status, AR_CLBCK_GET_PLFT_INFO, SUPPORT_HC, p_sw_entry);

    } else {
        struct SMP_PrivateLFTInfo *p_plft_info =
            (struct SMP_PrivateLFTInfo *)p_attribute_data;

        u_int8_t ModeCap = p_plft_info->ModeCap;

        if (ModeCap == 0) {
            OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
                    "Switch GUID " U64H_FMT ", LID %u, "  \
                    "Do not support PrivateLFT\n",
                    p_sw_entry->m_general_sw_info.m_guid,
                    p_sw_entry->m_general_sw_info.m_lid);

            p_sw_entry->MarkSWNotSupport(m_algorithm_feature, DF_PLFT_CAP_ZERO); //TODO: remove DF prefix
            m_errcnt[AR_CLBCK_GET_PLFT_INFO]++;

            OSM_AR_LOG_RETURN_VOID(this->m_p_osm_log);
        }

        SMP_PrivateLFTInfo *p_sw_plft_info = (SMP_PrivateLFTInfo *)clbck_data.m_data2;
        *p_sw_plft_info = *p_plft_info;
    }
    OSM_AR_LOG_RETURN_VOID(this->m_p_osm_log);
}

