/*                  - Mellanox Confidential and Proprietary -
 *
 *  Copyright (C) 2010-2013, Mellanox Technologies Ltd.  ALL RIGHTS RESERVED.
 *
 *  Except as specifically permitted herein, no portion of the information,
 *  including but not limited to object code and source code, may be reproduced,
 *  modified, distributed, republished or otherwise exploited in any form or by
 *  any means for any purpose without the prior written permission of Mellanox
 *  Technologies Ltd. Use of software subject to the terms and conditions
 *  detailed in the file "COPYING".
 *
 */

/* osm_config.h needs to be before all other includes */
#include <opensm/osm_config.h>
#include <errno.h>
#include <ctype.h>
#include <stdlib.h>
#include <sys/time.h>

#include <opensm/osm_sm.h>
#include <opensm/osm_opensm.h>
#include <opensm/osm_base.h>
#include <complib/cl_qmap.h>

#include <algorithm>

#include <iostream>
#include <sstream>

#include "ar_mgr.h"

//VlCap 2 is : VL0, VL1
#define MIN_NUM_VLS 2


#define SW_TYPE_TO_STR(sw_type) \
        ((sw_type == SW_TYPE_UNKNOWN) ? "UNKNOWN" : \
            ((sw_type == SW_TYPE_LEAF) ? "LEAF" : "SPINE"))


/****************************************************/
/**************** private methods *******************/
/****************************************************/

int AdaptiveRoutingManager::ARDragonFlyCycle()
{
    OSM_AR_LOG_ENTER(this->m_p_osm_log);

    int rc = 0;
    AnalizeDFSetupData setup_data;

    for (int i=0 ; i<1 ; i++) {

        int unsupported_num = this->ARInfoGetProcess();
        if (unsupported_num) {
            OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_INFO,
                    "%d switches do not support AR.\n" , unsupported_num);
        }

        rc = InitDragonFlyPlus(setup_data);
        if (rc)
            break;

        if(!SetDragonFlyPlusCapable()) {
            rc = -1;
            break;
        }

        rc = AnalizeDragonFlyPlusSetup(setup_data);
        if (rc) {
            m_is_permanent_error = true;
            break;
        }

        unsupported_num = this->ARInfoGetGroupCapProcess();
        if (unsupported_num) {
            OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_INFO,
                       "%d switches do not support DFP Cap.\n" ,
                       unsupported_num);
        }

        rc = ARCalculatePortGroupsDF(setup_data);
        ARCalculatePortGroupsDFCleanup();
        if (rc) {
            m_is_permanent_error = true;
            break;
        }

        unsupported_num = this->ARInfoSetProcess();
        if (unsupported_num) {
            OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
                    "%d switches do not support AR Set.\n" , unsupported_num);
            m_is_temporary_error = true; //Need to recalculate AR groups and LFTS
        }

        ARMapSL2VLOnHosts();

        rc = ARDefinePLFTs();
        if (rc)
            break;

        rc = ARMapPLFTsAndVL2VLs();
        if (rc)
            break;

        rc = ARGroupTableProcessDF();
        if (rc)
            break;

        rc = ARLFTTableProcessDF();
        if (rc)
            break;

        UpdateSmDbSwInfo();
    }

    ARDragonFlyCycleEnd(rc);

    OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_VERBOSE,
               "ARDragonFlyCycle Ended. rc:%d\n", rc);
    OSM_AR_LOG_RETURN(this->m_p_osm_log, rc);
}

//Short UpdateVlids cycle
void AdaptiveRoutingManager::UpdateVlidsLftsDF()
{
    OSM_AR_LOG_ENTER(m_p_osm_log);
    OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG, "UpdateVlidsLftsDF.\n");

    LidsToLidsVecVec sw_lid_to_vlids;
    BuildSwToVlidsMap(sw_lid_to_vlids);

    if (sw_lid_to_vlids.empty()) {
        OSM_AR_LOG_RETURN_VOID(m_p_osm_log);
        return;
    }

    map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it =
        m_sw_db.m_sw_map.begin();
    for ( ; sw_node_it != this->m_sw_db.m_sw_map.end(); ++sw_node_it) {


        for (u_int8_t  plft_id=0;
             plft_id < sw_node_it->second.m_p_df_data->plft_number;
             plft_id++) {

            CalculateVlidsLft(
                sw_lid_to_vlids,
                sw_node_it->second.m_general_sw_info.m_p_osm_sw,
                sw_node_it->second.m_p_df_data->m_plft[plft_id].m_ar_lft,
                sw_node_it->second.m_p_df_data->m_plft[plft_id].m_to_set_lft_table);

            if (sw_node_it->second.m_p_df_data->m_plft[plft_id].m_max_lid <
                 m_p_osm_subn->max_lid) {
                sw_node_it->second.m_p_df_data->m_plft[plft_id].m_max_lid =
                 m_p_osm_subn->max_lid;
                sw_node_it->second.m_p_df_data->m_plft[plft_id].m_set_lft_top =
                    true;
            }
        }
    }

    OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_VERBOSE,
               "ARLFTTableProcessDF.\n");
    ARLFTTableProcessDF();

    OSM_AR_LOG_RETURN_VOID(m_p_osm_log);
}

void AdaptiveRoutingManager::ARDragonFlyCycleEnd(int rc)
{
    OSM_AR_LOG_ENTER(this->m_p_osm_log);
    OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_VERBOSE, "ARDragonFlyCycleEnd.\n");

    if (m_is_permanent_error)
        m_is_temporary_error = false;
    else if(rc && !m_is_temporary_error)
        m_is_permanent_error = true;

    map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it = this->m_sw_db.m_sw_map.begin();
    for ( ; sw_node_it != this->m_sw_db.m_sw_map.end();  ++sw_node_it) {

        if (sw_node_it->second.m_p_df_data == NULL)
            continue;

        //set priv to point to it's origin value
        sw_node_it->second.m_general_sw_info.m_p_osm_sw->priv =
            sw_node_it->second.m_p_df_data->m_saved_priv;

    }

    if (m_is_permanent_error) {
        ClearAllDragonflyConfiguration();
    }

    OSM_AR_LOG_RETURN_VOID(this->m_p_osm_log);
}

void AdaptiveRoutingManager::ClearAllDragonflyConfiguration()
{
    OSM_AR_LOG_ENTER(this->m_p_osm_log);

    if (m_df_configured == false)
        return;

    osm_switch_t *p_osm_switch;

    m_ar_clbck.ResetErrorCount();

    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &this->m_ar_clbck;
    clbck_data.m_handle_data_func = SetExtendedSwitchInfoClbckDlg;

    struct SMP_ExtSWInfo ext_sw_info;
    ext_sw_info.SL2VL_Act = 0; //SL2VL

    OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
               "Clear All Dragonfly Configuration\n");

    if (m_sw_info_configured) {

        OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_INFO,
                   "Remove SW info rank and coord configuration "
                   "from all switches and force_heavy_sweep\n");

        //We clear the rank and coord therefore we need to recalculate them
        //because we might overwrite valuble data set by the routing engine
        m_p_osm->subn.force_heavy_sweep = TRUE;
    }

    map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it = this->m_sw_db.m_sw_map.begin();
    for ( ; sw_node_it != this->m_sw_db.m_sw_map.end();  ++sw_node_it) {

        // clear AR configuration from sw - SM should configure this sw
        sw_node_it->second.m_app_data = ARMGR_SW_NOT_UPDATED;

        if (sw_node_it->second.m_p_df_data == NULL)
            continue;

        if (m_sw_info_configured) {
            //Clear rank and coordination from smdb switch info
            p_osm_switch = sw_node_it->second.m_general_sw_info.m_p_osm_sw;
            p_osm_switch->rank = OSM_SW_NO_RANK;
            p_osm_switch->coord = OSM_SW_NO_COORD;
        }

        //Remove VLToVL configuration
        if (sw_node_it->second.m_p_df_data != NULL &&
            sw_node_it->second.m_p_df_data->m_get_ext_sw_info == false &&
            sw_node_it->second.m_p_df_data->m_set_ext_sw_info == false) {

            OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
                    "Disable VL2VL on Switch GUID " U64H_FMT ", LID %u:\n",
                    sw_node_it->second.m_general_sw_info.m_guid,
                    sw_node_it->second.m_general_sw_info.m_lid);


            clbck_data.m_data1 = &sw_node_it->second;
            //!!! TAMIR add to OSM
            clbck_data.m_data2 = &sw_node_it->second.m_p_df_data->m_set_ext_sw_info;

            this->ExtendedSwitchInfoMadGetSetByDirect(
                &sw_node_it->second.m_general_sw_info.m_direct_route,
                IBIS_IB_MAD_METHOD_SET,
                &ext_sw_info, &clbck_data);

        }
    }
    m_sw_info_configured = false;

    m_ibis_obj.MadRecAll();

    if (m_ar_clbck.m_errcnt[AR_CLBCK_SET_EXT_SW_INFO]) {

        OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
                   "Clear Dragonfly configuration was not completed.\n");
        OSM_AR_LOG_RETURN_VOID(this->m_p_osm_log);
    }

    sw_node_it = this->m_sw_db.m_sw_map.begin();
    for ( ; sw_node_it != this->m_sw_db.m_sw_map.end();  ++sw_node_it) {

        if (sw_node_it->second.m_p_df_data != NULL) {
            delete sw_node_it->second.m_p_df_data;
            sw_node_it->second.m_p_df_data = NULL;
            //Setting/Unsetting PLFT invalidate groups data
            sw_node_it->second.ClearARGroupsData();
        }
    }
    m_df_configured = false;

    OSM_AR_LOG_RETURN_VOID(this->m_p_osm_log);

}

int AdaptiveRoutingManager::InitDragonFlyPlus(AnalizeDFSetupData &setup_data)
{
    OSM_AR_LOG_ENTER(this->m_p_osm_log);
    OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_VERBOSE, "InitDragonFlyPlus.\n");

    m_ar_clbck.ResetErrorCount();
    m_ar_clbck.m_algorithm_feature = SUPPORT_DF; //!!! TAMIR add to OSM
    m_df_configured = true;
    m_group_discovered = false;

    map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it = this->m_sw_db.m_sw_map.begin();
    for ( ; sw_node_it != this->m_sw_db.m_sw_map.end();  ++sw_node_it) {
        if (sw_node_it->second.m_p_df_data == NULL) {
            //Allocate Dragon Flay Data for all switches
            sw_node_it->second.m_p_df_data = new DfSwData();
            if (sw_node_it->second.m_p_df_data == NULL) {
                OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
                           "Failed to allocate Dragon Flay Data.\n");
                OSM_AR_LOG_RETURN(this->m_p_osm_log, -1);
            }
            sw_node_it->second.ClearDFData();
            //Setting/Unsetting PLFT invalidate groups data
            sw_node_it->second.ClearARGroupsData();
            //On moving to DF clear ARLFTs
            sw_node_it->second.ClearARLftData();
        }
        //when using DF, one can not disable AR on a specific switch
        sw_node_it->second.m_option_on = true;

        //Init DF Data
        sw_node_it->second.m_p_df_data->
            m_df_sw_setup[NEW_DATA].Clear();
        //consider port 0 as down port
        sw_node_it->second.m_p_df_data->
            m_df_sw_setup[NEW_DATA].m_down_ports.set(0);

        //set priv to point to current ARSWDataBaseEntry
        sw_node_it->second.m_p_df_data->m_saved_priv =
            sw_node_it->second.m_general_sw_info.m_p_osm_sw->priv;
        sw_node_it->second.m_general_sw_info.m_p_osm_sw->priv = &(sw_node_it->second);

        //Init group_number
        sw_node_it->second.m_p_df_data->m_df_prev_group_number =
            sw_node_it->second.m_p_df_data->m_df_group_number;
        sw_node_it->second.m_p_df_data->m_df_group_number =
            INVALID_DFP_GROUP_NUMBER;
        //update sw_lid_to_df_data_mapping
        setup_data.sw_lid_to_df_data_mapping[sw_node_it->second.m_general_sw_info.m_lid] =
            sw_node_it->second.m_p_df_data;

        if(sw_node_it->second.m_osm_update_needed) {
            sw_node_it->second.m_p_df_data->plft_number = 0;
            sw_node_it->second.m_p_df_data->m_plft_info.ModeCap = 0;
            sw_node_it->second.m_p_df_data->m_get_ext_sw_info = true;
            sw_node_it->second.m_p_df_data->m_set_ext_sw_info = true;
            sw_node_it->second.m_p_df_data->
                m_df_sw_setup[OLD_DATA].Clear();

            for (int i = 0 ; i < MAX_DF_PLFT_NUMBER ; ++i) { //!!! TAMIR add to OSM modified from MAX_PLFT_NUMBER
                sw_node_it->second.m_p_df_data->m_plft[i].m_max_lid = 0;
                //sw_node_it->second.m_p_df_data->m_plft[i].m_max_group_id = 0;
            }
        }

    }

    OSM_AR_LOG_RETURN(this->m_p_osm_log, 0);
}

bool AdaptiveRoutingManager::SetDragonFlyPlusCapable()
{
    OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_VERBOSE, "IsDragonFlyPlusCapable.\n");

    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &this->m_ar_clbck;

    struct SMP_PrivateLFTInfo plft_info;
    struct SMP_ExtSWInfo ext_sw_info;
    int unsupported_num = 0;
    int disable_by_tr_err = 0;
    bool new_dev = false;
    //!!! TAMIR add to OSM
    uint8_t num_vls = (1 << (m_p_osm_subn->opt.max_op_vls - 1));

    if (num_vls < MIN_NUM_VLS) {
        OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
                "Invalid maximal operational VLs %u configuration "\
                "for Dragonfly setup.\n",
                num_vls);

        return false;
    }

    map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it = this->m_sw_db.m_sw_map.begin();
    for ( ; sw_node_it != this->m_sw_db.m_sw_map.end(); ++sw_node_it) {

        if (sw_node_it->second.m_osm_update_needed) {
            sw_node_it->second.m_support[SUPPORT_DF] = SUPPORT_UNKNOWN;
        } else if(sw_node_it->second.m_support[SUPPORT_DF] == NOT_SUPPORTED) {
            unsupported_num++;
            continue;
        }

        if (!IsARActive(sw_node_it->second)) {
            OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
                    "Skip Switch GUID " U64H_FMT ", LID %u: " \
                    "AR not supported or not enabled.\n",
                    sw_node_it->second.m_general_sw_info.m_guid,
                    sw_node_it->second.m_general_sw_info.m_lid);
            unsupported_num++;
            continue;
        }

        clbck_data.m_data1 = &sw_node_it->second;
        //All switches should support PLFT at least 2 banks of at least 48K (PrivateLFTInfo)
        if (sw_node_it->second.m_p_df_data->m_plft_info.ModeCap == 0) {
            clbck_data.m_handle_data_func = GetPrivateLFTInfoClbckDlg;
            this->PLFTInfoMadGetSetByDirect(
                &sw_node_it->second.m_general_sw_info.m_direct_route,
                IBIS_IB_MAD_METHOD_GET,
                &plft_info, &clbck_data);
        }

        if (sw_node_it->second.m_support[SUPPORT_DF] == SUPPORTED) {
            continue;
        }

        if (sw_node_it->second.in_temporary_error)
            continue;

        new_dev = true;

        //All switches should support vl2vl configuration ( ExtendedSwitchInfo  SL2VLCap  bit 0 is on)
        if (&sw_node_it->second.m_p_df_data->m_get_ext_sw_info) {

            clbck_data.m_handle_data_func = GetExtendedSwitchInfoClbckDlg;
            //!!! TAMIR add to OSM
            clbck_data.m_data2 = &sw_node_it->second.m_p_df_data->m_get_ext_sw_info;
            clbck_data.m_data3 = &sw_node_it->second.m_p_df_data->m_set_ext_sw_info;

            this->ExtendedSwitchInfoMadGetSetByDirect(
                &sw_node_it->second.m_general_sw_info.m_direct_route,
                IBIS_IB_MAD_METHOD_GET,
                &ext_sw_info, &clbck_data);
        }

        if ((m_master_db.m_disable_tr_mask != 0x0) &&
            (!sw_node_it->second.m_ar_info.by_transp_cap)) {
            OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
                    "Switch GUID " U64H_FMT ", LID %u, " \
                    "do not support by_transport_disable.\n",
                    sw_node_it->second.m_general_sw_info.m_guid,
                    sw_node_it->second.m_general_sw_info.m_lid);
            disable_by_tr_err++;
        }

        //Bit 0 = 2 sub groups
        //Bit 1 = 3 sub groups
        //Bit 2 = 4 sub groups
        //Bits 3 to 7 are reserved
        if (!(sw_node_it->second.m_ar_info.sub_grps_supported & 0x1)) {
            OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
                       "Switch GUID " U64H_FMT ", LID %u: " \
                       "2 Subgroups not supported.\n",
                       sw_node_it->second.m_general_sw_info.m_guid,
                       sw_node_it->second.m_general_sw_info.m_lid);
            sw_node_it->second.MarkSWNotSupport(SUPPORT_DF, DF_GLOBAL_AR_GROUP);
        }

        //All switches should support adaptive Routing checked by ARInfoProcess
        if (sw_node_it->second.m_ar_info.by_sl_cap != 1) {
            OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
                       "Switch GUID " U64H_FMT ", LID %u: " \
                       "BySLCap not supported.\n",
                       sw_node_it->second.m_general_sw_info.m_guid,
                       sw_node_it->second.m_general_sw_info.m_lid);
            sw_node_it->second.MarkSWNotSupport(SUPPORT_DF, DF_NO_AR_BY_SL_CAP);
        }

        if (sw_node_it->second.m_ar_info.glb_groups != 1) {
            OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
                       "Switch GUID " U64H_FMT ", LID %u: " \
                       "Only global AR groups supported.\n",
                       sw_node_it->second.m_general_sw_info.m_guid,
                       sw_node_it->second.m_general_sw_info.m_lid);
            sw_node_it->second.MarkSWNotSupport(SUPPORT_DF, DF_GLOBAL_AR_GROUP);
        }

        //All switches should support at least 2 VLs in PortInfo VLCap
        osm_node_t *p_osm_node = sw_node_it->second.m_general_sw_info.m_p_osm_sw->p_node;
        osm_physp_t *p_physp_0 = osm_node_get_physp_ptr(p_osm_node, 0);
        if (!p_physp_0) {
            OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
                       "Switch GUID " U64H_FMT ", LID %u: " \
                       "Failed to get phys port 0.\n",
                       sw_node_it->second.m_general_sw_info.m_guid,
                       sw_node_it->second.m_general_sw_info.m_lid);
            sw_node_it->second.MarkSWNotSupport(SUPPORT_DF, DB_ERR);
        }

        if (p_physp_0 &&
            !(p_physp_0->port_info.capability_mask & IB_PORT_CAP_HAS_SL_MAP)) {
            OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
                       "Switch GUID " U64H_FMT ", LID %u: " \
                       "SL to VL map not supported.\n",
                       sw_node_it->second.m_general_sw_info.m_guid,
                       sw_node_it->second.m_general_sw_info.m_lid);
            sw_node_it->second.MarkSWNotSupport(SUPPORT_DF, DF_NO_SL_MAP_CAP);
        }

        for( u_int8_t port_num = 1 ; port_num <= p_osm_node->node_info.num_ports ; port_num++) {

            osm_physp_t *p_physp = osm_node_get_physp_ptr(p_osm_node, port_num);
            if (!p_physp || !osm_link_is_healthy(p_physp))
                continue;

            osm_physp_t *p_remote_physp = osm_physp_get_remote(p_physp);
            if (!p_remote_physp)
                continue;

            u_int8_t vl_cap = ib_port_info_get_vl_cap(&p_physp->port_info);
            num_vls = (1 << (vl_cap - 1));

            if (num_vls < MIN_NUM_VLS) {
                OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
                        "Invalid numberof VLs: %u on Switch GUID " U64H_FMT \
                        ", LID %u, port:%u\n",
                        num_vls,
                        sw_node_it->second.m_general_sw_info.m_guid,
                        sw_node_it->second.m_general_sw_info.m_lid,
                        port_num);

                sw_node_it->second.MarkSWNotSupport(SUPPORT_DF, DF_VL_CAP_LESS_THAN_MIN);
                break;

            }
        }

    }

    m_ibis_obj.MadRecAll();

    if (new_dev) {
        unsupported_num = 0;
        sw_node_it = this->m_sw_db.m_sw_map.begin();
        for ( ; sw_node_it != this->m_sw_db.m_sw_map.end(); ++sw_node_it) {

            if (sw_node_it->second.in_temporary_error)
                continue;

            if(sw_node_it->second.m_support[SUPPORT_DF] == NOT_SUPPORTED)
                unsupported_num++;
            else if (sw_node_it->second.m_support[SUPPORT_DF] ==
                     SUPPORT_UNKNOWN)
                sw_node_it->second.MarkSWSupport(SUPPORT_DF);
        }
    }

    if (unsupported_num || disable_by_tr_err) {
        if (unsupported_num) {
            OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_INFO,
                    "%d switches are not Dragonfly Capable.\n",
                    unsupported_num);
        }
        if (disable_by_tr_err) {
            OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_INFO,
                    "%d switches do not support by_transport_disable.\n",
                    disable_by_tr_err);
        }
    }


    return true;

}

/***************/


int AdaptiveRoutingManager::DiscoverLeaf(
    AnalizeDFSetupData &setup_data,
    ARSWDataBaseEntry* p_db_entry)
{
    OSM_AR_LOG_ENTER(this->m_p_osm_log);
    int rc = 0;
    /* For each connected port not in the DownPortsBitset
        and not in UpPortsBitset go to connected SW

        if remote is not SW - skip
        if remote is LEAF - report an error
        else set as SPINE

        If remote DfSwInfo. GroupNumber == 0 //new Spine in group
            Remote DfSwInfo. GroupNumber = local DfSwInfo. GroupNumber
            push remote DfSwInfo to GroupStack.
        Else If Remote DfSwInfo. GroupNumber != local DfSwInfo. GroupNumber
            Incorrect fabric architecture
    */

    osm_node_t *p_osm_node = p_db_entry->m_general_sw_info.m_p_osm_sw->p_node;
    DfSwSetup &curr_sw_setup = p_db_entry->m_p_df_data->
        m_df_sw_setup[NEW_DATA];

    for( u_int8_t port_num = 1 ; port_num <= p_osm_node->node_info.num_ports ; port_num++) {
        if (curr_sw_setup.m_up_ports.test(port_num) ||
            curr_sw_setup.m_down_ports.test(port_num))
            continue;

        osm_node_t *p_remote_node = osm_node_get_remote_node(p_osm_node, port_num, NULL);
        osm_physp_t *p_physp = osm_node_get_physp_ptr(p_osm_node, port_num);
        if (!p_physp || !osm_link_is_healthy(p_physp))
            continue;
        osm_physp_t *p_remote_physp = osm_physp_get_remote(p_physp);
        if (!p_remote_physp || !p_remote_node)
            continue;

        u_int8_t remote_node_type = osm_node_get_type(p_remote_node);

        if (remote_node_type != IB_NODE_TYPE_SWITCH) {
            OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
                   "Skip non sw node GUID " U64H_FMT "\n",
                   cl_ntoh64(osm_node_get_node_guid(p_remote_node)));

            continue;
        }

        if (p_remote_node->sw == NULL) {
            OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
                       "Invalid SW DB on node GUID " U64H_FMT "\n",
                   cl_ntoh64(osm_node_get_node_guid(p_remote_node)));
            rc = -1;
            break;
        }

        ARSWDataBaseEntry *p_remote_db_entry =
            (ARSWDataBaseEntry*)(p_remote_node->sw->priv);
        DfSwSetup &remote_sw_setup = p_remote_db_entry->m_p_df_data->
            m_df_sw_setup[NEW_DATA];

        if (remote_sw_setup.m_sw_type == SW_TYPE_LEAF) {
            OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
                       "Unexpected leaf GUID: " U64H_FMT " LID: %u"
                       " connected to leaf GUID: " U64H_FMT " LID: %u\n",
                       cl_ntoh64(osm_node_get_node_guid(p_osm_node)),
                       cl_ntoh16(osm_node_get_base_lid(p_osm_node, 0)),
                       cl_ntoh64(osm_node_get_node_guid(p_remote_node)),
                       cl_ntoh16(osm_node_get_base_lid(p_remote_node, 0)));
            rc = -1;
            break;
        }

        rc = SetSpine(setup_data, p_remote_node);
        if (rc) {
            break;
        }

        if (p_remote_db_entry->m_p_df_data->m_df_group_number ==
            INVALID_DFP_GROUP_NUMBER) {

            SetGroupNumber(p_remote_db_entry,
                           p_db_entry->m_p_df_data->m_df_group_number);

            OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
                       "Set SPINE GUID: " U64H_FMT " LID: %u group: %u\n",
                       p_remote_db_entry->m_general_sw_info.m_guid,
                       p_remote_db_entry->m_general_sw_info.m_lid,
                       p_remote_db_entry->m_p_df_data->m_df_group_number);

            setup_data.m_group_stack.push(p_remote_db_entry);

        } else if (p_remote_db_entry->m_p_df_data->m_df_group_number !=
                   p_db_entry->m_p_df_data->m_df_group_number) {

            OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
                       "Unexpected leaf GUID: " U64H_FMT " LID: %u from DF group %d"
                       " connected to spine GUID: " U64H_FMT " LID: %u from group %d\n",
                       cl_ntoh64(osm_node_get_node_guid(p_osm_node)),
                       cl_ntoh16(osm_node_get_base_lid(p_osm_node, 0)),
                       p_db_entry->m_p_df_data->m_df_group_number,
                       cl_ntoh64(osm_node_get_node_guid(p_remote_node)),
                       cl_ntoh16(osm_node_get_base_lid(p_remote_node, 0)),
                       p_remote_db_entry->m_p_df_data->m_df_group_number);

            rc = -1;
            break;
        }

    }

    OSM_AR_LOG_RETURN(this->m_p_osm_log, rc);
}

int AdaptiveRoutingManager::DiscoverSpine(
    AnalizeDFSetupData &setup_data,
    ARSWDataBaseEntry* p_db_entry)
{
    OSM_AR_LOG_ENTER(this->m_p_osm_log);
    int rc = 0;
    /* For each connected port not in the DownPortsBitset
        and not in UpPortsBitset go to connected SW

        Skip non sw node
        If remote type == Spine Set Spine
            If Remote DfSwInfo. GroupNumber == local DfSwInfo. GroupNumber
                Incorrect fabric architecture

        Else // line
            If remote DfSwInfo. GroupNumber == 0 //new line in group
                Remote DfSwInfo. GroupNumber = local DfSwInfo. GroupNumber
                push remote DfSwInfo to GroupStack.
            Else If Remote DfSwInfo. GroupNumber != local DfSwInfo. GroupNumber
                Incorrect fabric architecture
    */

    osm_node_t *p_osm_node = p_db_entry->m_general_sw_info.m_p_osm_sw->p_node;
    DfSwSetup &curr_sw_setup = p_db_entry->m_p_df_data->
        m_df_sw_setup[NEW_DATA];


    for( u_int8_t port_num = 1 ; port_num <= p_osm_node->node_info.num_ports ; port_num++) {
        if (curr_sw_setup.m_up_ports.test(port_num) ||
            curr_sw_setup.m_down_ports.test(port_num))
            continue;

        osm_node_t *p_remote_node = osm_node_get_remote_node(p_osm_node, port_num, NULL);
        osm_physp_t *p_physp = osm_node_get_physp_ptr(p_osm_node, port_num);
        if (!p_physp || !osm_link_is_healthy(p_physp))
            continue;
        osm_physp_t *p_remote_physp = osm_physp_get_remote(p_physp);
        if (!p_remote_physp || !p_remote_node)
            continue;

        u_int8_t remote_node_type = osm_node_get_type(p_remote_node);

        if (remote_node_type != IB_NODE_TYPE_SWITCH) {
            OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
                   "Skip non sw node GUID " U64H_FMT "\n",
                   cl_ntoh64(osm_node_get_node_guid(p_remote_node)));

            continue;
        }

        if (p_remote_node->sw == NULL) {
            OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
                       "Invalid SW DB on node GUID " U64H_FMT "\n",
                   cl_ntoh64(osm_node_get_node_guid(p_remote_node)));
            rc = -1;
            break;
        }

        ARSWDataBaseEntry *p_remote_db_entry = (ARSWDataBaseEntry*)(p_remote_node->sw->priv);
        DfSwSetup &remote_sw_setup = p_remote_db_entry->m_p_df_data->
            m_df_sw_setup[NEW_DATA];

        if (remote_sw_setup.m_sw_type == SW_TYPE_SPINE) {

            rc = SetSpine(setup_data, p_remote_node);
            if (rc)
                break;

            if (p_remote_db_entry->m_p_df_data->m_df_group_number ==
                p_db_entry->m_p_df_data->m_df_group_number) {

                OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
                           "Unexpected spine from DF group %d"
                           " connected to spine sw from the same group.\n",
                           p_db_entry->m_p_df_data->m_df_group_number);
                rc = -1;
                break;
            }

        } else if (remote_sw_setup.m_sw_type == SW_TYPE_LEAF) {

            if (p_remote_db_entry->m_p_df_data->m_df_group_number ==
                INVALID_DFP_GROUP_NUMBER) {


                SetGroupNumber(p_remote_db_entry,
                               p_db_entry->m_p_df_data->m_df_group_number);

                OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
                           "Set LEAF GUID: " U64H_FMT " LID: %u group: %u\n",
                           p_remote_db_entry->m_general_sw_info.m_guid,
                           p_remote_db_entry->m_general_sw_info.m_lid,
                           p_remote_db_entry->m_p_df_data->m_df_group_number);

                setup_data.m_group_stack.push(p_remote_db_entry);

            } else if (p_remote_db_entry->m_p_df_data->m_df_group_number !=
                       p_db_entry->m_p_df_data->m_df_group_number) {

                OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
                           "Unexpected spine GUID " U64H_FMT \
                           " LID:%u from DF group %d" \
                           " connected to line GUID " U64H_FMT \
                           " LID:%u from group %d\n",
                           p_db_entry->m_general_sw_info.m_guid,
                           p_db_entry->m_general_sw_info.m_lid,
                           p_db_entry->m_p_df_data->m_df_group_number,
                           p_remote_db_entry->m_general_sw_info.m_guid,
                           p_remote_db_entry->m_general_sw_info.m_lid,
                           p_remote_db_entry->m_p_df_data->m_df_group_number);

                rc = -1;
                break;
            }
        }
    }

    OSM_AR_LOG_RETURN(this->m_p_osm_log, rc);
}

int AdaptiveRoutingManager::SetLeaf(
    AnalizeDFSetupData &setup_data,
    SwDbEntryPrtList &leafs_list,
    osm_node_t *p_node)
{
    OSM_AR_LOG_ENTER(this->m_p_osm_log);
    int rc = 0;
    ARSWDataBaseEntry *p_db_entry = (ARSWDataBaseEntry*)(p_node->sw->priv);
    DfSwSetup &curr_sw_setup = p_db_entry->m_p_df_data->
        m_df_sw_setup[NEW_DATA];

    if (curr_sw_setup.m_sw_type == SW_TYPE_UNKNOWN) {
        curr_sw_setup.m_sw_type = SW_TYPE_LEAF;
        setup_data.m_leafs.push_back(p_db_entry);
        leafs_list.push_back(p_db_entry);
    } else {
        //!!! TODO check if else is valid
        OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
                   "SetLeaf Node (GUID " U64H_FMT ") with type: %s.\n",
                   p_node->node_info.node_guid,
                   SW_TYPE_TO_STR(curr_sw_setup.m_sw_type));
    }

    OSM_AR_LOG_RETURN(this->m_p_osm_log, rc);
}

int AdaptiveRoutingManager::SetSpine(
    AnalizeDFSetupData &setup_data,
    osm_node_t *p_node)
{
    OSM_AR_LOG_ENTER(this->m_p_osm_log);
    int rc = 0;
    ARSWDataBaseEntry *p_db_entry = (ARSWDataBaseEntry*)(p_node->sw->priv);
    DfSwSetup &curr_sw_setup = p_db_entry->m_p_df_data->
        m_df_sw_setup[NEW_DATA];

    if (curr_sw_setup.m_sw_type == SW_TYPE_UNKNOWN) {
        curr_sw_setup.m_sw_type = SW_TYPE_SPINE;
        setup_data.m_spines.push_back(p_db_entry);

    } else if (curr_sw_setup.m_sw_type != SW_TYPE_SPINE){
        OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
                   "SetLeaf Node (GUID " U64H_FMT ") with type: %s.\n",
                   p_node->node_info.node_guid,
                   SW_TYPE_TO_STR(curr_sw_setup.m_sw_type));
    }

    OSM_AR_LOG_RETURN(this->m_p_osm_log, rc);
}

#if 0
int AdaptiveRoutingManager::HandleLeafsWithoutHosts(AnalizeDFSetupData &setup_data,
                                                    int &max_group_number)
{
    /*
    For each SW in DfSpineMap
        if local DfSwInfo. GroupNumber != 0 continue
        For each connected port
            MinGroup = min(MinGroup, Remote DfSwInfo. GroupNumber)
            MaxGroup = max(MaxGroup, Remote DfSwInfo. GroupNumber)
            If MinGroup == MaxGroup // this is not a spine but a line without hosts
                Insert DfSwInfo to ChangeSpineToLine list
            Else this is definatly a spine. add all undiscovered neighbors to new_switches.
    */
    OSM_AR_LOG_ENTER(this->m_p_osm_log);

    int rc = 0;
    int min_group = max_group_number;
    int max_group = 0;
    uint8_t remote_port_num;
    list <ARSWDataBaseEntry*> new_switches;


    list <ARSWDataBaseEntry*>::iterator spines_it = setup_data.m_spines.begin();
    list <ARSWDataBaseEntry*>::iterator spines_remove_it = setup_data.m_spines.begin();

    for ( ; spines_it != setup_data.m_spines.end();) {
        if((*spines_it)->m_p_df_data->m_df_group_number != 0) {
            ++spines_it;
            continue;
        }

        osm_node_t *p_osm_node = (*spines_it)->m_general_sw_info.m_p_osm_sw->p_node;
        DfSwSetup &curr_sw_setup = (*spines_it)->m_p_df_data->
            m_df_sw_setup[NEW_DATA];

        /*
        //curr_sw_setup.m_down_ports[0] is always set.
        if (curr_sw_setup.m_down_ports.any()) {
            OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
                       "Unexpected Node (GUID " U64H_FMT " LID %u)"\
                       " with no group and with down ports.\n",
                       (*spines_it)->m_general_sw_info.m_guid,
                       (*spines_it)->m_general_sw_info.m_lid);
            rc = -1;
            break;
        }
        */

        min_group = max_group_number;
        max_group = 0;

        for( u_int8_t port_num = 1 ; port_num <= p_osm_node->node_info.num_ports ; port_num++) {

            osm_node_t *p_remote_node = osm_node_get_remote_node(
                p_osm_node, port_num, &remote_port_num);

            if (!curr_sw_setup.m_up_ports.test(port_num)) {
                if (p_remote_node != NULL) {
                    OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
                               "Usupportes spine GUID " U64H_FMT " LID %u." \
                               "May be connected to Leaf without hosts. (port:%u)\n",
                               (*spines_it)->m_general_sw_info.m_guid,
                               (*spines_it)->m_general_sw_info.m_lid, port_num);
                    OSM_AR_LOG_RETURN(this->m_p_osm_log, -1);
                }
                continue;
            }

            if (!p_remote_node) {
                OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
                           "Unexpected bad link at Node GUID " \
                           U64H_FMT " Lid %u ports:%u.\n",
                           (*spines_it)->m_general_sw_info.m_guid,
                           (*spines_it)->m_general_sw_info.m_lid,
                           port_num);
                return -1;
            }
            ARSWDataBaseEntry *p_remote_db_entry = (ARSWDataBaseEntry*)(p_remote_node->sw->priv);

            min_group = min(min_group, p_remote_db_entry->m_p_df_data->m_df_group_number);
            max_group = max(max_group, p_remote_db_entry->m_p_df_data->m_df_group_number);
        }

        if (min_group != max_group) {
            max_group_number++;
            (*spines_it)->m_p_df_data->m_df_group_number = max_group_number;
            OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
                       "Discover isolated Leaf. GUID "  \
                       U64H_FMT " LID %u group:%d\n",
                       (*spines_it)->m_general_sw_info.m_guid,
                       (*spines_it)->m_general_sw_info.m_lid,
                       max_group_number);
            continue;
        }

        //remove spines_it srom spines and push to leafs.
        //update remote m_up_ports to m_down_ports
        spines_remove_it = spines_it;
        ++spines_it;

        OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
                   "Handle leaf without hosts: Node GUID " \
                   U64H_FMT " LID %u\n",
                   (*spines_remove_it)->m_general_sw_info.m_guid,
                   (*spines_remove_it)->m_general_sw_info.m_lid);

        curr_sw_setup.m_sw_type = SW_TYPE_LEAF;
        (*spines_remove_it)->m_p_df_data->m_df_group_number = min_group;

        setup_data.m_leafs.push_back(*spines_remove_it);
        setup_data.m_spines.erase(spines_remove_it);

        //flip all remote ports from up do down
        for( u_int8_t port_num = 1 ; port_num <= p_osm_node->node_info.num_ports ; port_num++ ) {
            if (!curr_sw_setup.m_up_ports.test(port_num))
                continue;

            osm_node_t *p_remote_node = osm_node_get_remote_node(
                p_osm_node, port_num, &remote_port_num);

            ARSWDataBaseEntry *p_remote_db_entry =
                (ARSWDataBaseEntry*)(p_remote_node->sw->priv);
            DfSwSetup &remote_sw_setup = p_remote_db_entry->m_p_df_data->
                m_df_sw_setup[NEW_DATA];

            OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
                       "convert port %d from up to down on Node GUID " \
                       U64H_FMT " LID %u\n",
                       remote_port_num,
                       p_remote_db_entry->m_general_sw_info.m_guid,
                       p_remote_db_entry->m_general_sw_info.m_lid);

            remote_sw_setup.m_up_ports.reset(remote_port_num);
            remote_sw_setup.m_down_ports.set(remote_port_num);
        }

    }

    OSM_AR_LOG_RETURN(this->m_p_osm_log, rc);
}
#endif

int AdaptiveRoutingManager::MarkLeafsByCasNumber(AnalizeDFSetupData &setup_data,
                                                 SwDbEntryPrtList &leafs_list)
{
    int rc = 0;
    leafs_list.clear();

    OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
               "Mark leafs by CAs Number. "
               "Each switch with more than %u CAs is considered as leaf.\n",
               (u_int16_t)this->m_master_db.m_max_cas_on_spine);

    osm_node_t *p_osm_node;
    // Set Leaf SW
    for (p_osm_node = (osm_node_t *) cl_qmap_head(&m_p_osm_subn->node_guid_tbl);
         p_osm_node != (osm_node_t *) cl_qmap_end(&m_p_osm_subn->node_guid_tbl);
         p_osm_node = (osm_node_t *) cl_qmap_next(&p_osm_node->map_item)) {

        if (p_osm_node->node_info.node_type != IB_NODE_TYPE_SWITCH) {
            continue;
        }

        //Count the number of child CAs
        u_int8_t ca_count = 0;
        for (u_int8_t port_num = 1;
             port_num <= p_osm_node->node_info.num_ports;
             port_num++) {

            osm_node_t *p_remote_node = osm_node_get_remote_node(
                p_osm_node, port_num, NULL);
            if (!p_remote_node) {
                continue;
            }

            u_int8_t remote_node_type = osm_node_get_type(p_remote_node);

            if (remote_node_type != IB_NODE_TYPE_CA) {
                continue;
            }

            ca_count++;

            osm_physp_t *p_physp = osm_node_get_physp_ptr(p_osm_node, port_num);
            if (!p_physp || !osm_link_is_healthy(p_physp)) {
                continue;
            }
            osm_physp_t *p_remote_physp = osm_physp_get_remote(p_physp);
            if (!p_remote_physp) {
                continue;
            }

            rc = SetHcaToSwLidMapping(p_remote_physp, p_osm_node,
                                      setup_data.hca_to_sw_lid_mapping);
            if (rc) {
                OSM_AR_LOG_RETURN(this->m_p_osm_log, rc);
            }
        }

        if (ca_count > this->m_master_db.m_max_cas_on_spine) {
            OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
                       "Discover LEAF (connected to %u cas)."
                       " GUID: " U64H_FMT " LID: %u\n", ca_count,
                       cl_ntoh64(osm_node_get_node_guid(p_osm_node)),
                       cl_ntoh16(osm_node_get_base_lid(p_osm_node, 0)));

            rc = SetLeaf(setup_data, leafs_list, p_osm_node);
        }
    }
    OSM_AR_LOG_RETURN(this->m_p_osm_log, rc);
}

int AdaptiveRoutingManager::MarkLeafsByGroupsNumber(AnalizeDFSetupData &setup_data,
                                                    SwDbEntryPrtList &leafs_list)
{
    int rc = 0;
    leafs_list.clear();

    if (!m_group_discovered) {
        OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
                   "MarkLeafsByGroupsNumber No group was found.\n");
        return -1;
    }

    map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it =
        this->m_sw_db.m_sw_map.begin();
    for ( ; sw_node_it != this->m_sw_db.m_sw_map.end();  ++sw_node_it) {
        DfSwData * p_df_data = sw_node_it->second.m_p_df_data;
        if (p_df_data->m_df_group_number != INVALID_DFP_GROUP_NUMBER ||
            p_df_data->m_df_sw_setup[NEW_DATA].m_sw_type != SW_TYPE_UNKNOWN)
            continue;

        uint16_t min_group = m_max_df_group_number;
        uint16_t max_group = 0;

        osm_node_t *p_osm_node =
            sw_node_it->second.m_general_sw_info.m_p_osm_sw->p_node;

        for (u_int8_t port_num = 1; port_num <= p_osm_node->node_info.num_ports;
             port_num++) {

            osm_physp_t *p_physp = osm_node_get_physp_ptr(p_osm_node, port_num);
            if (!p_physp || !osm_link_is_healthy(p_physp))
                continue;
            osm_physp_t *p_remote_physp = osm_physp_get_remote(p_physp);
            if (!p_remote_physp)
                continue;
            osm_node_t *p_remote_node = osm_physp_get_node_ptr(p_remote_physp);
            u_int8_t remote_node_type = osm_node_get_type(p_remote_node);

            if (remote_node_type != IB_NODE_TYPE_SWITCH) {
                continue;
            }

            ARSWDataBaseEntry *p_remote_db_entry = (ARSWDataBaseEntry*)(p_remote_node->sw->priv);

            //if connected to LEAf this is not a LEAF
            if (p_remote_db_entry->m_p_df_data->m_df_sw_setup[NEW_DATA].m_sw_type
                == SW_TYPE_LEAF) {
                OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
                           "Discover SPINE (connected to LEAF)."
                           " GUID: " U64H_FMT " LID: %u\n",
                           sw_node_it->second.m_general_sw_info.m_guid,
                           sw_node_it->second.m_general_sw_info.m_lid);

                min_group = 0;
                max_group = m_max_df_group_number;
                break;
            }

            min_group = min(min_group, p_remote_db_entry->m_p_df_data->m_df_group_number);
            max_group = max(max_group, p_remote_db_entry->m_p_df_data->m_df_group_number);
        } //ports loop

        if (min_group == max_group) {
            OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
                       "Discover LEAF (connected to single group: %u)."
                       " GUID: " U64H_FMT " LID: %u\n", min_group,
                       sw_node_it->second.m_general_sw_info.m_guid,
                       sw_node_it->second.m_general_sw_info.m_lid);
            rc = SetLeaf(setup_data, leafs_list, p_osm_node);

            SetGroupNumber(&sw_node_it->second, min_group);

            OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
                       "Set LEAF GUID: " U64H_FMT " LID: %u group: %u\n",
                       sw_node_it->second.m_general_sw_info.m_guid,
                       sw_node_it->second.m_general_sw_info.m_lid,
                       p_df_data->m_df_group_number);
        } else {
            rc = SetSpine(setup_data, p_osm_node);
        }


        if (rc) {
            OSM_AR_LOG_RETURN(this->m_p_osm_log, rc);
        }

    } //switches loop

    OSM_AR_LOG_RETURN(this->m_p_osm_log, 0);
}

void AdaptiveRoutingManager::SetGroupNumber(ARSWDataBaseEntry* p_sw_entry,
                                            u_int16_t group_number)
{
    DfSwData *p_df_data = p_sw_entry->m_p_df_data;

    if (p_df_data->m_df_prev_group_number != group_number) {
        pair <GuidToGroupMapIter, bool> insert_iter =
            m_guid_to_dfp_group.insert(
                make_pair(p_sw_entry->m_general_sw_info.m_guid, group_number));
        insert_iter.first->second = group_number;

        OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
                   "SetGroupNumber insert: " U64H_FMT " -> %u.\n",
                   p_sw_entry->m_general_sw_info.m_guid, group_number);


    }
    p_df_data->m_df_group_number = group_number;
}

//Set previous group number (from df_data or from guid_to_dfp_group map)
//On success return 0
//On failure to set previous group number return 1
int AdaptiveRoutingManager::SetPrevGroupNumber(ARSWDataBaseEntry* p_sw_entry,
                                               BoolVec &used_group_numbers)
{
    DfSwData * p_df_data = p_sw_entry->m_p_df_data;

    u_int16_t prev_group_number = p_df_data->m_df_prev_group_number;
    if (prev_group_number == INVALID_DFP_GROUP_NUMBER) {
        GuidToGroupMapIter iter =
            m_guid_to_dfp_group.find(p_sw_entry->m_general_sw_info.m_guid);
        if (iter != m_guid_to_dfp_group.end()) {
            prev_group_number = iter->second;
        }

        OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
                   "SetGroupNumber map_prev:%u\n", prev_group_number);

    } else {
        OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
                   "SetGroupNumber prev:%u\n", prev_group_number);
    }

    if ((prev_group_number != INVALID_DFP_GROUP_NUMBER) &&
        (used_group_numbers[prev_group_number] == false)) {

        p_df_data->m_df_group_number = prev_group_number;
        used_group_numbers[prev_group_number] = true;

        OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
                   "SetGroupNumber set_prev\n");
        m_group_discovered = true;
        return 0;
    }

    return 1;
}

//On the first cycle we assigne previous group numbers for previous discovered switches
//On the second cycle we generates new groups numbers for switches in new groups
//We might lose group numbers on cases in wich groups are divided and rejoined
int AdaptiveRoutingManager::DiscoverGroups(AnalizeDFSetupData &setup_data,
                                           SwDbEntryPrtList &leafs_list,
                                           BoolVec &used_group_numbers,
                                           int cycle)
{
    int rc = 0;

    if (cycle > 2) {
        OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
                   "-A- DiscoverGroups: Invalid cycle number: %u\n",
                   cycle);
    }

    bool generate_new_group_number = (cycle != 1) ||
                                     (m_max_df_group_number == 0);
    SwDbEntryPrtList delayed_leafs_list;

    for (SwDbEntryPrtList::iterator iter = leafs_list.begin();
         iter !=  leafs_list.end(); ++iter) {

        DfSwData * p_df_data = (*iter)->m_p_df_data;
        if (p_df_data->m_df_group_number != INVALID_DFP_GROUP_NUMBER) {
            continue;
        }
        if (p_df_data->m_df_sw_setup[NEW_DATA].m_sw_type != SW_TYPE_LEAF) {
            OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
                       "DiscoverGroups: SW GUID " U64H_FMT "LID %u"
                       "Unexpected type %s.\n",
                       (*iter)->m_general_sw_info.m_guid,
                       (*iter)->m_general_sw_info.m_lid,
                       SW_TYPE_TO_STR(p_df_data->m_df_sw_setup[NEW_DATA].m_sw_type));
            continue;
        }

        if (SetPrevGroupNumber(*iter, used_group_numbers)) {
            if (generate_new_group_number) {
                m_max_df_group_number++;
                SetGroupNumber(*iter, m_max_df_group_number);
                m_group_discovered = true;
            } else {
                delayed_leafs_list.push_back(*iter);
                continue;
            }
        }
        OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
                   "Set LEAF GUID: " U64H_FMT " LID: %u group: %u\n",
                   (*iter)->m_general_sw_info.m_guid,
                   (*iter)->m_general_sw_info.m_lid,
                   p_df_data->m_df_group_number);

        setup_data.m_group_stack.push(*iter);

        while (!setup_data.m_group_stack.empty()) {
            ARSWDataBaseEntry* p_local_sw = setup_data.m_group_stack.top();
            setup_data.m_group_stack.pop();
            switch (p_local_sw->m_p_df_data->m_df_sw_setup[NEW_DATA].m_sw_type) {
                case SW_TYPE_LEAF:
                {
                    rc = DiscoverLeaf(setup_data, p_local_sw);
                    break;
                }
                case SW_TYPE_SPINE:
                {
                    rc = DiscoverSpine(setup_data, p_local_sw);
                    break;
                }
                default:
                {
                    OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
                               "Unexpected SW type.\n");
                    OSM_AR_LOG_RETURN(this->m_p_osm_log, -1);
                }
            }

            if (rc)
                OSM_AR_LOG_RETURN(this->m_p_osm_log, rc);
        } //same group

    }// all leafs in leafs_list

    if (!delayed_leafs_list.empty()) {
        rc = DiscoverGroups(setup_data, delayed_leafs_list,
                            used_group_numbers, ++cycle);
    }

    OSM_AR_LOG_RETURN(this->m_p_osm_log, rc);
}

int AdaptiveRoutingManager::SetPortsDirection()
{

    int rc = 0;
    int unknown_type_count = 0;

    map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it;

    for ( sw_node_it = m_sw_db.m_sw_map.begin();
          sw_node_it != this->m_sw_db.m_sw_map.end();
          ++sw_node_it) {

        DfSwData *p_df_data = sw_node_it->second.m_p_df_data;
        DfSwSetup &curr_sw_setup = p_df_data->m_df_sw_setup[NEW_DATA];

        if (curr_sw_setup.m_sw_type == SW_TYPE_UNKNOWN) {
            OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_INFO,
                       "Discover Unknown SW type. GUID "
                       U64H_FMT " LID %u\n",
                       sw_node_it->second.m_general_sw_info.m_guid,
                       sw_node_it->second.m_general_sw_info.m_lid);
            unknown_type_count++;
            continue;
        }

        osm_node_t *p_osm_node =
            sw_node_it->second.m_general_sw_info.m_p_osm_sw->p_node;

        for (u_int8_t port_num = 1; port_num <= p_osm_node->node_info.num_ports;
             port_num++) {

            osm_physp_t *p_physp = osm_node_get_physp_ptr(p_osm_node, port_num);
            if (!p_physp || !osm_link_is_healthy(p_physp))
                continue;
            osm_physp_t *p_remote_physp = osm_physp_get_remote(p_physp);
            if (!p_remote_physp)
                continue;
            osm_node_t *p_remote_node = osm_physp_get_node_ptr(p_remote_physp);

            u_int8_t remote_node_type = osm_node_get_type(p_remote_node);


            if (remote_node_type == IB_NODE_TYPE_SWITCH) {
                ARSWDataBaseEntry *p_remote_db_entry =
                    (ARSWDataBaseEntry*)(p_remote_node->sw->priv);
                DfSwSetup &remote_sw_setup =
                    p_remote_db_entry->m_p_df_data->m_df_sw_setup[NEW_DATA];

                if (remote_sw_setup.m_sw_type == SW_TYPE_LEAF) {
                    if (curr_sw_setup.m_sw_type == SW_TYPE_LEAF) {
                        OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
                                   "Invalid LEAF to LEAF connection. GUID "
                                   U64H_FMT " LID %u to GUID "
                                   U64H_FMT " LID %u\n",
                                   sw_node_it->second.m_general_sw_info.m_guid,
                                   sw_node_it->second.m_general_sw_info.m_lid,
                                   p_remote_db_entry->m_general_sw_info.m_guid,
                                   p_remote_db_entry->m_general_sw_info.m_lid);
                        OSM_AR_LOG_RETURN(this->m_p_osm_log, -1);
                    } else if (curr_sw_setup.m_sw_type == SW_TYPE_SPINE) {
                        curr_sw_setup.m_down_ports.set(port_num);
                    }
                } else if (remote_sw_setup.m_sw_type == SW_TYPE_SPINE) {
                    curr_sw_setup.m_up_ports.set(port_num);
                }
            } else if (remote_node_type == IB_NODE_TYPE_CA) {
                curr_sw_setup.m_down_ports.set(port_num);
            }
        } //ports
    }//switches

    OSM_AR_LOG_RETURN(this->m_p_osm_log, rc);
}

void AdaptiveRoutingManager::UpdateSmDbSwInfo()
{
    OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_VERBOSE,
               "UpdateSmDbSwInfo.\n");
    m_sw_info_configured = true;

    map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it =
        m_sw_db.m_sw_map.begin();
    for ( ; sw_node_it != this->m_sw_db.m_sw_map.end();  ++sw_node_it) {

        DfSwData * p_df_data = sw_node_it->second.m_p_df_data;
        osm_switch_t *p_osm_switch =
            sw_node_it->second.m_general_sw_info.m_p_osm_sw;

        if (p_df_data->m_df_group_number == INVALID_DFP_GROUP_NUMBER) {
            p_osm_switch->coord = OSM_SW_NO_COORD;
            p_osm_switch->rank = OSM_SW_NO_RANK;
            continue;
        }

        p_osm_switch->coord = p_df_data->m_df_group_number;

        switch (p_df_data->m_df_sw_setup[NEW_DATA].m_sw_type) {
            case SW_TYPE_LEAF:
                p_osm_switch->rank = 1;
                break;
            case SW_TYPE_SPINE:
                p_osm_switch->rank = 0;
                break;
            default:
                p_osm_switch->rank = OSM_SW_NO_RANK;
                p_osm_switch->coord = OSM_SW_NO_COORD;
        }
    }
}

int AdaptiveRoutingManager::AnalizeDragonFlyPlusSetup(AnalizeDFSetupData &setup_data)
{
    OSM_AR_LOG_ENTER(this->m_p_osm_log);
    OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_VERBOSE, "AnalizeDragonFlyPlusSetup.\n");

    SwDbEntryPrtList leafs_list;

    int rc = MarkLeafsByCasNumber(setup_data, leafs_list);
    if (rc) {
        OSM_AR_LOG_RETURN(this->m_p_osm_log, rc);
    }

    BoolVec used_group_numbers(m_max_df_group_number, false);
    rc = DiscoverGroups(setup_data, leafs_list, used_group_numbers);
    if (rc) {
        OSM_AR_LOG_RETURN(this->m_p_osm_log, rc);
    }

    rc = MarkLeafsByGroupsNumber(setup_data, leafs_list);
    if (rc) {
        OSM_AR_LOG_RETURN(this->m_p_osm_log, rc);
    }

    rc = DiscoverGroups(setup_data, leafs_list, used_group_numbers);
    if (rc) {
        OSM_AR_LOG_RETURN(this->m_p_osm_log, rc);
    }

    rc = SetPortsDirection();
    if (rc) {
        OSM_AR_LOG_RETURN(this->m_p_osm_log, rc);
    }

    //check if port moved from up to down or sw_type was changed
    map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it;

    for ( sw_node_it = m_sw_db.m_sw_map.begin();
          sw_node_it != this->m_sw_db.m_sw_map.end();
          ++sw_node_it) {

        DfSwData * p_df_data = sw_node_it->second.m_p_df_data;

        if ((p_df_data->m_df_sw_setup[NEW_DATA].m_sw_type !=
             p_df_data->m_df_sw_setup[OLD_DATA].m_sw_type) &&
            (p_df_data->m_df_sw_setup[OLD_DATA].m_sw_type !=
             SW_TYPE_UNKNOWN)) {

            OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_INFO,
                    "Switch type was changed for GUID " U64H_FMT \
                    ", LID %u: from %s to %s\n",
                    sw_node_it->second.m_general_sw_info.m_guid,
                    sw_node_it->second.m_general_sw_info.m_lid,
                    SW_TYPE_TO_STR(p_df_data->
                                   m_df_sw_setup[OLD_DATA].m_sw_type),
                    SW_TYPE_TO_STR(p_df_data->
                                   m_df_sw_setup[NEW_DATA].m_sw_type));

            p_df_data->m_df_sw_setup[OLD_DATA].Clear();
            continue;
        }

        //clear ports that change orientation
        //from up to down or from down to up
        p_df_data->m_df_sw_setup[OLD_DATA].m_up_ports &=
            ~p_df_data->m_df_sw_setup[NEW_DATA].m_down_ports;
        p_df_data->m_df_sw_setup[OLD_DATA].m_down_ports &=
            ~p_df_data->m_df_sw_setup[NEW_DATA].m_up_ports;

    }

    this->ARDumpDFAnalizedSetup();

    OSM_AR_LOG_RETURN(this->m_p_osm_log, rc);
}

//enforce routing between HCA to each node and back
static bool leaf_enforce_complete_route[LEAF_PLFT_NUMBER] = {false,true};

static PathDescription leaf_plft_paths_description[LEAF_PLFT_NUMBER][DLID_LOCATION_NUMBER][PATHS_NUMBER] = {
    /* Leaf Up (from Spine) */
    {
        /*Leaf Same Group*/
        {{2, 2, PATH_VALIDATION_NON}, {0, 0, PATH_VALIDATION_NON} , {0, 0, PATH_VALIDATION_NON} },
        /*Leaf Remote Group*/
        {{3, 3, PATH_VALIDATION_NON}, {0, 0, PATH_VALIDATION_NON} , {0, 0, PATH_VALIDATION_NON} },
        /*Spine Same Group*/
        {{1, 1, PATH_VALIDATION_NON}, {0, 0, PATH_VALIDATION_NON} , {0, 0, PATH_VALIDATION_NON} },
        /*Spine Remote Group*/
        {{2, 2, PATH_VALIDATION_NON}, {0, 0, PATH_VALIDATION_NON} , {0, 0, PATH_VALIDATION_NON} }
    },
    /* Leaf Down (from HCA) */
    {
        /*Leaf Same Group*/
        {{2, 2, PATH_VALIDATION_NON}, {0, 0, PATH_VALIDATION_NON} , {0, 0, PATH_VALIDATION_NON} },
        /*Leaf Remote Group*/
        {{3, 3, PATH_VALIDATION_NON}, {4, 4, PATH_VALIDATION_REMOTE_DF} , {6, 5, PATH_VALIDATION_REMOTE} },
        /*Spine Same Group*/
        {{1, 1, PATH_VALIDATION_NON}, {3, 3, PATH_VALIDATION_REMOTE} , {0, 0, PATH_VALIDATION_NON} },
        /*Spine Remote Group*/
        {{2, 2, PATH_VALIDATION_NON}, {3, 3, PATH_VALIDATION_REMOTE_DF} , {5, 4, PATH_VALIDATION_REMOTE} }
    }
};

//enforce routing between HCA to each node and back
static bool spine_enforce_complete_route[SPINE_PLFT_NUMBER] = {false, true/*, false*/};

static PathDescription spine_plft_paths_description[SPINE_PLFT_NUMBER][DLID_LOCATION_NUMBER][PATHS_NUMBER] = {
    /* Spine Up (from Spine) */
    {
        /*Leaf Same Group*/
        {{1, 1, PATH_VALIDATION_NON}, {0, 0, PATH_VALIDATION_NON}, {0, 0, PATH_VALIDATION_NON} },
        /*Leaf Remote Group*/
        {{2, 2, PATH_VALIDATION_NON}, {4, 4, PATH_VALIDATION_HOPS_DOWN}, {0, 0, PATH_VALIDATION_NON} },
        /*Spine Same Group*/
        {{2, 2, PATH_VALIDATION_DOWN}, {0, 0, PATH_VALIDATION_NON}, {0, 0, PATH_VALIDATION_NON} },
        /*Spine Remote Group*/
        {{1, 1, PATH_VALIDATION_NON}, {3, 3, PATH_VALIDATION_HOPS_DOWN}, {0, 0, PATH_VALIDATION_NON} }
    },
    /* Spine Down (from Leaf) VL0*/
    {
        /*Leaf Same Group*/
        {{1, 1, PATH_VALIDATION_NON}, {3, 3, PATH_VALIDATION_HOPS_DOWN} , {0, 0, PATH_VALIDATION_NON} },
        /*Leaf Remote Group*/
        {{2, 2, PATH_VALIDATION_NON}, {3, 3, PATH_VALIDATION_REMOTE_DF} , {5, 4, PATH_VALIDATION_REMOTE_UP} },
        /*Spine Same Group*/
        {{2, 2, PATH_VALIDATION_DOWN}, {0, 0, PATH_VALIDATION_NON} , {0, 0, PATH_VALIDATION_NON} },
        /*Spine Remote Group*/
        {{1, 1, PATH_VALIDATION_NON}, {2, 2, PATH_VALIDATION_REMOTE_DF} , {4, 3, PATH_VALIDATION_REMOTE_UP} }
    }
#if 0
    /* Spine Down (from Leaf) VL1*/
    ,{
        /*Leaf Same Group*/
        {{0, 0, PATH_VALIDATION_NON}, {0, 0, PATH_VALIDATION_NON} , {0, 0, PATH_VALIDATION_NON} },
        /*Leaf Remote Group*/
        {{2, 2, PATH_VALIDATION_NON}, {0, 0, PATH_VALIDATION_NON} , {0, 0, PATH_VALIDATION_NON} },
        /*Spine Same Group*/
        {{0, 0, PATH_VALIDATION_NON}, {0, 0, PATH_VALIDATION_NON} , {0, 0, PATH_VALIDATION_NON} },
        /*Spine Remote Group*/
        {{1, 1, PATH_VALIDATION_NON}, {0, 0, PATH_VALIDATION_NON} , {0, 0, PATH_VALIDATION_NON} }
    }
#endif //we might need 3 plft for pelican
};

int AdaptiveRoutingManager::ARCalculatePortGroupsDF(AnalizeDFSetupData &setup_data)
{
    OSM_AR_LOG_ENTER(this->m_p_osm_log);
    OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_VERBOSE, "ARCalculatePortGroupsDF.\n");

    int rc = 0;

    for (int plft_id = 0 ; plft_id < SPINE_PLFT_NUMBER ; ++plft_id) {
        rc = ARCalculatePLFTPortGroups(
            spine_plft_paths_description[plft_id],
            spine_enforce_complete_route[plft_id],
            setup_data.m_spines,
            plft_id,
            setup_data);
        if (rc)
            OSM_AR_LOG_RETURN(this->m_p_osm_log, rc);
    }

    for (int plft_id = 0 ; plft_id < LEAF_PLFT_NUMBER ; ++plft_id) {
        rc = ARCalculatePLFTPortGroups(
            leaf_plft_paths_description[plft_id],
            leaf_enforce_complete_route[plft_id],
            setup_data.m_leafs,
            plft_id,
            setup_data);
        if (rc)
            OSM_AR_LOG_RETURN(this->m_p_osm_log, rc);
    }

    SavePortGroupsAndDump();

    OSM_AR_LOG_RETURN(this->m_p_osm_log, rc);
}

void AdaptiveRoutingManager::ARCalculatePortGroupsDFCleanup()
{
    OSM_AR_LOG_ENTER(this->m_p_osm_log);

    map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it = this->m_sw_db.m_sw_map.begin();

    for ( ; sw_node_it != this->m_sw_db.m_sw_map.end(); ++sw_node_it) {

        if (sw_node_it->second.in_temporary_error)
            continue;

        sw_node_it->second.m_p_df_data->m_algorithm_data.Clear();

        for (int plft_id = 0 ; plft_id < MAX_DF_PLFT_NUMBER ; ++plft_id) { //!!! TAMIR add to OSM modified from MAX_PLFT_NUMBER
            ClearNoValidDFRouteToLid(sw_node_it->second, plft_id);
        }
    }

    OSM_AR_LOG_RETURN_VOID(this->m_p_osm_log);
}
//[is Spine][is Remote]
static ar_dlid_location_t ar_dlid_location[2][2] = {
    {LEAF_SAME_GROUP,LEAF_REMOTE_GROUP},
    {SPINE_SAME_GROUP, SPINE_REMOTE_GROUP}
};


int AdaptiveRoutingManager::ARCalculatePLFTPortGroups(
    PathDescription paths_description[DLID_LOCATION_NUMBER][PATHS_NUMBER],
    bool enforce_complete_route,
    list <ARSWDataBaseEntry*> &switches,
    int plft_id,
    AnalizeDFSetupData &setup_data)
{
    OSM_AR_LOG_ENTER(this->m_p_osm_log);

    OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG, "ARCalculatePLFTPortGroups\n");


    // some variables for calculation
    SMP_AR_LFT calculated_ar_lft;
    //SMP_ARLinearForwardingTable calculated_ar_lft_table[AR_LFT_TABLE_NUM_BLOCKS];
    unsigned calculated_max_lid;
    //int rc = 0;

    //On new devices groups are set globally and not per pLFT
    //there for algorithm_data moved to m_p_df_data
    //DFAlgorithmData algorithm_data;
    LidToPSGroupDataMap     lid_to_ps_group_map;

    uint16_t max_lid = 0;
    uint8_t **hops = NULL;
    PortsBitset group_bitmask[PATHS_NUMBER];

    //will be used to select the static bind port
    uint16_t ports_load[MAX_SUPPORTED_PORTS+1];

    PSGroupData ps_group_data;
    LidPortMappingPtrSX     lft_entry_ptr_sx;


    //Rewrite paths_description in more convenient data structure
    PathDescriptionEntry path_desc[DLID_LOCATION_NUMBER][DF_MAX_HOPS+1];
    memset(path_desc, 0, sizeof(path_desc));
    uint8_t max_hops_to_location[DLID_LOCATION_NUMBER];
    memset(max_hops_to_location, 0, sizeof(max_hops_to_location));

    int desc_max_hops_number = 0;
    int desc_min_hops_number = 0;
    for (int location_iter = 0 ; location_iter < DLID_LOCATION_NUMBER ; ++location_iter) {
        for (int path_iter = 0 ; path_iter < PATHS_NUMBER ; ++path_iter) {
            desc_max_hops_number = paths_description[location_iter][path_iter].m_max_hops_number;
            desc_min_hops_number = paths_description[location_iter][path_iter].m_min_hops_number;


            if (desc_max_hops_number == 0)
                break;

            if (desc_max_hops_number > DF_MAX_HOPS) {
                OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
                           "Invalid paths_description[%d][%d]"
                           " min_hops:%d-max_hops:%d\n",
                           location_iter, path_iter,
                           desc_min_hops_number, desc_max_hops_number);
                OSM_AR_LOG_RETURN(this->m_p_osm_log, -1);
            }

            for (int i = desc_min_hops_number ;
                 i <= desc_max_hops_number    ; ++i) {
                path_desc[location_iter][i].m_p_desk =
                    &paths_description[location_iter][path_iter];
                path_desc[location_iter][i].m_path_index = path_iter;
            }

            max_hops_to_location[location_iter] =
                max(desc_max_hops_number, (int)max_hops_to_location[location_iter]);

        }

        OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
                   "max_hops_to_location[%d] = %d\n",location_iter,
                   max_hops_to_location[location_iter]);
    }


    /*
     * Go through all the switches and create their port groups.
     * In this algorithm ports groups are created out of minimal hops
     * links - each lid will recieve a group containing all the ports
     * with minimal hops to this lid.
     * So here's what we're doing here:
     *   - for each switch
     *       - for each port of the switch (port_num)
     *           - calculate group bitmask for this lid
     *           - asisign groups acording to device limitations
     */
    uint16_t group_cap = 0;

    list <ARSWDataBaseEntry*>::iterator sw_entry_it = switches.begin();
    for ( ; sw_entry_it != switches.end() ; ++sw_entry_it) {

        if ((*sw_entry_it)->in_temporary_error)
            continue;

        if(!IsDFActive(*(*sw_entry_it))) {
            OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG, "Switch GUID " U64H_FMT ", LID %u: " \
                    "DF/AR not supported or not enabled, CalculatePLFTPortGroups skipped.\n",
                    (*sw_entry_it)->m_general_sw_info.m_guid,
                    (*sw_entry_it)->m_general_sw_info.m_lid);
            continue;
        }

        // (&&pLFT number support&&) We are currently support max of 2 pLFT
        // at this point m_p_df_data->plft_number is not assigned
        //
        //On Spines with only 2 lfts we will configure no to use AR for VL1
        //if ((*sw_entry_it)->m_p_df_data->plft_number <= plft_id)
        //    continue;

        group_cap = (*sw_entry_it)->GetGroupCap();
        OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
                "group_cap for GUID " U64H_FMT \
                ", LID %u: is:%u\n",
                (*sw_entry_it)->m_general_sw_info.m_guid,
                (*sw_entry_it)->m_general_sw_info.m_lid,
                group_cap);
        /*
         * Step 1 -
         *  Set default values for calculations
         */
        this->ARLftTableSetDefault(calculated_ar_lft.m_ar_lft_table_sx);

        calculated_max_lid = 0;
        //algorithm_data.Clear();
        lid_to_ps_group_map.clear();
        DFAlgorithmData &algorithm_data =
            (*sw_entry_it)->m_p_df_data->m_algorithm_data;
        unsigned &calculated_groups_number =
            algorithm_data.m_calculated_groups_number;

        ClearNoValidDFRouteToLid(*(*sw_entry_it), plft_id);
        memset(ports_load, 0, sizeof(ports_load));

        OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
                "Calculate group bitmask for GUID " U64H_FMT \
                ", LID %u: pLFT_ID:%u\n",
                (*sw_entry_it)->m_general_sw_info.m_guid,
                (*sw_entry_it)->m_general_sw_info.m_lid,
                plft_id);

        /*
         * Step 2.1 -
         *  Iterate through all lids and calculate group bitmask for each lid.
         */
        max_lid = (*sw_entry_it)->m_general_sw_info.m_p_osm_sw->max_lid_ho;
        hops = (*sw_entry_it)->m_general_sw_info.m_p_osm_sw->hops;
        //u_int8_t last_port_num;

        uint16_t sw_lid_num;
        int max_hops = 0;
        for (uint16_t lid_num = 1; lid_num <= max_lid; ++lid_num) {

            if (!hops[lid_num]) {
                //this is not sw lid. get the sw lid connected to this lid
                sw_lid_num = setup_data.hca_to_sw_lid_mapping[lid_num];
                if (sw_lid_num == 0) {
                    continue;
                }
                OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
                           "hops[lid:%u] = hops[sw_lid:%u]+1 \n",lid_num , sw_lid_num);
            } else
                sw_lid_num = lid_num;

            ps_group_data.Clear();
            //ps_group_data.m_ps_group_bitmask.reset();
            //ps_group_data.m_group_size = 0;

            if ((*sw_entry_it)->m_general_sw_info.m_lid == sw_lid_num)
                continue; //Same Switch

            if (!hops[sw_lid_num] || hops[sw_lid_num][0] < 1 ||
                hops[sw_lid_num][0] == OSM_NO_PATH) {
                continue;
            }

            if (sw_lid_num != lid_num)
                continue; //it is enough to go over switches and skip HCAs

            if (setup_data.sw_lid_to_df_data_mapping[sw_lid_num] == NULL) {
                OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
                           "No DF Data for SW lid::%u\n", sw_lid_num);
                OSM_AR_LOG_RETURN(this->m_p_osm_log, -1);
            }

            // Calculate DLID Location
            int type_idx = (int)(setup_data.sw_lid_to_df_data_mapping[sw_lid_num]->
               m_df_sw_setup[NEW_DATA].m_sw_type == SW_TYPE_SPINE);
            int dist_idx = (int)(setup_data.sw_lid_to_df_data_mapping[sw_lid_num]->
               m_df_group_number != (*sw_entry_it)->m_p_df_data->m_df_group_number);

            ar_dlid_location_t dlid_location = ar_dlid_location[type_idx][dist_idx];
            max_hops = max_hops_to_location[dlid_location];
            if (dist_idx == 0)  //on same group use only primary route
                max_hops = hops[sw_lid_num][0];

            OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
                   "Calculate group bitmask for LID:%u->LID:%u"
                   " min_hops:%u max_hops:%d dlid_location:%d\n",
                   (*sw_entry_it)->m_general_sw_info.m_lid, sw_lid_num,
                   hops[sw_lid_num][0], max_hops, dlid_location);

            /*
             *  Calculate group bitmask.
             */
            u_int8_t last_port_num = 0;
            int curr_hops = 0;

            for (u_int8_t port_num = 1;
                  port_num <= (*sw_entry_it)->m_general_sw_info.m_num_ports;
                  ++port_num) {

                curr_hops = hops[sw_lid_num][port_num];

                if(max_hops < curr_hops)
                    continue;

                //the maximum value of curr_hops is min_hopes+2
                //so the real hops number (withowt return on the same link)
                //might be higher so it might be equal max_hops
                /*
                if ((hops[sw_lid_num][0]+2 == curr_hops) &&
                     (curr_hops < max_hops)) {
                    OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
                               "we will treat hops[%u][%u]:%u as:%u\n",
                               sw_lid_num, port_num,
                               curr_hops, max_hops);
                    curr_hops = max_hops;
                }
                */
                PathDescriptionEntry & path_desc_entry =
                    path_desc[dlid_location][curr_hops];

                if( path_desc_entry.m_p_desk == NULL)
                    continue;

                switch (path_desc_entry.m_p_desk->m_path_validation_t) {
                    case PATH_VALIDATION_NON:
                        break;
                    case PATH_VALIDATION_REMOTE_DF:
                    {
                        OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
                                   "PATH_VALIDATION_REMOTE_DF hops[%u][%u]:%u \n",
                                   sw_lid_num, port_num ,curr_hops);

                        if (!IsRemoteSupportsDF(*(*sw_entry_it), port_num)) {
                            continue;
                        }
                        break;
                    }
                    case PATH_VALIDATION_REMOTE_UP:
                    {
                        OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
                                   "PATH_VALIDATION_REMOTE_UP hops[%u][%u]:%u \n",
                                   sw_lid_num, port_num ,curr_hops);
                        if(! (*sw_entry_it)->m_p_df_data->
                                m_df_sw_setup[NEW_DATA].m_up_ports.test(port_num))
                            continue;

                        if(!IsRouteOnRemote(*(*sw_entry_it), plft_id, port_num, sw_lid_num))
                            continue;

                        break;
                    }
                    case PATH_VALIDATION_REMOTE:
                    {
                        if(!IsRouteOnRemote(*(*sw_entry_it), plft_id, port_num, sw_lid_num))
                            continue;

                        break;
                    }
                    case PATH_VALIDATION_DOWN:
                    {
                        OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
                                   "PATH_VALIDATION_DOWN hops[%u][%u]:%u \n",
                                   sw_lid_num, port_num ,curr_hops);
                        if(! (*sw_entry_it)->m_p_df_data->
                                m_df_sw_setup[NEW_DATA].m_down_ports.test(port_num)) {
                            continue;
                        }

                        if (!IsRemoteSupportsDF(*(*sw_entry_it), port_num)) {
                            continue;
                        }

                        break;
                    }
                    case PATH_VALIDATION_HOPS_DOWN:
                    {
                        OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
                                   "PATH_VALIDATION_HOPS_DOWN hops[%u][%u]:%u \n",
                                   sw_lid_num, port_num ,curr_hops);
                        if(! (*sw_entry_it)->m_p_df_data->
                                m_df_sw_setup[NEW_DATA].m_down_ports.test(port_num))
                            continue;

                        if (!IsRemoteSupportsDF(*(*sw_entry_it), port_num)) {
                            continue;
                        }

                        //int real_hops = hops[sw_lid_num][port_num];

                        if ((hops[sw_lid_num][0]+2 == curr_hops) &&
                            !IsTrueHopsOnRemote(
                                *(*sw_entry_it), port_num,
                                sw_lid_num,
                                path_desc_entry.m_p_desk->m_max_hops_number-1))
                            continue;

                        break;
                    }
                    default:
                    {
                        OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
                                   "Unexpected path_validation::%u\n",
                                   path_desc_entry.m_p_desk->m_path_validation_t);
                        OSM_AR_LOG_RETURN(this->m_p_osm_log, -1);
                    }
                }

                group_bitmask[path_desc_entry.m_path_index].set(port_num);
                OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
                           "Add port:%d to bitset:%d \n",
                           port_num, path_desc_entry.m_path_index);

                last_port_num = port_num;

                ps_group_data.m_group_size++;
            }

            if (ps_group_data.m_group_size != 0) {

                //set ps_group_data
                bool set_pri = true;
                for (int i = 0 ; i < PATHS_NUMBER ; ++i) {
                    if(group_bitmask[i].any()){
                        if (set_pri) {
                            ps_group_data.m_ps_group_bitmask.set_pri(group_bitmask[i]);
                            set_pri = false;
                        } else {
                            ps_group_data.m_ps_group_bitmask.add_sec(group_bitmask[i]);
                            OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
                                       "Add secondary bitset index:%d value:%s \n",
                                       i, group_bitmask[i].to_string().c_str());
                        }

                        group_bitmask[i].reset();
                    }
                }

                if (!ps_group_data.m_ps_group_bitmask.any()) {
                    OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
                               "Failed to set group_bitmask for GUID "
                               U64H_FMT ", LID %u: \n",
                               (*sw_entry_it)->m_general_sw_info.m_guid,
                               (*sw_entry_it)->m_general_sw_info.m_lid);

                    OSM_AR_LOG_RETURN(this->m_p_osm_log, -1);
                }

                PSGroupBitmaskToDataMapInsertRes res =
                    algorithm_data.m_ps_groups_map.insert(make_pair(
                        ps_group_data.m_ps_group_bitmask,ps_group_data));

                if (ps_group_data.m_group_size > 1 &&
                    res.second) { //if new group set group number)

                    if (calculated_groups_number >= group_cap ) {
                        OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_INFO,
                                   "Group number exceeded. "
                                   "Use static port:%u for for LID:%u->LID:%u\n",
                                   res.first->second.m_port_num,
                                   (*sw_entry_it)->m_general_sw_info.m_lid, sw_lid_num);

                        //use only static
                        res.first->second.m_group_size = 1;

                    } else {
                        res.first->second.m_group_number =
                            calculated_groups_number;
                        calculated_groups_number++;
                    }

                } else if (ps_group_data.m_group_size == 1)
                    res.first->second.m_port_num =  last_port_num;

                //GroupBitmaskToDataMapIter groups_map_iter = groups_map.find(group_bitmask);
                //res.first->second.m_lids_list.push_back(sw_lid_num);
                lid_to_ps_group_map.insert(
                        make_pair(sw_lid_num,&(res.first->second)));
                //OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG, "m_lid_to_ps_group_map lid:%u group:%p \n", lid_num , &(res.first->second));

                PrintPSGroupData("Add/Update Group Data: ", res.first->second);

            } else {
                if (enforce_complete_route &&
                    (type_idx == 0 ||// dest sw is leaf
                     (*sw_entry_it)->m_p_df_data->
                        m_df_sw_setup[NEW_DATA].m_sw_type == SW_TYPE_LEAF)) {

                    OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
                               "No Valid Route from leaf switch (or) to leaf switch "
                               "pLFT:%u GUID " U64H_FMT " LID:%u->LID:%u. "
                               "Probably some links are missing.\n",
                               plft_id,
                               (*sw_entry_it)->m_general_sw_info.m_guid,
                               (*sw_entry_it)->m_general_sw_info.m_lid,
                               sw_lid_num);

                    if ((*sw_entry_it)->m_p_df_data->
                        m_df_sw_setup[NEW_DATA].m_sw_type == SW_TYPE_LEAF) {
                        OSM_AR_LOG_RETURN(this->m_p_osm_log, -1);
                    } else {
                        SetNoValidDFRouteToLid(*(*sw_entry_it), plft_id, sw_lid_num);
                    }
                } else {

                    OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
                               "No Valid Route pLFT:%u LID:%u->LID:%u\n",
                               plft_id,
                               (*sw_entry_it)->m_general_sw_info.m_lid,
                               sw_lid_num);

                    SetNoValidDFRouteToLid(*(*sw_entry_it), plft_id, sw_lid_num);
                }
            }
        }

        /*
         * Step 3 -
         *  Now let's configure AR LFTs.
         */
        OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG, "Configure AR LFTs \n");
        for (unsigned dest_lid = 1; dest_lid <= m_p_osm_subn->max_lid; ++dest_lid) {

            // Find out which port was used to route to lid_num in the usual unicast routing
            u_int8_t ucast_lft_port =
                    OsmSwitchGetPortByLidUseNewLft((*sw_entry_it)->m_general_sw_info.m_p_osm_sw,
                            dest_lid);
            if (ucast_lft_port == OSM_NO_PATH)  //no route to this lid_num from this switch
                continue;

            u_int8_t lft_entry_num = dest_lid % AR_LFT_TABLE_BLOCK_SIZE_SX;
            u_int16_t lft_block_num = dest_lid / AR_LFT_TABLE_BLOCK_SIZE_SX;
            lft_entry_ptr_sx.SetPtr(&calculated_ar_lft.m_ar_lft_table_sx[lft_block_num].LidEntry[lft_entry_num]);
            //p_lft_entry_ptr = &lft_entry_ptr_sx;

            if (!hops[dest_lid]) {
                //this is not sw lid. get the sw lid connected to this lid
                sw_lid_num = setup_data.hca_to_sw_lid_mapping[dest_lid];
                if (sw_lid_num == 0) {
                    continue;
                }
                OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
                           "dest_lid:%u ,  sw_lid:%u \n",dest_lid , sw_lid_num);
            } else
                sw_lid_num = dest_lid;


            LidToPSGroupDataMapIter iter = lid_to_ps_group_map.end();

            //if not same switch
            if ((*sw_entry_it)->m_general_sw_info.m_lid != sw_lid_num)
                iter = lid_to_ps_group_map.find(sw_lid_num);

            if (iter == lid_to_ps_group_map.end()) {

                if ((*sw_entry_it)->m_general_sw_info.m_lid == sw_lid_num) {

                    OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
                               "LID %u dest_lid:%u - Set route on same switch static port:%u\n",
                               (*sw_entry_it)->m_general_sw_info.m_lid,
                               dest_lid,
                               ucast_lft_port);
                } else if ((setup_data.sw_lid_to_df_data_mapping[sw_lid_num]->
                            m_df_sw_setup[NEW_DATA].m_sw_type == SW_TYPE_SPINE) &&
                           ((*sw_entry_it)->m_p_df_data->
                            m_df_sw_setup[NEW_DATA].m_sw_type == SW_TYPE_SPINE) &&
                           (m_master_db.m_op_mode &
                            AR_OP_MODE_ENABLE_SPINE_TO_SPINE_NONE_DF_ROUTE)) {

                    OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
                               "LID %u dest_lid:%u - Set none DF route "
                               "(Spine to Spine) static port:%u plft_id:%u\n",
                               (*sw_entry_it)->m_general_sw_info.m_lid,
                               dest_lid,
                               ucast_lft_port,
                               plft_id);

                } else {
                    OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
                               "No route from LID %u to dest_lid:%u plft_id: %u\n",
                               (*sw_entry_it)->m_general_sw_info.m_lid,
                               dest_lid, plft_id);
                    continue; //no route for this lid
                }

                lft_entry_ptr_sx.SetData(AR_IB_LID_STATE_STATIC, 0, 0, ucast_lft_port);

            } else if (iter->second->m_group_size == 1) {

                u_int8_t static_port = iter->second->m_port_num;

                if (iter->second->m_port_num == 0) {
                    //group number exeaded maximum
                    if (iter->second->m_ps_group_bitmask.test_pri(ucast_lft_port))
                        static_port = ucast_lft_port;
                    else
                        static_port = GetNextStaticPort(
                            ports_load,
                            iter->second->m_ps_group_bitmask,
                            sw_lid_num != dest_lid, //is HCA
                            (*sw_entry_it)->m_general_sw_info.m_num_ports);
                }

                OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
                           "LID %u dest_lid:%u - Set lid sate Static static_port:%u (p_group:%p)\n",
                           (*sw_entry_it)->m_general_sw_info.m_lid,
                           dest_lid,
                           static_port,
                           iter->second);

                lft_entry_ptr_sx.SetData(AR_IB_LID_STATE_STATIC, 0, 0, static_port);
            } else {
                u_int8_t static_port =0;

                if (iter->second->m_ps_group_bitmask.test_pri(ucast_lft_port))
                    static_port = ucast_lft_port;
                else
                    static_port = GetNextStaticPort(
                        ports_load,
                        iter->second->m_ps_group_bitmask,
                        sw_lid_num != dest_lid, //is HCA
                        (*sw_entry_it)->m_general_sw_info.m_num_ports);

                //SMP_AR_LID_STATE lidState = this->m_master_db.m_ar_mode;
                //if (((*sw_entry_it)->m_p_df_data->m_df_sw_setup[NEW_DATA].
                //     m_sw_type == SW_TYPE_SPINE) &&
                //    plft_id == 0 )
                //    lidState = AR_IB_LID_STATE_BOUNDED;

                OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
                           "LID %u dest_lid:%u - Set lid state %u Group:%u static_port:%u (p_group:%p)\n",
                           (*sw_entry_it)->m_general_sw_info.m_lid, dest_lid,
                           this->m_master_db.m_ar_mode,
                           iter->second->m_group_number,
                           static_port,
                           iter->second);

                lft_entry_ptr_sx.SetData(this->m_master_db.m_ar_mode,
                                         0,
                                         iter->second->m_group_number,
                                         static_port);
            }

            calculated_max_lid = dest_lid;

            if (this->m_p_osm_log->level & OSM_LOG_DEBUG) {
                OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
                           "LID %u dest_lid:%u - calculated_ar_lft_table[%u].LidEntry[%u]" \
                           " GroupNumber:%u TableNumber:%u LidState:%u\n",
                           (*sw_entry_it)->m_general_sw_info.m_lid,
                           dest_lid,
                           lft_block_num, lft_entry_num,
                           lft_entry_ptr_sx.GetGroupNumber(),
                           lft_entry_ptr_sx.GetTableNumber(),
                           lft_entry_ptr_sx.GetLidState());
            }

        }      /* End of loop destination lids */

        /*
         * Step 4 -
         *  Now we will decide if the calculated settings are different from old,
         *  or if we need to ignore old settings and use the new ones
         */
        this->ARUpdateDFLFTTable(*(*sw_entry_it),
                calculated_ar_lft.m_ar_lft_table_sx,
                calculated_max_lid,
                plft_id);
    }

    OSM_AR_LOG_RETURN(this->m_p_osm_log, 0);
}

u_int8_t AdaptiveRoutingManager::GetNextStaticPort(
                            uint16_t *ports_load,
                            PSPortsBitset &ps_group_bitmask,
                            bool isHCA,
                            u_int8_t num_ports)
{
    OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
               "GetNextStaticPort\n");

    bool set_load = isHCA || m_p_osm_subn->opt.port_profile_switch_nodes;

    uint16_t minLoad = 0xFFFF;
    u_int8_t static_port = 0;

    for (u_int8_t port_num = 1 ; port_num <= num_ports ; port_num++) {
        if (!ps_group_bitmask.test_pri(port_num))
            continue;
        if (!set_load)
            return port_num;
        if (ports_load[port_num] < minLoad) {
            minLoad = ports_load[port_num] ;
            static_port = port_num;
        }
    }

    OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
               "GetNextStaticPort returned:%u\n", static_port);

    ports_load[static_port]++;
    return static_port;
}


void AdaptiveRoutingManager::PrintPSGroupData(const char * str,PSGroupData &group_data)
{

    if(this->m_p_osm_log->level & OSM_LOG_DEBUG) {

        stringstream sstr;
        /*
        list<uint16_t>::const_iterator iter =  group_data.m_lids_list.begin();
        for (; iter != group_data.m_lids_list.end() ; ++iter) {
            sstr << " " << *iter;
        }
        */

        if (group_data.m_group_size == 1)
            sstr << " port:" << (int)group_data.m_port_num;
        else
            sstr << " ps bitset:" << group_data.m_ps_group_bitmask;

        /*
        OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
                   "%s Group:%p size:%d lids_number:%u lids:%s g_umber:%u\n",
                   str,
                   &group_data,
                   group_data.m_group_size,
                   (uint16_t)group_data.m_lids_list.size(),
                   sstr.str().c_str(),
                   group_data.m_group_number);
                   */
        OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
                   "%s Group:%p size:%d %s g_umber:%u\n",
                   str,
                   &group_data,
                   group_data.m_group_size,
                   sstr.str().c_str(),
                   group_data.m_group_number);
    }
}

bool AdaptiveRoutingManager::IsRemoteSupportsDF(ARSWDataBaseEntry &sw_db_entry,
                                                u_int8_t port_num)
{
    //get remote sw
    osm_node_t *p_osm_node = sw_db_entry.m_general_sw_info.m_p_osm_sw->p_node;
    osm_node_t *p_remote_node = osm_node_get_remote_node(p_osm_node, port_num, NULL);
	if(p_remote_node == NULL ||
       p_remote_node->sw == NULL) {
        OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
                   "IsRemoteSupportsDFP: Invalid remote node to Switch GUID "
                   U64H_FMT ", LID %u though port: %u\n",
                   sw_db_entry.m_general_sw_info.m_guid,
                   sw_db_entry.m_general_sw_info.m_lid,
                   port_num);
        return false;
    }

    ARSWDataBaseEntry *p_remote_db_entry = (ARSWDataBaseEntry*)(p_remote_node->sw->priv);

    if(!IsDFActive(*p_remote_db_entry)){
        OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
                   "Next switche from Switch GUID "
                   U64H_FMT ", LID %u though port: %u do not support DF\n",
                   sw_db_entry.m_general_sw_info.m_guid,
                   sw_db_entry.m_general_sw_info.m_lid,
                   port_num);

        return false;
    }
    return true;
}

bool AdaptiveRoutingManager::IsRouteOnRemote(ARSWDataBaseEntry &sw_db_entry,
        int plft_id, u_int8_t port_num, uint16_t lid_num)
{
    //get remote sw
    osm_node_t *p_osm_node = sw_db_entry.m_general_sw_info.m_p_osm_sw->p_node;
    osm_node_t *p_remote_node = osm_node_get_remote_node(p_osm_node, port_num, NULL);
	if(p_remote_node == NULL ||
       p_remote_node->sw == NULL) {
        OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
                   "Invalid remote node in IsRouteOnRemote"\
                   " from LID %u: plft:%d port:%u to LID:%u\n",
                   sw_db_entry.m_general_sw_info.m_lid,
                   plft_id, port_num, lid_num);
        return false;
    }

    ARSWDataBaseEntry *p_remote_db_entry = (ARSWDataBaseEntry*)(p_remote_node->sw->priv);

    if (!IsDFActive(*p_remote_db_entry)) {
        OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
                   "Next switch from LID %u: plft:%d though port:%u to LID:%u "
                   "do not support DF\n",
                   sw_db_entry.m_general_sw_info.m_lid,
                   plft_id, port_num, lid_num);
        return false;
    }

    //get remote plft_id according to the HDL for DF algorithm
    int remote_plft_number = 0;
    int rc =0;
    if (sw_db_entry.m_p_df_data->m_df_sw_setup[NEW_DATA].m_sw_type == SW_TYPE_LEAF) {
        if (plft_id != 1)
            rc = -1;
        remote_plft_number = 1;
    } else {
        if (plft_id != 1)
            rc = -1;
        remote_plft_number = 0;
    }
    if (rc) {
        OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
                   "Invalid call to IsRouteOnRemote"\
                   " from LID %u: plft:%d port:%u to LID:%u\n",
                   sw_db_entry.m_general_sw_info.m_lid,
                   plft_id, port_num, lid_num);
        return false;
    }

    set <u_int16_t> & no_df_valid_route =
        p_remote_db_entry->m_p_df_data->m_plft[remote_plft_number].m_no_df_valid_route;

    if (no_df_valid_route.find(lid_num) != no_df_valid_route.end()) {
        OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
                   "No route from LID %u: plft:%d though port:%u to LID:%u\n",
                   sw_db_entry.m_general_sw_info.m_lid,
                   plft_id, port_num, lid_num);
        return false;
    }
    return true;
}

bool AdaptiveRoutingManager::IsTrueHopsOnRemote(ARSWDataBaseEntry &sw_db_entry,
        u_int8_t local_port_num, uint16_t lid_num, int hops_num)
{
    //get remote sw
    uint8_t remote_port_num;
    osm_node_t *p_osm_node = sw_db_entry.m_general_sw_info.m_p_osm_sw->p_node;
    osm_node_t *p_remote_node = osm_node_get_remote_node(
        p_osm_node, local_port_num, &remote_port_num);

	if(p_remote_node == NULL ||
       p_remote_node->sw == NULL ) {
        OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
                   "Invalid remote node in IsTrueHopsOnRemote from LID %u: to LID:%u\n",
                   sw_db_entry.m_general_sw_info.m_lid,
                   lid_num);
        return false;
    }


    if (!p_remote_node->sw->hops[lid_num] ||
        p_remote_node->sw->hops[lid_num][0] > hops_num)
        return false;

    for (u_int8_t port_num = 1;
         port_num <= p_remote_node->node_info.num_ports;
         ++port_num) {

        if (p_remote_node->sw->hops[lid_num][port_num] != hops_num)
            continue;

        if (port_num == remote_port_num)
            continue;

        return true;
    }

    return false;
}

void AdaptiveRoutingManager::ARUpdateDFLFTTable(ARSWDataBaseEntry &sw_db_entry,
        struct SMP_ARLinearForwardingTable_SX p_ar_calculated_lft_table[],
        u_int16_t calculated_max_lid,
        int plft_id)
{
    OSM_AR_LOG_ENTER(this->m_p_osm_log);

    PLFTMads   &plft_mads = sw_db_entry.m_p_df_data->m_plft[plft_id];

    // Update blocks
    // In fact we go over only on blocks we need  - until max lid
    unsigned supported_lft_blocks =
            calculated_max_lid / AR_LFT_TABLE_BLOCK_SIZE_SX;
    for (unsigned lft_block = 0; lft_block <= supported_lft_blocks; ++lft_block) {
        if (sw_db_entry.m_osm_update_needed ||
                !this->IsEqualSMPARLftTableBlock(
                        &p_ar_calculated_lft_table[lft_block],
                        &plft_mads.m_ar_lft.m_ar_lft_table_sx[lft_block])) {

                plft_mads.m_ar_lft.m_ar_lft_table_sx[lft_block] =
                    p_ar_calculated_lft_table[lft_block];
                plft_mads.m_to_set_lft_table[lft_block] = true;

                OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
                           "Need to set LFT block: %u plft_id: %u on "
                           "Switch GUID " U64H_FMT ", LID %u:\n",
                           lft_block,
                           plft_id,
                           sw_db_entry.m_general_sw_info.m_guid,
                           sw_db_entry.m_general_sw_info.m_lid);
        }
    }

    if (plft_mads.m_max_lid != calculated_max_lid) {
        plft_mads.m_set_lft_top = true;
        plft_mads.m_max_lid = calculated_max_lid;

        OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
                   "Need to set LFT top: %u on "
                   "Switch GUID " U64H_FMT ", LID %u:\n",
                   plft_mads.m_max_lid,
                   sw_db_entry.m_general_sw_info.m_guid,
                   sw_db_entry.m_general_sw_info.m_lid);

    }

    OSM_AR_LOG_RETURN_VOID(this->m_p_osm_log);
}

void AdaptiveRoutingManager::SavePortGroupsAndDump()
{
    OSM_AR_LOG_ENTER(this->m_p_osm_log);

    SMP_ARGroupTable calculated_ar_group_table[AR_GROUP_TABLE_NUM_BLOCKS];
    map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it = this->m_sw_db.m_sw_map.begin();

    for ( ; sw_node_it != this->m_sw_db.m_sw_map.end(); ++sw_node_it) {

        OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
                   "SavePortGroups for Switch GUID " U64H_FMT ", LID %u:\n",
                    sw_node_it->second.m_general_sw_info.m_guid,
                    sw_node_it->second.m_general_sw_info.m_lid);

        this->ARGroupTableSetDefault(calculated_ar_group_table);
        DFAlgorithmData &algorithm_data =
            sw_node_it->second.m_p_df_data->m_algorithm_data;

        /*
         *  Update calculated_ar_group_table:
         *  At this point we have all assigned groups in algorithm_data.m_assign_groups list
         */
        if (!algorithm_data.m_ps_groups_map.empty()) {

            PSGroupBitmaskToDataMapIter groups_iter = algorithm_data.m_ps_groups_map.begin();
            PortGroup_Block_Element *p_element = NULL;
            for ( ;
                  groups_iter != algorithm_data.m_ps_groups_map.end();
                  ++groups_iter) {

                if (groups_iter->second.m_group_size == 1)
                    continue; //this is Static port not a group

                p_element = &calculated_ar_group_table[groups_iter->second.m_group_number].Group[0];

                PortsBitset &primary_bitset = groups_iter->second.m_ps_group_bitmask.get_primary();
                PortsBitset &secondary_bitset = groups_iter->second.m_ps_group_bitmask.get_secondary();

                primary_bitset.get_data(p_element->SubGroup_0,
                                        p_element->SubGroup_1,
                                        p_element->SubGroup_2,
                                        p_element->SubGroup_3);

                OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
                           "Set Group:%u primary:" U64H_FMT "\n",
                           groups_iter->second.m_group_number,
                           p_element->SubGroup_0);

                if (secondary_bitset.any()) {
                    p_element = &calculated_ar_group_table[groups_iter->second.m_group_number].Group[1];

                    secondary_bitset.get_data(p_element->SubGroup_0,
                                              p_element->SubGroup_1,
                                              p_element->SubGroup_2,
                                              p_element->SubGroup_3);

                    OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
                               "Set Group:%u secondary:" U64H_FMT "\n",
                               groups_iter->second.m_group_number,
                               p_element->SubGroup_0);
                }
            }
        }

        /*
         *  Now we will decide if the calculated settings are different from old,
         *  or if we need to ignore old settings and use the new ones
         */
        if (!algorithm_data.m_ps_groups_map.empty()) {
            this->ARUpdateSWGroupTable(sw_node_it->second,
                    calculated_ar_group_table,
                    sw_node_it->second.m_p_df_data->m_algorithm_data.
                                       m_calculated_groups_number);
        }

        /*
         * Now dump debug information to osm log
         */
        this->ARDumpDFSettings(sw_node_it->second);

    }

    OSM_AR_LOG_RETURN_VOID(this->m_p_osm_log);
}

void AdaptiveRoutingManager::ARDumpDFAnalizedSetup()
{
    OSM_AR_LOG_ENTER(this->m_p_osm_log);

    if (!osm_log_is_active(this->m_p_osm_log,OSM_LOG_ROUTING))
        return;

    map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it =
        this->m_sw_db.m_sw_map.begin();
    for ( ; sw_node_it != this->m_sw_db.m_sw_map.end();  ++sw_node_it) {

        DfSwData * p_df_data = sw_node_it->second.m_p_df_data;

        char buff[1024];
        string str = "---------------\n";
        sprintf(buff,
                "\t\t\t\tDragonFly Analized Setup::\n" \
                "\t\t\t\tSwitch GUID............" U64H_FMT "\n" \
                "\t\t\t\tLID....................%u\n"   \
                "\t\t\t\tType...................%s\n"   \
                "\t\t\t\tGroup..................%d\n"   \
                "\t\t\t\tUp Ports...............%s\n"   \
                "\t\t\t\tDown Ports.............%s\n",
                sw_node_it->second.m_general_sw_info.m_guid,
                sw_node_it->second.m_general_sw_info.m_lid,
                SW_TYPE_TO_STR(p_df_data->
                               m_df_sw_setup[NEW_DATA].m_sw_type),
                p_df_data->m_df_group_number,
                p_df_data->m_df_sw_setup[NEW_DATA].
                                m_up_ports.to_string().c_str(),
                p_df_data->m_df_sw_setup[NEW_DATA].
                                m_down_ports.to_string().c_str());
        str += buff;
        OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ROUTING, "%s", str.c_str());
    }
    OSM_AR_LOG_RETURN_VOID(this->m_p_osm_log);
}

void AdaptiveRoutingManager::ARDumpDFSettings(ARSWDataBaseEntry &sw_db_entry)
{
    OSM_AR_LOG_ENTER(this->m_p_osm_log);

    if (!osm_log_is_active(this->m_p_osm_log,OSM_LOG_ROUTING))
        return;

    char buff[1024];
    string str = "---------------\n";
    ConvertARInfoToStr(sw_db_entry.m_general_sw_info,
                       sw_db_entry.m_ar_info, buff);
    str += buff;
    OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ROUTING, "%s", str.c_str());

    //u_int16_t max_group_id =
    //    (sw_db_entry.m_total_groups_number > 0 ?
    //     sw_db_entry.m_total_groups_number -1 : 0);

    for (unsigned group_block = 0; group_block <= sw_db_entry.m_group_top ;
         ++group_block) {

        sprintf(buff, "Group Table Settings : Block " U32D_FMT "/" U32D_FMT "\n",
                group_block,
                sw_db_entry.m_group_top);
        str = buff;
        str += this->ConvertARGroupTableBlockToStr(
            &sw_db_entry.m_ar_group_table[group_block],
            sw_db_entry.m_ar_info, group_block);
        OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ROUTING, "%s", str.c_str());
    }

    for (int plft_id = 0 ; plft_id < MAX_DF_PLFT_NUMBER ; ++plft_id) { //!!! TAMIR add to OSM modified from MAX_PLFT_NUMBER

        PLFTMads   &plft_mads = sw_db_entry.m_p_df_data->m_plft[plft_id];

        unsigned supported_lft_blocks =
                plft_mads.m_max_lid / AR_LFT_TABLE_BLOCK_SIZE_SX;
        for (unsigned lft_block = 0; lft_block <= supported_lft_blocks; ++lft_block) {
            sprintf(buff, "SX LFT Table Settings : pLFT:%u Block " U32D_FMT \
                    "/" U32D_FMT "\n", plft_id, lft_block,
                    AR_LFT_TABLE_NUM_BLOCKS_SX);
            str = this->ConvertARLFTTableBlockToStr(
                  &plft_mads.m_ar_lft.m_ar_lft_table_sx[lft_block],
                  lft_block);

            if (!str.empty())
                OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ROUTING,
                           "%s%s", buff, str.c_str());
        }
    }

    str = "---------------\n";
    OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ROUTING, "%s", str.c_str());

    OSM_AR_LOG_RETURN_VOID(this->m_p_osm_log);
}

int AdaptiveRoutingManager::ARGroupTableProcessDF()
{
    OSM_AR_LOG_ENTER(this->m_p_osm_log);
    OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_VERBOSE, "ARGroupTableProcessDF.\n");
    int rc = 0;

    map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it = this->m_sw_db.m_sw_map.begin();
    for ( ; sw_node_it != this->m_sw_db.m_sw_map.end(); ++sw_node_it) {

        if (sw_node_it->second.in_temporary_error)
            continue;

        if(!IsDFActive(sw_node_it->second)) {
            OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG, "Switch GUID " U64H_FMT ", LID %u: " \
                    "DF/AR not supported or not enabled, ARGroupTableProcessDF skipped.\n",
                    sw_node_it->second.m_general_sw_info.m_guid,
                    sw_node_it->second.m_general_sw_info.m_lid);
            continue;
        }

        ARGroupTableProcess(
            sw_node_it->second,
            //sw_node_it->second.m_total_groups_number,
            sw_node_it->second.m_group_top,
            false, //reset_direction
            sw_node_it->second.m_ar_group_table);
    }

    //Receive all Set ARGroupTable reply
    m_ibis_obj.MadRecAll();

    if (m_ar_clbck.m_errcnt[AR_CLBCK_SET_AR_GROUP_TABLE]) {
        OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
                "Set AR Groups error count: %u.\n",
                m_ar_clbck.m_errcnt[AR_CLBCK_SET_AR_GROUP_TABLE]);
        m_is_temporary_error = true; //Need to recalculate AR groups and LFTS
    }


    OSM_AR_LOG_RETURN(this->m_p_osm_log, rc);
}

int AdaptiveRoutingManager::ARLFTTableProcessDF()
{
    OSM_AR_LOG_ENTER(this->m_p_osm_log);
    OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_VERBOSE, "ARLFTTableProcessDF.\n");
    int rc = 0;

    map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it = this->m_sw_db.m_sw_map.begin();
    for ( ; sw_node_it != this->m_sw_db.m_sw_map.end(); ++sw_node_it) {

        if (sw_node_it->second.in_temporary_error)
            continue;

        if(!IsDFActive(sw_node_it->second)) {
            OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG, "Switch GUID " U64H_FMT ", LID %u: " \
                    "DF/AR not supported or not enabled, ARLFTTableProcessDF skipped.\n",
                    sw_node_it->second.m_general_sw_info.m_guid,
                    sw_node_it->second.m_general_sw_info.m_lid);
            continue;
        }

        for (u_int8_t  plft_id=0;
             plft_id < sw_node_it->second.m_p_df_data->plft_number;
             plft_id++) {

            ARLFTTableProcess(
                sw_node_it->second,
                sw_node_it->second.m_p_df_data->m_plft[plft_id].m_max_lid,
                plft_id,
                sw_node_it->second.m_p_df_data->m_plft[plft_id].m_to_set_lft_table,
                sw_node_it->second.m_p_df_data->m_plft[plft_id].m_ar_lft);

            if (sw_node_it->second.m_p_df_data->m_plft[plft_id].m_set_lft_top)
                SetLftTop(sw_node_it->second,
                          &sw_node_it->second.m_p_df_data->m_plft[plft_id], //!!! TAMIR add to OSM
                          plft_id);

        }
    }

    //Receive all Set ARLFTTable reply
    m_ibis_obj.MadRecAll();

    if (this->m_ar_clbck.m_errcnt[AR_CLBCK_SET_AR_LFT_TABLE] ||
        this->m_ar_clbck.m_errcnt[AR_CLBCK_SET_PLFT_MAP]) {
        OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
                "Set AR LFT error count: %u.\n",
                m_ar_clbck.m_errcnt[AR_CLBCK_SET_AR_LFT_TABLE]+
                m_ar_clbck.m_errcnt[AR_CLBCK_SET_PLFT_MAP]);
        m_is_temporary_error = true; //Need to recalculate AR groups and LFTS
    }

    OSM_AR_LOG_RETURN(this->m_p_osm_log, rc);
}

int AdaptiveRoutingManager::ARMapPLFTsAndVL2VLs()
{
    OSM_AR_LOG_ENTER(this->m_p_osm_log);
    OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_VERBOSE, "ARMapPLFTsAndVL2VLs.\n");

    int rc = 0;

    PortsBitset    new_up_ports;
    PortsBitset    new_down_ports;

    list<ARSWDataBaseEntry*> update_list;

    //clbck
    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &this->m_ar_clbck;
    clbck_data.m_handle_data_func = SetExtendedSwitchInfoClbckDlg;

    map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it =
        this->m_sw_db.m_sw_map.begin();
    for ( ; sw_node_it != this->m_sw_db.m_sw_map.end();  ++sw_node_it) {
        DfSwData * p_df_data = sw_node_it->second.m_p_df_data;

        if (sw_node_it->second.in_temporary_error)
            continue;

        if(!IsDFActive(sw_node_it->second)) {
            OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG, "Switch GUID " U64H_FMT ", LID %u: " \
                    "DF/AR not supported or not enabled, ARMapPLFTsAndVL2VLs skipped.\n",
                    sw_node_it->second.m_general_sw_info.m_guid,
                    sw_node_it->second.m_general_sw_info.m_lid);
            continue;
        }

        OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
                   "Map pLFT and VL2VL on Switch GUID " U64H_FMT ", LID %u" \
                   " Old up:%s down:%s New up:%s down:%s \n",
                   sw_node_it->second.m_general_sw_info.m_guid,
                   sw_node_it->second.m_general_sw_info.m_lid,
                   p_df_data->m_df_sw_setup[OLD_DATA].m_up_ports.to_string().c_str(),
                   p_df_data->m_df_sw_setup[OLD_DATA].m_down_ports.to_string().c_str(),
                   p_df_data->m_df_sw_setup[NEW_DATA].m_up_ports.to_string().c_str(),
                   p_df_data->m_df_sw_setup[NEW_DATA].m_down_ports.to_string().c_str());

        new_up_ports =
            p_df_data->m_df_sw_setup[NEW_DATA].m_up_ports &
            ~p_df_data->m_df_sw_setup[OLD_DATA].m_up_ports;

        new_down_ports =
            p_df_data->m_df_sw_setup[NEW_DATA].m_down_ports &
            ~p_df_data->m_df_sw_setup[OLD_DATA].m_down_ports;

        if (new_up_ports.none() && new_down_ports.none())
            continue;

        PortsBitset new_ports = new_up_ports | new_down_ports;
        update_list.push_back(&sw_node_it->second);

        if (p_df_data->m_set_ext_sw_info) {
            struct SMP_ExtSWInfo ext_sw_info;
            ext_sw_info.SL2VL_Act = 1;

            clbck_data.m_data1 = &sw_node_it->second;
            //!!! TAMIR add to OSM
            clbck_data.m_data2 = &sw_node_it->second.m_p_df_data->m_set_ext_sw_info;

            this->ExtendedSwitchInfoMadGetSetByDirect(
                &sw_node_it->second.m_general_sw_info.m_direct_route,
                IBIS_IB_MAD_METHOD_SET,
                &ext_sw_info, &clbck_data);

        }

        u_int8_t port_block;
        u_int8_t prev_port_block = (u_int8_t)-1;
        for (u_int8_t port_num = 0;
             port_num <= sw_node_it->second.m_general_sw_info.m_num_ports;
             ++port_num) {

            if (!new_ports.test(port_num))
                continue;

            port_block = port_num / PLFT_MAP_PORT_BLOCK_SIZE;

            if (port_block != prev_port_block) {
                ARMapPLFTs(sw_node_it->second, port_block);
                prev_port_block = port_block;
            }

            ARMapVL2VL(sw_node_it->second, port_num);
        }
    }

    //Receive all Set Private LFTMap and VL2VL reply
    m_ibis_obj.MadRecAll();

    if (this->m_ar_clbck.m_errcnt[AR_CLBCK_SET_EXT_SW_INFO] ||
        this->m_ar_clbck.m_errcnt[AR_CLBCK_SET_PLFT_MAP]    ||
        this->m_ar_clbck.m_errcnt[AR_CLBCK_SET_VL_TO_VL_MAP]){

        OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
                   "MapPLFTsAndVL2VLs error count: %u.\n",
                   m_ar_clbck.m_errcnt[AR_CLBCK_SET_EXT_SW_INFO]+
                   m_ar_clbck.m_errcnt[AR_CLBCK_SET_PLFT_MAP]+
                   m_ar_clbck.m_errcnt[AR_CLBCK_SET_VL_TO_VL_MAP]);
        m_is_temporary_error = true; //Need to recalculate AR groups and LFTS
    }

    list<ARSWDataBaseEntry*>::iterator update_list_iter = update_list.begin();
    for (; update_list_iter != update_list.end() ; ++update_list_iter) {

        if ((*update_list_iter)->in_temporary_error)
            continue;

        (*update_list_iter)->m_p_df_data->m_df_sw_setup[OLD_DATA].m_down_ports |=
            (*update_list_iter)->m_p_df_data->m_df_sw_setup[NEW_DATA].m_down_ports;

        (*update_list_iter)->m_p_df_data->m_df_sw_setup[OLD_DATA].m_up_ports |=
            (*update_list_iter)->m_p_df_data->m_df_sw_setup[NEW_DATA].m_up_ports;
    }

    OSM_AR_LOG_RETURN(this->m_p_osm_log, rc);
}
//VL:  7  6  5  4  3  2  1  0  15 13 12 11 10  9  8  7
SMP_SLToVLMappingTable slvl_mapping[]     =
   {{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, //port 0 SL2VL
    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, // 1 vls ???
    {1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0}, // 2 vls
    {3, 2, 1, 0, 3, 2, 1, 0, 3, 2, 1, 0, 3, 2, 1, 0}, // 4 vls
    {7, 6, 5, 4, 3, 2, 1, 0, 7, 6, 5, 4, 3, 2, 1, 0}, // 8 vls
    {7, 6, 5, 4, 3, 2, 1, 0, 15, 14, 13, 12, 11, 10, 9, 8}}; // 16 vls

//origival value {7, 6, 5, 4, 3, 2, 1, 0, 15, 6, 5, 4, 3, 2, 1, 0};

SMP_SLToVLMappingTable inc_slvl_mapping[] =
   {{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, //reserved
    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, // 1 vls ???
    {1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1}, // 2 vls
    {3, 3, 1, 1, 3, 3, 1, 1, 3, 3, 1, 1, 3, 3, 1, 1}, // 4 vls
    {7, 7, 5, 5, 3, 3, 1, 1, 7, 7, 5, 5, 3, 3, 1, 1}, // 8 vls
    {7, 7, 5, 5, 3, 3, 1, 1, 15, 15, 13, 13, 11, 11, 9, 9}}; // 16 vls

//origival value {7, 7, 5, 5, 3, 3, 1, 1, 15, 15, 5, 5, 3, 3, 1, 1};

//validate that op_vl >1 or that out port is not switch
int AdaptiveRoutingManager::GetOpVlForVL2VL(ARSWDataBaseEntry &sw_db_entry,                                            
                                            u_int8_t out_port,
                                            u_int8_t &op_vls)
{
    op_vls = 0;
    osm_node_t *p_osm_node = sw_db_entry.m_general_sw_info.m_p_osm_sw->p_node;
    osm_physp_t *p_physp = osm_node_get_physp_ptr(p_osm_node, out_port);
    if (!p_physp || !osm_link_is_healthy(p_physp)) {
        OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
                   "Switch GUID " U64H_FMT ", LID %u, " \
                   "Fail to get physp:%u\n",
                   sw_db_entry.m_general_sw_info.m_guid,
                   sw_db_entry.m_general_sw_info.m_lid,
                   out_port);
        return -1;
    }

    //nuber of vls opcodes
    op_vls = ib_port_info_get_op_vls(&p_physp->port_info);

    if (op_vls < 2) {
        //no problem if remote is CA
        osm_physp_t *p_remote_physp = osm_physp_get_remote(p_physp);
        if (!p_remote_physp) {
            OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
                       "Switch GUID " U64H_FMT ", LID %u, " \
                       "Fail to get remote_physp:%u\n",
                       sw_db_entry.m_general_sw_info.m_guid,
                       sw_db_entry.m_general_sw_info.m_lid,
                       out_port);

            return -1;
        }
        u_int8_t remote_node_type = osm_node_get_type(p_remote_physp->p_node);

        if (remote_node_type == IB_NODE_TYPE_SWITCH) {
            OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
                       "Switch GUID " U64H_FMT ", LID %u, " \
                       "invalid op_vl: %u on port: %u for DFP. "
                       "Skip setting SL2VL\n",
                       sw_db_entry.m_general_sw_info.m_guid,
                       sw_db_entry.m_general_sw_info.m_lid,
                       op_vls, out_port);
            return -1;
        }
    }
    return 0;
}

void AdaptiveRoutingManager::ARMapVL2VL(ARSWDataBaseEntry &sw_db_entry,
                                        u_int8_t port_x)
{
    OSM_AR_LOG_ENTER(this->m_p_osm_log);

    osm_node_t *p_osm_node = sw_db_entry.m_general_sw_info.m_p_osm_sw->p_node;

    int rc;
    bool set_vl2vl_to_x = true;
    u_int8_t op_vls_x = 0;
    if (port_x == 0) {
        if(!ib_switch_info_is_enhanced_port0(&p_osm_node->sw->switch_info)) {
            //set port0 only if it is enhanced_port
            return;
        }
    } else {
        //set VL2VL considering op_vl unless it is port0
        //set all zerro for port0 VL2VL 
        rc = GetOpVlForVL2VL(sw_db_entry, port_x, op_vls_x);
        if(rc) {
            set_vl2vl_to_x = false;
        }
    }

    PortsBitset total_up_ports =
        sw_db_entry.m_p_df_data->m_df_sw_setup[NEW_DATA].m_up_ports |
        sw_db_entry.m_p_df_data->m_df_sw_setup[OLD_DATA].m_up_ports;

    PortsBitset total_down_ports =
        sw_db_entry.m_p_df_data->m_df_sw_setup[NEW_DATA].m_down_ports |
        sw_db_entry.m_p_df_data->m_df_sw_setup[OLD_DATA].m_down_ports;

    PortsBitset total_old_ports =
        sw_db_entry.m_p_df_data->m_df_sw_setup[OLD_DATA].m_up_ports |
        sw_db_entry.m_p_df_data->m_df_sw_setup[OLD_DATA].m_down_ports;

    //clbck
    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &this->m_ar_clbck;
    clbck_data.m_handle_data_func = SetVL2VLMapClbckDlg;
    clbck_data.m_data1 = &sw_db_entry;

    bool is_up_port_x = total_up_ports.test(port_x);
    SMP_SLToVLMappingTable *p_slvl_mapping;
    bool is_turn;

    for (u_int8_t port_y = 1;
         port_y <= sw_db_entry.m_general_sw_info.m_num_ports;
         ++port_y) {

        if (port_y == port_x)
            continue;

        //p_slvl_mapping = &slvl_mapping[op_vls_x];
        is_turn = false; //if turn use inc slvl mapping
        if (total_up_ports.test(port_y)) {
            if (is_up_port_x) {
                is_turn = true;
            }
        } else if (!total_down_ports.test(port_y))
            continue;

        if (set_vl2vl_to_x) {
            p_slvl_mapping = 
                (is_turn ? 
                 &inc_slvl_mapping[op_vls_x] : &slvl_mapping[op_vls_x]);

            OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
                       "Set VL2VL on Switch GUID " U64H_FMT ", LID %u, " \
                       "out port:%u in port:%u \n",
                       sw_db_entry.m_general_sw_info.m_guid,
                       sw_db_entry.m_general_sw_info.m_lid,
                       port_x, port_y);

            clbck_data.m_data2 = (void *)(intptr_t)port_x;
            clbck_data.m_data3 = (void *)(intptr_t)port_y;
            this->SMPSLToVLMappingTableGetSetByDirect(
                &sw_db_entry.m_general_sw_info.m_direct_route,
                IBIS_IB_MAD_METHOD_SET,
                port_x, port_y,
                p_slvl_mapping, &clbck_data);
        }

        //if y is new set y->x later when configure y port
        if (total_old_ports.test(port_y)) {

            bool set_vl2vl_to_y = true;
            u_int8_t op_vls_y;
            rc = GetOpVlForVL2VL(sw_db_entry, port_y, op_vls_y);
            if(rc) {
                set_vl2vl_to_y = false;
            }

            if (set_vl2vl_to_y) {
                p_slvl_mapping =
                    (is_turn ?
                     &inc_slvl_mapping[op_vls_y] : &slvl_mapping[op_vls_y]);

                OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
                           "Set VL2VL on Switch GUID " U64H_FMT ", LID %u, " \
                           "out port:%u in port:%u \n",
                           sw_db_entry.m_general_sw_info.m_guid,
                           sw_db_entry.m_general_sw_info.m_lid,
                           port_y, port_x);

                clbck_data.m_data2 = (void *)(intptr_t)port_y;
                clbck_data.m_data3 = (void *)(intptr_t)port_x;
                this->SMPSLToVLMappingTableGetSetByDirect(
                    &sw_db_entry.m_general_sw_info.m_direct_route,
                    IBIS_IB_MAD_METHOD_SET,
                    port_y, port_x,
                    p_slvl_mapping, &clbck_data);
            }
        }
    }

    OSM_AR_LOG_RETURN_VOID(this->m_p_osm_log);
}

SMP_SLToVLMappingTable op_vls_to_slvl_host_mapping[] =
   {{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, //reserved
    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, // 1 vls
    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, // 2 vls
    {2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0}, // 4 vls
    {6, 4, 2, 0, 6, 4, 2, 0, 6, 4, 2, 0, 6, 4, 2, 0}, // 8 vls
    {14, 12, 10, 8, 6, 4, 2, 0, 14, 12, 10, 8, 6, 4, 2, 0}}; // 16 vls

#define MAX_SL_NUMBER 16

void SetSLToVLMappingTable(SMP_SLToVLMappingTable *p_slvl_mapping,
                           u_int8_t slvl_array[MAX_SL_NUMBER])
{
    p_slvl_mapping->SL0ToVL = slvl_array[0];
    p_slvl_mapping->SL1ToVL = slvl_array[1];
    p_slvl_mapping->SL2ToVL = slvl_array[2];
    p_slvl_mapping->SL3ToVL = slvl_array[3];
    p_slvl_mapping->SL4ToVL = slvl_array[4];
    p_slvl_mapping->SL5ToVL = slvl_array[5];
    p_slvl_mapping->SL6ToVL = slvl_array[6];
    p_slvl_mapping->SL7ToVL = slvl_array[7];
    p_slvl_mapping->SL8ToVL = slvl_array[8];
    p_slvl_mapping->SL9ToVL = slvl_array[9];
    p_slvl_mapping->SL10ToVL = slvl_array[10];
    p_slvl_mapping->SL11ToVL = slvl_array[11];
    p_slvl_mapping->SL12ToVL = slvl_array[12];
    p_slvl_mapping->SL13ToVL = slvl_array[13];
    p_slvl_mapping->SL14ToVL = slvl_array[14];
    p_slvl_mapping->SL15ToVL = slvl_array[15];
}

void AdaptiveRoutingManager::ARMapSL2VLOnHosts()
{
    OSM_AR_LOG_ENTER(m_p_osm_log);
    OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_VERBOSE, "ARMapSL2VLOnHosts.\n");

    osm_node_t *p_osm_node;
    SMP_SLToVLMappingTable slvl_mapping;
    u_int8_t slvl_array[MAX_SL_NUMBER];
    u_int8_t op_vls; //nuber of vls opcode
    u_int64_t guid;
    u_int16_t lid;
    ARCADataBaseEntry ca_db_entry;

    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &m_ar_clbck;
    clbck_data.m_handle_data_func = SetSL2VLMapOnHostsClbckDlg;

    direct_route_t direct_route;

    //if all need update, invalidate all data
    if (m_p_osm_subn->need_update) {
        for (GuidToCADataBaseEntryMapIter iter = m_ca_db.m_ca_map.begin();
             iter != m_ca_db.m_ca_map.end();
             ++iter) {
            iter->second.m_to_set_slvl_mapping = true;
        }
    }

    // Set Leaf SW
    for (p_osm_node = (osm_node_t *) cl_qmap_head(&m_p_osm_subn->node_guid_tbl);
         p_osm_node != (osm_node_t *) cl_qmap_end(&m_p_osm_subn->node_guid_tbl);
         p_osm_node = (osm_node_t *) cl_qmap_next(&p_osm_node->map_item)) {

        //TODO should we configure AggNodes
        if (p_osm_node->node_info.node_type != IB_NODE_TYPE_CA) {
            continue;
        }

        for (u_int8_t port_num = 1;
             port_num <= p_osm_node->node_info.num_ports;
             port_num++) {

            osm_log_level_t dbg_log_level =
                (m_master_db.m_en_sl_mask == ALL_SL_VL_AR_ENABLED ?
                 OSM_LOG_VERBOSE : OSM_LOG_DEBUG);

            osm_physp_t *p_physp = osm_node_get_physp_ptr(p_osm_node, port_num);
            if (!p_physp || !osm_link_is_healthy(p_physp)) {
                continue;
            }
            guid = ADB2C_BE64_TO_CPU(p_physp->port_guid);
            lid = ADB2C_BE16_TO_CPU(p_physp->port_info.base_lid);

            if (!(p_physp->port_info.capability_mask & IB_PORT_CAP_HAS_SL_MAP)) {

                OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
                           "skip slvl_mapping on CA port GUID " U64H_FMT
                           ", LID %u SL_MAP not supported\n",
                           guid, lid);
                continue;
            }

            op_vls = ib_port_info_get_op_vls(&p_physp->port_info);

            OSM_AR_LOG(m_p_osm_log, dbg_log_level,
                       "Calculate slvl_mapping Host GUID " U64H_FMT
                       ", LID %u op_vls:%u en_sl_mask:0x%x\n",
                       guid, lid, op_vls, m_master_db.m_en_sl_mask);

            //Map SL to VL
            if ((m_master_db.m_en_sl_mask == ALL_SL_VL_AR_ENABLED) ||
                (op_vls <= 2)) {
                memcpy(&slvl_mapping, &op_vls_to_slvl_host_mapping[op_vls],
                       sizeof(slvl_mapping));
                if (dbg_log_level & m_p_osm_log->level) {
                    OSM_AR_LOG(m_p_osm_log, dbg_log_level,
                               "Static calculate slvl_mapping %s\n",
                               AdaptiveRoutingManager::ConvertSLToVLMappingToStr(
                                   &slvl_mapping).c_str());
                }

            } else {
                //Disabled SLs will be mapped (round robin) to VLs 0,1
                //Enabled SLs will be mapped (round robin) to VLs 2, .. max op vl
                u_int8_t ar_enabled_vl = 2;
                u_int8_t num_vls = ( 1 << (op_vls-1));
                u_int8_t max_ar_enabled_vl = num_vls - 2;
                u_int8_t ar_disabled_vl = 0;
                for (u_int8_t sl = 0; sl < MAX_SL_NUMBER; ++sl) {
                    if ((1 << sl) & m_master_db.m_en_sl_mask) {
                        //sl is AR enabled
                        slvl_array[sl] = ar_enabled_vl;

                        OSM_AR_LOG(m_p_osm_log, dbg_log_level,
                                   "slvl_mapping[%u]=%u AR enabled\n",
                                   sl, slvl_array[sl]);

                        ar_enabled_vl += 2;
                        if (ar_enabled_vl > max_ar_enabled_vl) {
                            ar_enabled_vl = 2;
                        }
                    } else {
                        slvl_array[sl] = ar_disabled_vl % 2;
                        OSM_AR_LOG(m_p_osm_log, dbg_log_level,
                                   "slvl_mapping[%u]=%u AR enabled\n",
                                   sl, slvl_array[sl]);
                        ar_disabled_vl++;
                    }
                }
                SetSLToVLMappingTable(&slvl_mapping, slvl_array);

                if (dbg_log_level & m_p_osm_log->level) {
                    OSM_AR_LOG(m_p_osm_log, dbg_log_level,
                               "Dynamic calculate slvl_mapping %s\n",
                               AdaptiveRoutingManager::ConvertSLToVLMappingToStr(
                                   &slvl_mapping).c_str());
                }

            }

            GuidToCADataBaseEntryMapIter iter = m_ca_db.m_ca_map.find(guid);

            if (iter != m_ca_db.m_ca_map.end()) {
                if (!p_physp->need_update &&
                    !iter->second.m_to_set_slvl_mapping &&
                    !memcmp(&(iter->second.m_slvl_mapping),
                            &slvl_mapping,
                            sizeof(slvl_mapping))){
                    continue;
                }
            } else {
                GuidToCADataBaseEntryMapInsertRes res =
                    m_ca_db.m_ca_map.insert(make_pair(guid, ca_db_entry));
                iter = res.first;
            }

            memcpy(&(iter->second.m_slvl_mapping), &slvl_mapping,
                   sizeof(slvl_mapping));
            iter->second.m_to_set_slvl_mapping = true;

            OSM_AR_LOG(m_p_osm_log, dbg_log_level,
                       "Set SL2VL on Host GUID " U64H_FMT ", LID %u \n",
                       guid, lid);

            ConvertOSMDRToIbisDR(osm_physp_get_dr_path_ptr(p_physp),
                                 &direct_route);

            clbck_data.m_data1 = &(iter->second); //CADataBaseEntry
            clbck_data.m_data2 = (void *)(intptr_t)guid;
            clbck_data.m_data3 = (void *)(intptr_t)lid;

            SMPSLToVLMappingTableGetSetByDirect(
                &direct_route,
                IBIS_IB_MAD_METHOD_SET,
                0, 0,
                &slvl_mapping, &clbck_data);

        } //CA's ports
    } // CAs

    m_ibis_obj.MadRecAll();

    OSM_AR_LOG_RETURN_VOID(m_p_osm_log);
}

int AdaptiveRoutingManager::ARDefinePLFTs()
{
    OSM_AR_LOG_ENTER(this->m_p_osm_log);
    OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_VERBOSE, "ARDefinePLFTs.\n");

    int rc = 0;
    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &this->m_ar_clbck;

    struct SMP_PrivateLFTInfo plft_info;
    CLEAR_STRUCT(plft_info);

    map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it = this->m_sw_db.m_sw_map.begin();

    //Set pLFT Info
    for ( ; sw_node_it != this->m_sw_db.m_sw_map.end(); ++sw_node_it) {

        if (sw_node_it->second.in_temporary_error)
            continue;

        if(!IsDFActive(sw_node_it->second)) {
            OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG, "Switch GUID " U64H_FMT ", LID %u: " \
                    "DF/AR not supported or not enabled, Set pLFT Info skipped.\n",
                    sw_node_it->second.m_general_sw_info.m_guid,
                    sw_node_it->second.m_general_sw_info.m_lid);

            // clear AR configuration from sw - SM should configure this sw
            sw_node_it->second.m_app_data = ARMGR_SW_NOT_UPDATED;

            continue;
        }


        if (sw_node_it->second.m_p_df_data->m_plft_info.ModeCap == 0) {

            //Set plft_info Write Values
            plft_info.Active_Mode =
                sw_node_it->second.m_p_df_data->m_plft_info.Active_Mode;
            //plft_info.no_fallback = 0;

            clbck_data.m_handle_data_func = SetPrivateLFTInfoClbckDlg;
            clbck_data.m_data1 = &sw_node_it->second;
            clbck_data.m_data2 = &sw_node_it->second.m_p_df_data->m_plft_info; //!!! TAMIR add to OSM

            this->PLFTInfoMadGetSetByDirect(
                &sw_node_it->second.m_general_sw_info.m_direct_route,
                IBIS_IB_MAD_METHOD_SET,
                &plft_info, &clbck_data);
        }

    }

    //Receive all Set Private LFTInfo reply
    m_ibis_obj.MadRecAll();

    if (this->m_ar_clbck.m_errcnt[AR_CLBCK_SET_PLFT_INFO]) {
        OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
                   "Set PLFT info error count: %u.\n",
                   m_ar_clbck.m_errcnt[AR_CLBCK_SET_PLFT_INFO]);
        m_is_temporary_error = true; //Need to recalculate AR groups and LFTS
    }

    SMP_PrivateLFTDef plft_def_2;

    CLEAR_STRUCT(plft_def_2);
    plft_def_2.LFT[0].FDB_LID_Offset = 0;
    plft_def_2.LFT[0].FDB_LID_Space = PLFT_LID_SPACE;
    plft_def_2.LFT[0].FDB_Table_IX = 0;

    plft_def_2.LFT[1].FDB_LID_Offset = 0;
    plft_def_2.LFT[1].FDB_LID_Space = PLFT_LID_SPACE;
    plft_def_2.LFT[1].FDB_Table_IX = 1;

    /*
    // for use only if we use 3 pLFTs
    SMP_PrivateLFTDef plft_def_3;
    plft_def_3 = plft_def_2;
    plft_def_3.LFT[2].FDB_LID_Offset = 0;
    plft_def_3.LFT[2].FDB_LID_Space = PLFT_LID_SPACE;
    plft_def_3.LFT[2].FDB_Table_IX = 2;
    */

    //Set pLFT Def
    u_int8_t   plft_number = 0;
    SMP_PrivateLFTDef *p_plft_def = NULL;
    clbck_data.m_handle_data_func = SetPrivateLFTDefClbckDlg;

    for (sw_node_it = this->m_sw_db.m_sw_map.begin();
         sw_node_it != this->m_sw_db.m_sw_map.end(); ++sw_node_it) {

        if(!IsDFActive(sw_node_it->second)) {
            OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG, "Switch GUID " U64H_FMT ", LID %u: " \
                    "DF/AR not supported or not enabled, Set pLFT Def skipped.\n",
                    sw_node_it->second.m_general_sw_info.m_guid,
                    sw_node_it->second.m_general_sw_info.m_lid);

            continue;
        }

        //We will currently use only 2 pLFTs for LEAFs and SPINEs
        plft_number = MIN_DF_PLFT_NUMBER; //!!! TAMIR add to OSM modified from MIN_PLFT_NUMBER
        p_plft_def = &plft_def_2;

        /*
        SMP_PrivateLFTInfo *p_plft_info =
            &sw_node_it->second.m_p_df_data->m_plft_info;

        if (sw_node_it->second.m_p_df_data->
            m_df_sw_setup[NEW_DATA].m_sw_type == SW_TYPE_LEAF) {

            plft_number = LEAF_PLFT_NUMBER;
            p_plft_def = &plft_def_2;

        } else if (p_plft_info->Description_Mode[p_plft_info->Active_Mode].
                   NumOfLFTables < MAX_PLFT_NUMBER) {

            plft_number = MIN_PLFT_NUMBER;
            p_plft_def = &plft_def_2;

        } else {
            plft_number = MAX_PLFT_NUMBER;
            p_plft_def = &plft_def_3;
        }
        */
        if (sw_node_it->second.in_temporary_error)
            continue;

        if (sw_node_it->second.m_p_df_data->plft_number == plft_number)
            continue; //already defined

        clbck_data.m_data1 = &sw_node_it->second;
        clbck_data.m_data2 = (void *)(intptr_t)plft_number;
        //!!! TAMIR add to OSM
        clbck_data.m_data3 = &sw_node_it->second.m_p_df_data->plft_number;

        this->PLFTDefMadGetSetByDirect(
            &sw_node_it->second.m_general_sw_info.m_direct_route,
            IBIS_IB_MAD_METHOD_SET,
            0, //block 0
            p_plft_def, &clbck_data);
    }

    //Receive all Set Private LFTDef reply
    m_ibis_obj.MadRecAll();

    if (m_ar_clbck.m_errcnt[AR_CLBCK_SET_PLFT_DEF]) {
        OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
                "Set pLFT Def error count: %u.\n",
                m_ar_clbck.m_errcnt[AR_CLBCK_SET_PLFT_DEF]);
        m_is_temporary_error = true; //Need to recalculate AR groups and LFTS
    }

    OSM_AR_LOG_RETURN(this->m_p_osm_log, rc);
}

//	PLFTMap for vl15 should be equal to vl0 settings
//PortSLToPLFT port_vl2plft_leaf_up      = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
// {SL3,SL2,SL1,SL0,SL7,SL6,SL5,SL4,SL11,SL10,SL9,SL8,SL15,SL14,SL13,SL12}
PortSLToPLFT port_vl2plft_leaf_down      = {1,1,1,1,1,1,1,1,0,0,0,0,1,0,0,0};

//PortSLToPLFT port_vl2plft_spine_up     = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
PortSLToPLFT port_vl2plft_spine_down     = {1,1,1,1,1,1,1,1,0,0,0,0,1,0,0,0};
PortSLToPLFT port_vl2plft_spine_down_vl  = {1,2,1,2,1,2,1,2,0,0,0,0,1,0,0,0};

void AdaptiveRoutingManager::ARMapPLFTs(ARSWDataBaseEntry &sw_db_entry,
                                        u_int8_t port_block)
{
    OSM_AR_LOG_ENTER(this->m_p_osm_log);

    //PortsBitset total_up_ports =
    //    sw_db_entry.m_p_df_data->m_df_sw_setup[NEW_DATA].m_up_ports |
    //    sw_db_entry.m_p_df_data->m_df_sw_setup[OLD_DATA].m_up_ports;

    PortsBitset total_down_ports =
        sw_db_entry.m_p_df_data->m_df_sw_setup[NEW_DATA].m_down_ports |
        sw_db_entry.m_p_df_data->m_df_sw_setup[OLD_DATA].m_down_ports;

    PortSLToPLFT *p_port_vl2plft_down = NULL;
    if (sw_db_entry.m_p_df_data->m_df_sw_setup[NEW_DATA].m_sw_type ==
        SW_TYPE_LEAF) {
        p_port_vl2plft_down = &port_vl2plft_leaf_down;
    } else if (sw_db_entry.m_p_df_data->plft_number == 2) {
        p_port_vl2plft_down = &port_vl2plft_spine_down;
    } else
        p_port_vl2plft_down = &port_vl2plft_spine_down_vl;


    //Set pLFT Map
    SMP_PortSLToPrivateLFTMap port_sl_to_plft_map;
    CLEAR_STRUCT(port_sl_to_plft_map);

    //clbck
    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &this->m_ar_clbck;
    clbck_data.m_handle_data_func = SetPortSlToPLFTMapClbckDlg;
    clbck_data.m_data1 = &sw_db_entry;
    clbck_data.m_data2 = (void *)(intptr_t)port_block;

    u_int8_t port_num = port_block*PLFT_MAP_PORT_BLOCK_SIZE;


    for (int port_idx = 0 ; port_idx < PLFT_MAP_PORT_BLOCK_SIZE ;
         ++port_num, ++port_idx) {

        if(total_down_ports.test(port_num)) {
            port_sl_to_plft_map.PortSLToPLFT[port_idx] = *p_port_vl2plft_down;
        }

        OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
                   "Map pLFT on Switch GUID " U64H_FMT ", LID %u, " \
                   "port_block:%u port_num:%u idx:%u is_down:%u total_down:%s\n",
                   sw_db_entry.m_general_sw_info.m_guid,
                   sw_db_entry.m_general_sw_info.m_lid,
                   port_block, port_num, port_idx,
                   total_down_ports.test(port_num),
                   total_down_ports.to_string().c_str());

    }

    this->PortSLToPrivateLFTMapGetSetByDirect(
        &sw_db_entry.m_general_sw_info.m_direct_route,
        IBIS_IB_MAD_METHOD_SET,
        port_block,
        &port_sl_to_plft_map,
        &clbck_data);


    OSM_AR_LOG_RETURN_VOID(this->m_p_osm_log);
}

void AdaptiveRoutingManager::SetLftTop(ARSWDataBaseEntry &sw_db_entry,
                                       PLFTMads *p_plft_mads, //!!! TAMIR add to OSM
                                       u_int8_t  pLFTID)
{
    OSM_AR_LOG_ENTER(this->m_p_osm_log);

    //Set pLFT Map
    SMP_PrivateLFTMap plft_map;
    CLEAR_STRUCT(plft_map);

    plft_map.LFT_TopEn = 1;
    plft_map.LFT_Top = p_plft_mads->m_max_lid; //!!! TAMIR add to OSM (code changed)

    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &this->m_ar_clbck;
    clbck_data.m_handle_data_func = SetPrivateLFTMapClbckDlg;
    clbck_data.m_data1 = &sw_db_entry;
    clbck_data.m_data2 = (void *)(intptr_t)pLFTID;
    clbck_data.m_data3 = p_plft_mads;  //!!! TAMIR add to OSM

    OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
            "Setting pLFT Top %u to "   \
            "Switch GUID " U64H_FMT ", LID %u pLFTID %u\n",
            plft_map.LFT_Top,
            sw_db_entry.m_general_sw_info.m_guid,
            sw_db_entry.m_general_sw_info.m_lid,
            pLFTID);

    AR_MGR_LOG(TT_LOG_LEVEL_INFO, "Setting pLFT Top %u to "   \
            "Switch GUID " U64H_FMT ", LID %u pLFTID %u\n",
            plft_map.LFT_Top,
            sw_db_entry.m_general_sw_info.m_guid,
            sw_db_entry.m_general_sw_info.m_lid,
            pLFTID);

    PLFTMapMadGetSetByDirect(
        &sw_db_entry.m_general_sw_info.m_direct_route,
        IBIS_IB_MAD_METHOD_SET,
        pLFTID,
        &plft_map,
        &clbck_data);

    OSM_AR_LOG_RETURN_VOID(this->m_p_osm_log);
}

