/*                  - Mellanox Confidential and Proprietary -
 *
 *  Copyright (C) 2010-2019, Mellanox Technologies Ltd.  ALL RIGHTS RESERVED.
 *
 *  Except as specifically permitted herein, no portion of the information,
 *  including but not limited to object code and source code, may be reproduced,
 *  modified, distributed, republished or otherwise exploited in any form or by
 *  any means for any purpose without the prior written permission of Mellanox
 *  Technologies Ltd. Use of software subject to the terms and conditions
 *  detailed in the file "COPYING".
 *
 */

/* osm_config.h needs to be in all cpp files before all other includes */
#include <opensm/osm_config.h>
#include <sstream>

#include "ar_kdor_algorithm.h"
#include "ar_mgr.h"
#include "ar_clbck.h"

int ArKdorAlgorithm::Preprocess()
{
    OSM_AR_LOG_ENTER(m_p_osm_log);
    int rc = 0;

    int unsupported_num = m_ar_mgr_.ARInfoGetProcess();
    if (unsupported_num) {
        OSM_AR_LOG(m_p_osm_log, OSM_LOG_INFO,
                   "%d switches do not support AR.\n" , unsupported_num);
    }

    rc = Init();

    //TODO: check if can be done before Init
    if (!rc && !SetCapable()) {
        rc = 1;
    }

    if (!rc) {
        SetPlftConfiguration();

        unsupported_num = m_ar_mgr_.ARInfoGetGroupCapProcess();
        if (unsupported_num) {
            OSM_AR_LOG(m_p_osm_log, OSM_LOG_INFO,
                       "%d switches do not support kDOR Cap.\n" ,
                       unsupported_num);
        }
    }

    OSM_AR_LOG(m_p_osm_log, OSM_LOG_VERBOSE,
               "ArKdorAlgorithm::Preprocess Ended. rc:%d\n", rc);
    OSM_AR_LOG_RETURN(m_p_osm_log, rc);
}

int ArKdorAlgorithm::RunCycle()
{
    OSM_AR_LOG_ENTER(m_p_osm_log);

    int rc = 0;

    for (int i=0 ; i<1 ; i++) {

        if(Preprocess()) {
            rc = -1;
            break;
        }

        rc = BuildStaticRouteInfo();
        if (rc) {
            m_ar_mgr_.m_is_permanent_error = true;
            break;
        }

        rc = CalculatePortGroups();
        if (rc) {
            break;
        }

        CalculateSl2VlMappingOnHosts(m_ar_mgr_.m_master_db.m_en_sl_mask);
        CalculateVl2VlMappingnOnSwitches();

        OSM_AR_LOG(m_p_osm_log, OSM_LOG_VERBOSE,
                   "ArKdorAlgorithm cycle calculation Ended.\n");

        int unsupported_num = m_ar_mgr_.ARInfoSetProcess();
        if (unsupported_num) {
            OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
                    "%d switches do not support AR Set.\n" , unsupported_num);
            m_ar_mgr_.m_is_temporary_error = true; //Need to recalculate AR groups and LFTS
        }

        SetExtendedNodeInfoProcess();
        MapHostsSl2VlProcess();
        MapSwitchesVl2VlProcess();
        SetPlftInfoProcess();
        SetPlftDefProcess();
        MapPlftsProcess();

        rc = GroupTableProcess();
        if (rc) {
            break;
        }

        rc = PlftProcess();
        if (rc){
            break;
        }
    }

    CycleEnd(rc);

    OSM_AR_LOG(m_p_osm_log, OSM_LOG_VERBOSE,
               "ArKdorAlgorithm::RunCycle Ended. rc:%d\n", rc);
    OSM_AR_LOG_RETURN(m_p_osm_log, rc);
}

int ArKdorAlgorithm::Init()
{
    OSM_AR_LOG_ENTER(m_p_osm_log);
    OSM_AR_LOG(m_p_osm_log, OSM_LOG_VERBOSE, "Init kDOR Algorithm.\n");

    if (!m_ar_mgr_.m_p_osm_subn->opt.dor_hyper_cube_mode) {
        OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
                "dor_hyper_cube_mode opensm configuration parameter must be set "
                "to TRUE for kDOR setup.\n");

        OSM_AR_LOG_RETURN(m_p_osm_log, -1);
    }

    m_ar_mgr_.m_ar_clbck.ResetErrorCount();
    m_ar_mgr_.m_ar_clbck.m_algorithm_feature = m_algorithm_feature_;

    memset(m_sw_lid_to_kdor_data_, 0, sizeof(m_sw_lid_to_kdor_data_));

    uint32_t sw_idx = 0;
    uint32_t switches_number = (uint32_t)m_sw_map_.size();
    map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it = m_sw_map_.begin();
    for ( ; sw_node_it != m_sw_map_.end();  ++sw_node_it) {
        if (sw_node_it->second.m_kdor_data == NULL) {
            //Allocate Dragon Flay Data for all switches
            sw_node_it->second.m_kdor_data = new KdorSwData(
                sw_node_it->second, &m_ar_mgr_.m_port_groups_calculator);
            if (sw_node_it->second.m_kdor_data == NULL) {
                OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
                           "Failed to allocate kDOR Data. kDOR failed.\n");
                OSM_AR_LOG_RETURN(m_p_osm_log, -1);
            }

            sw_node_it->second.m_kdor_data->Resize(
                sw_node_it->second.m_general_sw_info.m_p_osm_sw->p_node->
                physp_tbl_size);

            //Setting/Unsetting PLFT invalidate groups data and ARLFTs
            sw_node_it->second.ClearSubnetArTables();
        }

        int rc = sw_node_it->second.m_kdor_data->Init(
            *this, sw_node_it->second, sw_idx, switches_number);

        if (rc) {
            OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
                       "Failed to init kDOR Data. kDOR failed.\n");
            OSM_AR_LOG_RETURN(m_p_osm_log, rc);
        }
        m_sw_lid_to_kdor_data_[sw_node_it->second.m_general_sw_info.m_lid] =
            sw_node_it->second.m_kdor_data;
        sw_idx++;

        if(sw_node_it->second.m_osm_update_needed) {
            sw_node_it->second.m_kdor_data->m_plft_number_configured = 0;
            sw_node_it->second.m_kdor_data->m_required_plft_mode = 0;
            sw_node_it->second.m_kdor_data->m_get_ext_sw_info = true;
            sw_node_it->second.m_kdor_data->m_set_ext_sw_info = true;

            for (int i = 0 ; i < MAX_KDOR_PLFT_NUMBER ; ++i) {
                sw_node_it->second.m_kdor_data->m_plft[i].m_max_lid = 0;
            }
        }
    }

    OSM_AR_LOG_RETURN(this->m_p_osm_log, 0);
}

bool ArKdorAlgorithm::SetCapable()
{
    OSM_AR_LOG_ENTER(m_p_osm_log);

    OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,"SetCapable algorithm feature: %u\n",
               m_algorithm_feature_);

    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &m_ar_mgr_.m_ar_clbck;
    clbck_data.m_handle_data_func = SetExtendedSwitchInfoClbckDlg;

    struct SMP_PrivateLFTInfo plft_info;
    struct SMP_ExtSWInfo ext_sw_info;
    int unsupported_num = 0;
    int disable_by_tr_err = 0;
    bool new_dev = false;
    bool algorithm_supported = true;

    uint8_t num_vls = (1 << (m_ar_mgr_.m_p_osm_subn->opt.max_op_vls - 1));

    if (num_vls < m_min_vl_number_) {
        OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
                "Invalid maximal operational VLs %u < %u configuration "\
                "for kDOR setup.\n",
                num_vls, m_min_vl_number_);

        OSM_AR_LOG_RETURN(m_p_osm_log, false);
    }

    map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it = m_sw_map_.begin();
    for ( ; sw_node_it != m_sw_map_.end(); ++sw_node_it) {

        if (sw_node_it->second.m_osm_update_needed) {
            sw_node_it->second.m_support[m_algorithm_feature_] = SUPPORT_UNKNOWN;
        } else if(sw_node_it->second.m_support[m_algorithm_feature_] == NOT_SUPPORTED) {
            unsupported_num++;
            continue;
        }

        if (!m_ar_mgr_.IsARActive(sw_node_it->second)) {
            OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
                    "Skip Switch GUID: " U64H_FMT ", LID: %u - " \
                    "AR not supported or not enabled.\n",
                    sw_node_it->second.m_general_sw_info.m_guid,
                    sw_node_it->second.m_general_sw_info.m_lid);
            unsupported_num++;
            continue;
        }

        clbck_data.m_data1 = &sw_node_it->second;
        clbck_data.m_data2 = &sw_node_it->second.m_kdor_data->m_plft_info;
        clbck_data.m_data3 = (void *)(intptr_t)m_algorithm_feature_;

        //All switches should support PLFT at least 2 banks of at least 48K (PrivateLFTInfo)
        if (sw_node_it->second.m_kdor_data->m_required_plft_mode == 0) {
            clbck_data.m_handle_data_func = GetPrivateLFTInfoKdorClbckDlg;
            m_ar_mgr_.PLFTInfoMadGetSetByDirect(
                &sw_node_it->second.m_general_sw_info.m_direct_route,
                IBIS_IB_MAD_METHOD_GET,
                &plft_info, &clbck_data);
        }

        if (sw_node_it->second.m_support[m_algorithm_feature_] == SUPPORTED) {
            continue;
        }

        if (sw_node_it->second.in_temporary_error)
            continue;

        new_dev = true;

        if (sw_node_it->second.m_general_sw_info.m_p_osm_sw->coord ==
            OSM_SW_NO_COORD) {
            OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
                    "Switch GUID " U64H_FMT ", LID %u, "
                    "do not have coordinate. "
                    "Cannot guarantee credit loop free.\n",
                    sw_node_it->second.m_general_sw_info.m_guid,
                    sw_node_it->second.m_general_sw_info.m_lid);
        }
        //All switches should support vl2vl configuration ( ExtendedSwitchInfo  SL2VLCap  bit 0 is on)
        if (&sw_node_it->second.m_kdor_data->m_get_ext_sw_info) {

            clbck_data.m_handle_data_func = GetExtendedSwitchInfoClbckDlg;
            clbck_data.m_data2 = &sw_node_it->second.m_kdor_data->m_get_ext_sw_info;
            clbck_data.m_data3 = &sw_node_it->second.m_kdor_data->m_set_ext_sw_info;

            m_ar_mgr_.ExtendedSwitchInfoMadGetSetByDirect(
                &sw_node_it->second.m_general_sw_info.m_direct_route,
                IBIS_IB_MAD_METHOD_GET,
                &ext_sw_info, &clbck_data);
        }

        if ((m_ar_mgr_.m_master_db.m_disable_tr_mask != 0x0) &&
            (!sw_node_it->second.m_ar_info.by_transp_cap)) {
            OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
                    "Switch GUID " U64H_FMT ", LID %u, "
                    "do not support by_transport_disable.\n",
                    sw_node_it->second.m_general_sw_info.m_guid,
                    sw_node_it->second.m_general_sw_info.m_lid);
            disable_by_tr_err++;
        }

        if (sw_node_it->second.m_ar_info.glb_groups != 1) {
            OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
                       "Switch GUID: " U64H_FMT ", LID: %u - "
                       "Only global AR groups supported.\n",
                       sw_node_it->second.m_general_sw_info.m_guid,
                       sw_node_it->second.m_general_sw_info.m_lid);
            sw_node_it->second.MarkSWNotSupport(m_algorithm_feature_, DF_GLOBAL_AR_GROUP);
        }

        //All links between switches should support at least m_min_vl_number_
        //VLs in PortInfo VLCap
        //We check VL cap and not operational VLs because configuring
        //VLs comes after routing calculation
        osm_node_t *p_osm_node = sw_node_it->second.m_general_sw_info.m_p_osm_sw->p_node;
        osm_physp_t *p_physp_0 = osm_node_get_physp_ptr(p_osm_node, 0);
        if (!p_physp_0) {
            OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
                       "Switch GUID: " U64H_FMT ", LID: %u - "
                       "Failed to get phys port 0.\n",
                       sw_node_it->second.m_general_sw_info.m_guid,
                       sw_node_it->second.m_general_sw_info.m_lid);
            sw_node_it->second.MarkSWNotSupport(m_algorithm_feature_, DB_ERR);
        }

        if (p_physp_0 &&
            !(p_physp_0->port_info.capability_mask & IB_PORT_CAP_HAS_SL_MAP)) {
            OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
                       "Switch GUID: " U64H_FMT ", LID: %u - "
                       "SL to VL map not supported. kDOR algorithm not supported.\n",
                       sw_node_it->second.m_general_sw_info.m_guid,
                       sw_node_it->second.m_general_sw_info.m_lid);
            sw_node_it->second.MarkSWNotSupport(m_algorithm_feature_, DF_NO_SL_MAP_CAP);
            algorithm_supported = false;
        }

        for(u_int8_t port_num = 1; port_num <= p_osm_node->node_info.num_ports;
            port_num++) {

            osm_physp_t *p_physp = osm_node_get_physp_ptr(p_osm_node, port_num);
            if (!p_physp || !osm_link_is_healthy(p_physp))
                continue;

            osm_physp_t *p_remote_physp = osm_physp_get_remote(p_physp);
            if (!p_remote_physp)
                continue;

            u_int8_t remote_node_type = osm_node_get_type(p_remote_physp->p_node);
            if (remote_node_type != IB_NODE_TYPE_SWITCH) {
                continue;
            }

            u_int8_t vl_cap = ib_port_info_get_vl_cap(&p_physp->port_info);
            num_vls = (1 << (vl_cap - 1));

            if (num_vls < m_min_vl_number_) {
                OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
                        "Invalid number of VLs: %u < %u on Switch GUID " U64H_FMT \
                        ", LID %u, port:%u\n",
                        vl_cap, m_min_vl_number_,
                        sw_node_it->second.m_general_sw_info.m_guid,
                        sw_node_it->second.m_general_sw_info.m_lid,
                        port_num);

                sw_node_it->second.MarkSWNotSupport(m_algorithm_feature_,
                                                    DF_VL_CAP_LESS_THAN_MIN);
                break;

            }
        }
    }

    m_ar_mgr_.m_ibis_obj.MadRecAll();

    if (new_dev) {
        OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
                "Mark support algorithm feature:%u \n", m_algorithm_feature_);

        unsupported_num = 0;
        sw_node_it = m_sw_map_.begin();
        for ( ; sw_node_it != m_sw_map_.end(); ++sw_node_it) {

            if (sw_node_it->second.in_temporary_error)
                continue;

            if (m_ar_mgr_.m_ar_clbck.m_errcnt[AR_CLBCK_GET_EXT_SW_INFO]) {
                //m_errcnt is set if failed to get node info from one of the
                //switches or if one of them do not supports VL as SL mode
                OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
                           "VL as SL (VL to VL mapping) not supported on all switches. "
                           "kDOR algorithm not supported.\n");

                algorithm_supported = false;
            } else if(sw_node_it->second.m_support[m_algorithm_feature_] ==
                      NOT_SUPPORTED) {
                unsupported_num++;
            } else if (sw_node_it->second.m_support[m_algorithm_feature_] ==
                     SUPPORT_UNKNOWN) {
                OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
                        "Mark Switch GUID: " U64H_FMT ", LID: %u "
                        " support algorithm feature:%u \n",
                        sw_node_it->second.m_general_sw_info.m_guid,
                        sw_node_it->second.m_general_sw_info.m_lid,
                        m_algorithm_feature_);

                sw_node_it->second.MarkSWSupport(m_algorithm_feature_);
            }
        }
    }

    if (unsupported_num || disable_by_tr_err) {
        if (unsupported_num) {
            OSM_AR_LOG(m_p_osm_log, OSM_LOG_INFO,
                    "%d switches are not kDOR Capable.\n",
                    unsupported_num);
        }
        if (disable_by_tr_err) {
            OSM_AR_LOG(m_p_osm_log, OSM_LOG_INFO,
                    "%d switches do not support by_transport_disable.\n",
                    disable_by_tr_err);
        }
    }

    OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,"SetCapable returns %u\n",
               algorithm_supported);
    OSM_AR_LOG_RETURN(m_p_osm_log, algorithm_supported);
}

void ArKdorAlgorithm::SetPlftConfiguration()
{
    OSM_AR_LOG_ENTER(m_p_osm_log);
    int unsupported_num = 0;
    map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it = m_sw_map_.begin();
    for ( ; sw_node_it != m_sw_map_.end(); ++sw_node_it) {

        if (sw_node_it->second.in_temporary_error)
            continue;

        if(sw_node_it->second.m_support[m_algorithm_feature_] ==
                  NOT_SUPPORTED) {
            continue;
        }

        if ((sw_node_it->second.m_kdor_data->m_required_plft_mode == 0 ||
             sw_node_it->second.m_kdor_data->GetMaxLidNumber() <
             m_ar_mgr_.m_p_osm_subn->max_lid) &&
             SetPlftConfiguration(sw_node_it->second)) {
            unsupported_num++;
            sw_node_it->second.MarkSWNotSupport(m_algorithm_feature_,
                                                DF_PLFT_CAP_LESS_THAN_MIN);
        }
    }

    if (unsupported_num) {
        OSM_AR_LOG(m_p_osm_log, OSM_LOG_INFO,
                   "Failed to set PlftConfiguration for %u switches.\n",
                unsupported_num);
    }
    OSM_AR_LOG_RETURN_VOID(m_p_osm_log);
}

int ArKdorAlgorithm::SetPlftConfiguration(ARSWDataBaseEntry &db_entry)
{
    OSM_AR_LOG_ENTER(m_p_osm_log);

    //Set m_required_plft_mode
    struct SMP_PrivateLFTInfo &plft_info =
        db_entry.m_kdor_data->m_plft_info;

    u_int8_t ModeCap = plft_info.ModeCap;
    Description_Mode_Block_Element *p_block_element = NULL;
    u_int8_t Mode = 0;

    uint16_t max_lid_number = 0;
    uint8_t min_lftable_size = (m_ar_mgr_.m_p_osm_subn->max_lid / 1024) + 1;
    uint8_t max_lftable_size = 0;
    uint8_t lftable_size;
    uint8_t plfts_per_table;

    OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
            "SetPlftConfiguration: Switch GUID: " U64H_FMT ", LID: %u, "  \
            "ModeCap: %u, max_lid: %u, min_lftable_size: %u planes_number: %u.\n",
            db_entry.m_general_sw_info.m_guid,
            db_entry.m_general_sw_info.m_lid,
            ModeCap,
            m_ar_mgr_.m_p_osm_subn->max_lid,
            min_lftable_size,
            m_planes_number_);

    for (int i = 0 ; i < ModeCap ; i++) {
        p_block_element =
            &plft_info.Description_Mode[i];

        OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
                "Switch GUID: " U64H_FMT ", LID: %u, "  \
                "ModeCap: %u, Mode: %u, LFTables: %u, LFTables size: %u.\n",
                db_entry.m_general_sw_info.m_guid,
                db_entry.m_general_sw_info.m_lid,
                ModeCap, i,
                p_block_element->NumOfLFTables,
                p_block_element->LFTableSize);

        if ((p_block_element->LFTableSize / min_lftable_size) *
             p_block_element->NumOfLFTables < m_planes_number_) {
            continue;
        }

        //old code - one plft per table
        //if ( p_block_element->NumOfLFTables < m_planes_number_ )
        //    continue;
        //if (p_block_element->LFTableSize > max_lftable_size) {
        //    max_lftable_size = p_block_element->LFTableSize;
        //
        //    max_lid_number = LFTableSize * 1024 - 1;
        //
        //    if (max_lid_number > m_ar_mgr_.m_p_osm_subn->max_lid) {
        //        Mode = i;
        //        break;
        //    }
        //}

        if (p_block_element->NumOfLFTables < m_planes_number_) {
            plfts_per_table = ((m_planes_number_ - 1) /
                               p_block_element->NumOfLFTables) + 1;
        } else {
            plfts_per_table = 1;
        }

        lftable_size = p_block_element->LFTableSize / plfts_per_table;

        if (lftable_size > max_lftable_size) {
            max_lftable_size  = lftable_size;
            Mode = i;
            if (max_lftable_size > min_lftable_size) {
                break; //it is enough to have extra of 1K lids
            }
        }
    }

    if (max_lftable_size == 0) {
        OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_ERROR,
                "Switch GUID " U64H_FMT ", LID %u, "  \
                "Do not have sufficient PrivateLFT cap: "
                "(num tables < %u or max_lid_number < %u). \n",
                db_entry.m_general_sw_info.m_guid,
                db_entry.m_general_sw_info.m_lid,
                m_planes_number_,
                m_ar_mgr_.m_p_osm_subn->max_lid);

        db_entry.MarkSWNotSupport(m_algorithm_feature_, DF_PLFT_CAP_LESS_THAN_MIN);

        OSM_AR_LOG_RETURN(m_p_osm_log, 1);
    }

    db_entry.m_kdor_data->m_required_plft_mode = Mode+1;

    if (db_entry.m_kdor_data->m_required_plft_mode !=
        db_entry.m_kdor_data->m_plft_info.Active_Mode) {
        db_entry.ClearKdorPlftData();
    }

    OSM_AR_LOG(this->m_p_osm_log, OSM_LOG_DEBUG,
        "PLFT Info Get Switch GUID: " U64H_FMT ", LID: %u"
        " desired mode:%d Active_Mode:%d.\n",
        db_entry.m_general_sw_info.m_guid,
        db_entry.m_general_sw_info.m_lid,
        db_entry.m_kdor_data->m_required_plft_mode,
        plft_info.Active_Mode);

    //max lid number is restricted by plft table size
    max_lid_number = max_lftable_size * 1024 - 1;
    db_entry.m_kdor_data->SetMaxLidNumber(max_lid_number);

    //db_entry.m_kdor_data->m_plft_number_configured is set after
    //recieve MAD Set reply

    int rc = SetPlftDef(db_entry,
                        plft_info.Description_Mode[Mode].LFTableSize,
                        plft_info.Description_Mode[Mode].NumOfLFTables,
                        max_lftable_size, m_planes_number_,
                        db_entry.m_kdor_data->m_plft_def,
                        db_entry.m_kdor_data->m_set_plft_def);

    if (rc) {
        OSM_AR_LOG_RETURN(m_p_osm_log, rc);
    }

    SetPlftMap(db_entry);

    OSM_AR_LOG_RETURN(m_p_osm_log, 0);
}

int ArKdorAlgorithm::BuildStaticRouteInfo()
{
    OSM_AR_LOG_ENTER(m_p_osm_log);

    SwDbEntryPrtList bfs_q;

    map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it = m_sw_map_.begin();
    for ( ; sw_node_it != m_sw_map_.end();  ++sw_node_it) {

        if(BuildStaticRouteInfoToSwitch(bfs_q, sw_node_it->second)) {
            OSM_AR_LOG_RETURN(m_p_osm_log, 1);
        }
    }

    if (m_max_vlinc_ >= m_planes_number_) {
        OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
                   "Max required VL increase in static routing (%u) >= "
                   "max allowed kDOR planes (%u). "
                   "Cannot guarantee credit loop free.",
                   m_max_vlinc_, m_planes_number_);
        //TODO: Should we abort algorithm (return 1) ???
    }

    OSM_AR_LOG_RETURN(m_p_osm_log, 0);
}

int ArKdorAlgorithm::BuildStaticRouteInfoToSwitch(
    SwDbEntryPrtList &bfs_q,
    ARSWDataBaseEntry &dst_db_entry)
{
    ARSWDataBaseEntry *p_sw_db_entry;

    OSM_AR_LOG_ENTER(m_p_osm_log);
    OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
               "Build static routing info for dest switch GUID " U64H_FMT
               ", LID %u: \n",
               dst_db_entry.m_general_sw_info.m_guid,
               dst_db_entry.m_general_sw_info.m_lid);

    map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it = m_sw_map_.begin();
    for ( ; sw_node_it != m_sw_map_.end();  ++sw_node_it) {

        sw_node_it->second.m_kdor_data->InitRouteInfo();
    }

    bfs_q.push_back(&dst_db_entry);
    while (!bfs_q.empty()) {
        p_sw_db_entry = bfs_q.front();
        bfs_q.pop_front();

        OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
                   "Popped from BFS Q: Switch GUID: " U64H_FMT ", LID: %u \n",
                   p_sw_db_entry->m_general_sw_info.m_guid,
                   p_sw_db_entry->m_general_sw_info.m_lid);
        if (ProcessNeighborsBfs(bfs_q, p_sw_db_entry, &dst_db_entry)) {
            OSM_AR_LOG_RETURN(m_p_osm_log, 1);
        }
        p_sw_db_entry->m_kdor_data->SetRouteStatus(BFS_DONE);
    }
    OSM_AR_LOG_RETURN(m_p_osm_log, 0);
}

// Adds unvisited neighbors
int ArKdorAlgorithm::ProcessNeighborsBfs(SwDbEntryPrtList &bfs_q,
                                         ARSWDataBaseEntry *p_sw_db_entry,
                                         ARSWDataBaseEntry *p_dst_db_entry)
{
    OSM_AR_LOG_ENTER(m_p_osm_log);

    KdorRouteInfo min_route_into;
    KdorRouteInfo tmp_route_info;

    SwToConnectionMap &connections = p_sw_db_entry->m_kdor_data->GetConnections();
    uint16_t sw_distance = p_sw_db_entry->m_kdor_data->GetRouteDistance();

    OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
               "switch GUID: " U64H_FMT ", LID %u: connections num :%u\n",
               p_sw_db_entry->m_general_sw_info.m_guid,
               p_sw_db_entry->m_general_sw_info.m_lid,
               (int)connections.size());

    for (SwToConnectionMap::iterator connections_iter = connections.begin();
         connections_iter != connections.end(); ++connections_iter) {

        KdorConnection &connection = connections_iter->second;
        if (connection.m_remote_switch == NULL) {
            continue; //CAs connection
        }
        ARSWDataBaseEntry *p_remote_db_entry = connection.GetRemoteSwitchDbEntry();
        KdorSwData *p_remote_switch_data = p_remote_db_entry->m_kdor_data;

        OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
                   "Processing neighbor Switch GUID: " U64H_FMT ", LID: %u \n",
                   p_remote_db_entry->m_general_sw_info.m_guid,
                   p_remote_db_entry->m_general_sw_info.m_lid);

        if (p_remote_switch_data->GetRouteStatus() == BFS_INIT) {
            bfs_q.push_back(p_remote_db_entry);
            p_remote_switch_data->SetRouteStatus(BFS_QUEUED);
            p_remote_switch_data->SetRouteDistance(sw_distance + 1);
            OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
                       "Neighbor in BFS_INIT, adding to queue, distance: %d\n",
                       p_remote_switch_data->GetRouteDistance());
        }
        else if ((p_remote_switch_data->GetRouteStatus() == BFS_DONE) &&
                 (p_remote_switch_data->GetRouteDistance() + 1 == sw_distance)) {
            // Considering shortest path candidate only
            CalculateRouteInfo(connection,
                               p_remote_switch_data,
                               p_dst_db_entry->m_kdor_data,
                               tmp_route_info);

            OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
                       "Neighbor in BFS_DONE, "
                       "on shortest path to destination switch, "
                       "vl_inc: %u turn_type: %u\n",
                       tmp_route_info.m_vl_inc, tmp_route_info.m_turn_type);

            if (min_route_into.m_connection == NULL) {
                OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
                           "No best connection so far, take this neighbor as best\n");
                min_route_into = tmp_route_info;
            }
            else if (tmp_route_info < min_route_into) {
                OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
                           "This neighbor is better\n");
                min_route_into = tmp_route_info;
            }
        }
        else {
            OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
                       "Neighbor in BFS_QUEUED, do nothing\n");
        }
    }


    if (min_route_into.m_connection != NULL) { // Found the best neighbor for routing
        KdorRouteInfo &sw_route_info =
            p_sw_db_entry->m_kdor_data->GetRouteInfo(p_dst_db_entry->m_kdor_data);
        sw_route_info = min_route_into;

        uint16_t next_hop_lid = 0;
        if ((sw_route_info.m_connection != NULL) &&
            (sw_route_info.m_connection->m_remote_switch != NULL)) {

            next_hop_lid = ADB2C_BE16_TO_CPU(osm_node_get_base_lid(
                sw_route_info.m_connection->m_remote_switch->p_node, 0));
        }

        OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
               "Set RouteInfo from "
               "Switch GUID: " U64H_FMT ", LID: %u "
               "to switch GUID: " U64H_FMT ", LID: %u next_hop_lid: %u\n",
               p_sw_db_entry->m_general_sw_info.m_guid,
               p_sw_db_entry->m_general_sw_info.m_lid,
               p_dst_db_entry->m_general_sw_info.m_guid,
               p_dst_db_entry->m_general_sw_info.m_lid,
               next_hop_lid);

        if (m_max_vlinc_ < min_route_into.m_vl_inc) {
            m_max_vlinc_ = min_route_into.m_vl_inc;
            OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
                       "Max vlinc of %d for path from "
                       "Switch GUID: " U64H_FMT ", LID: %u "
                       "to switch GUID: " U64H_FMT ", LID: %u \n",
                       m_max_vlinc_,
                       p_sw_db_entry->m_general_sw_info.m_guid,
                       p_sw_db_entry->m_general_sw_info.m_lid,
                       p_dst_db_entry->m_general_sw_info.m_guid,
                       p_dst_db_entry->m_general_sw_info.m_lid);
        }
        if (min_route_into.m_vl_inc >= m_planes_number_) {
            OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
                       "Max required VL increase (%u) for path from "
                       "Switch GUID: " U64H_FMT ", LID: %u "
                       "to switch GUID " U64H_FMT ", LID %u >= "
                       "max allowed kDOR planes (%u). \n",
                       m_max_vlinc_,
                       p_sw_db_entry->m_general_sw_info.m_guid,
                       p_sw_db_entry->m_general_sw_info.m_lid,
                       p_dst_db_entry->m_general_sw_info.m_guid,
                       p_dst_db_entry->m_general_sw_info.m_lid,
                       m_planes_number_);
        }
    }

    OSM_AR_LOG_RETURN(m_p_osm_log, 0);
}

//    KDOR_TURN_TYPE_0,   //  +-Low   -> +-High
//    KDOR_TURN_TYPE_1,   //  +-High  ->  +Low
//    KDOR_TURN_TYPE_2    //  +-High  ->  -Low
kdor_turn_t ArKdorAlgorithm::GetTurnType(KdorConnection &from_connection,
                                         KdorConnection &to_connection)
{
    if (from_connection.m_remote_switch == NULL ||
        to_connection.m_remote_switch == NULL) {
        //from or to non switch port
        return KDOR_TURN_TYPE_0;
    }
    if (from_connection.m_dim_idx <= to_connection.m_dim_idx) {
        return KDOR_TURN_TYPE_0;
    }
    if (to_connection.m_dim_sign > 0) {
        return KDOR_TURN_TYPE_1;
    }
    return KDOR_TURN_TYPE_2;
}

//calculate new vl increase and turn type for given connection to sw
//
// Vl calculation if last turn type is T1:
//   T1 | T2 | T1 | T2 | T1
//   ----------------------
//      |    |    |    | 0      total vl_inc = 0
//   ----------------------
//      |    |    | 1  | 2      total vl_inc = 2 (prev vl inc + 2)
//   ----------------------
//      |    |  0 | 1  | 2      total vl_inc = 2 (prev vl inc)
//   ----------------------
//      |  1 |  2 | 3  | 4      total vl_inc = 4 (prev vl inc + 2)
//   ----------------------
//    0 |  1 |  2 | 3  | 4      total vl_inc = 4 (prev vl inc)
//
// Vl calculation if last turn type is T2:
//   T2 | T1 | T2 | T1 | T2
//   ----------------------
//      |    |    |    | 1      total vl_inc = 1
//   ----------------------
//      |    |    | 0  | 1      total vl_inc = 1 (prev vl inc)
//   ----------------------
//      |    |  1 | 2  | 3      total vl_inc = 2 (prev vl inc + 2)
//   ----------------------
//      |  0 |  1 | 2  | 3      total vl_inc = 4 (prev vl inc)
//   ----------------------
//    1 |  2 |  3 | 4  | 5      total vl_inc = 5 (prev vl inc + 2)
void ArKdorAlgorithm::CalculateRouteInfo(KdorConnection &connection,
                                         KdorRouteInfo &remote_route_info,
                                         KdorRouteInfo &route_info)
{
    OSM_AR_LOG_ENTER(m_p_osm_log);
    route_info.m_connection = &connection;

    kdor_turn_t turn_type = KDOR_TURN_TYPE_0;

    if (remote_route_info.m_connection != NULL) {
        //if remote is destination remote_route_info.m_connection == NULL
        turn_type = GetTurnType(connection, *remote_route_info.m_connection);
    }

    if(turn_type == KDOR_TURN_TYPE_0) { //TURN_TYPE_0
        route_info.m_vl_inc = remote_route_info.m_vl_inc;
        route_info.m_turn_type = remote_route_info.m_turn_type;
    } else if (turn_type == KDOR_TURN_TYPE_1) {//TURN_TYPE_1
        route_info.m_turn_type = KDOR_TURN_TYPE_1;
        //TURN_TYPE_1 starts on lowest vl (TURN_TYPE_2 VL -1) there for it
        //does not change vl_inc
        route_info.m_vl_inc = remote_route_info.m_vl_inc;

        OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
                   "TURN_TYPE_1, vl_inc remains: %u\n", route_info.m_vl_inc);
    } else { //TURN_TYPE_2
        route_info.m_turn_type = KDOR_TURN_TYPE_2;
        if (remote_route_info.m_turn_type == KDOR_TURN_TYPE_0) {
            route_info.m_vl_inc = remote_route_info.m_vl_inc + 1;
            OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
                       "TURN_TYPE_2 to TURN_TYPE_0, inc by 1 vl_inc: %u\n",
                       route_info.m_vl_inc);
        } else if (remote_route_info.m_turn_type == KDOR_TURN_TYPE_1) {
            route_info.m_vl_inc = remote_route_info.m_vl_inc + 2;
            OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
                       "TURN_TYPE_2 to TURN_TYPE_1, inc by 2 vl_inc: %u\n",
                       route_info.m_vl_inc);
        } else {
            route_info.m_vl_inc = remote_route_info.m_vl_inc;
            OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
                       "TURN_TYPE_2 to same turn type, vl_inc remains: %u\n",
                       route_info.m_vl_inc);
        }
    }
    OSM_AR_LOG_RETURN_VOID(m_p_osm_log);
}

ThreadPoolTask *ArKdorAlgorithm::GetCalculatePortGroupsTask(
        ARSWDataBaseEntry &db_entry)
{
    return db_entry.m_kdor_data->GetCalculatePortGroupsTask();
}

bool *ArKdorAlgorithm::GetIsToSetExtSwInfoPtr(ARSWDataBaseEntry &sw_db_entry)
{
    if (sw_db_entry.m_kdor_data) {
        return &sw_db_entry.m_kdor_data->m_set_ext_sw_info;
    }
    return NULL;
}

SlToVlPtrVecVec &ArKdorAlgorithm::GetVl2VlMapping(ARSWDataBaseEntry &sw_db_entry)
{
    return sw_db_entry.m_kdor_data->m_vl2vl;
}

BoolVecVec &ArKdorAlgorithm::GetIsToSetVl2VlMapping(ARSWDataBaseEntry &sw_db_entry)
{
    return sw_db_entry.m_kdor_data->m_is_to_set_vl2vl;
}

SMP_PrivateLFTInfo &ArKdorAlgorithm::GetPlftInfo(ARSWDataBaseEntry &sw_db_entry)
{
    return sw_db_entry.m_kdor_data->m_plft_info;
}

uint8_t ArKdorAlgorithm::GetRequiredPlftMode(ARSWDataBaseEntry &sw_db_entry)
{
    return sw_db_entry.m_kdor_data->m_required_plft_mode;
}

bool ArKdorAlgorithm::IsToSetPlftDef(ARSWDataBaseEntry &sw_db_entry)
{
    return (sw_db_entry.m_kdor_data->m_set_plft_def);
}

bool ArKdorAlgorithm::IsToSetPlftMap(ARSWDataBaseEntry &sw_db_entry)
{
    return (sw_db_entry.m_kdor_data->m_set_plft_map);
}

void ArKdorAlgorithm::ToSetPlftMap(ARSWDataBaseEntry &sw_db_entry, bool is_to_set)
{
    sw_db_entry.m_kdor_data->m_set_plft_map = is_to_set;
}

SMP_PrivateLFTDef *ArKdorAlgorithm::GetPlftDef(ARSWDataBaseEntry &sw_db_entry)
{
    return &sw_db_entry.m_kdor_data->m_plft_def;
}

SMP_PortSLToPrivateLFTMap **ArKdorAlgorithm::GetPlftMap(
    ARSWDataBaseEntry &sw_db_entry)
{
    return sw_db_entry.m_kdor_data->m_plft_map;
}

uint8_t *ArKdorAlgorithm::GetPlftNumberConfiguredPtr(
    ARSWDataBaseEntry &sw_db_entry)
{
    return &sw_db_entry.m_kdor_data->m_plft_number_configured;
}

PLFTMads *ArKdorAlgorithm::GetPlft(ARSWDataBaseEntry &sw_db_entry)
{
    return sw_db_entry.m_kdor_data->m_plft;
}

void ArKdorAlgorithm::RestorePriv(ARSWDataBaseEntry &sw_db_entry)
{
    if (sw_db_entry.m_kdor_data != NULL) {
        //set priv to point to it's origin value
        sw_db_entry.m_general_sw_info.m_p_osm_sw->priv =
            sw_db_entry.m_kdor_data->GetSavedPriv();
    }
}

void ArKdorAlgorithm::ClearSwitchAlgotithmData(ARSWDataBaseEntry &sw_db_entry)
{
    delete sw_db_entry.m_kdor_data;
    sw_db_entry.m_kdor_data = NULL;
}

int ArKdorAlgorithm::CalculatePortGroups()
{
    OSM_AR_LOG_ENTER(m_p_osm_log);
    int rc = m_ar_mgr_.m_port_groups_calculator.CalculatePortGroups(this);
    OSM_AR_LOG_RETURN(m_p_osm_log, rc);
}

void ArKdorAlgorithm::CalculateSwitchPortGroups(
    ARSWDataBaseEntry &sw_db_entry,
    LidMapping *p_lid_mapping)
{
    OSM_AR_LOG_ENTER(m_p_osm_log);

    /*
     * Step 1.1 -
     *  Ensure that this switch supports the algorithm.
     */
    if (!IsFeatureActive(sw_db_entry, m_algorithm_feature_)) {
        OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
                   "Skip switch GUID: " U64H_FMT ", LID: %u - "
                   "Algorithm not supported or not enabled.\n",
                   sw_db_entry.m_general_sw_info.m_guid,
                   sw_db_entry.m_general_sw_info.m_lid);

        OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
                   "Algorithm %u not supported or not enabled. "
                   "sw_db_entry.m_support[SUPPORT_AR]: %u "
                   "sw_db_entry.m_support[m_algorithm_feature_]: %u "
                   "sw_db_entry.m_option_on: %u\n",
                   m_algorithm_feature_,
                   sw_db_entry.m_support[SUPPORT_AR],
                   sw_db_entry.m_support[m_algorithm_feature_],
                   sw_db_entry.m_option_on);

        OSM_AR_LOG_RETURN_VOID(m_p_osm_log);
    }

    /*
     * Step 1.2 -
     *  Set default values for calculations
     */
    SMP_ARGroupTable calculated_ar_group_table[AR_GROUP_TABLE_NUM_BLOCKS];
    m_ar_mgr_.ARGroupTableSetDefault(calculated_ar_group_table);
    SMP_AR_LFT calculated_ar_plft[MAX_KDOR_PLFT_NUMBER];
    for (int i = 0; i < MAX_KDOR_PLFT_NUMBER ; ++i) {
        m_ar_mgr_.ARLftTableSetDefault(calculated_ar_plft[i].m_ar_lft_table_sx);
    }

    KdorAlgorithmData algorithm_data;

    /*
     * Step 2 - Calculate Ar Groups
     */
    CalculateArGroups(sw_db_entry,
                      p_lid_mapping->m_lid_to_sw_lid_mapping,
                      algorithm_data,
                      calculated_ar_group_table);

    /*
     * Step 3 - Calculate Ar PLFTs
     */
    CalculateArPlfts(sw_db_entry,
                     p_lid_mapping,
                     algorithm_data,
                     calculated_ar_plft);


    OSM_AR_LOG_RETURN_VOID(m_p_osm_log);
}

void ArKdorAlgorithm::CalculateArGroups(
    ARSWDataBaseEntry &sw_db_entry,
    u_int16_t *lid_to_sw_lid_mapping,
    KdorAlgorithmData &algorithm_data,
    SMP_ARGroupTable calculated_ar_group_table[])
{
    OSM_AR_LOG_ENTER(m_p_osm_log);

    OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
            "Calculate group bitmask for GUID " U64H_FMT ", LID %u: \n",
            sw_db_entry.m_general_sw_info.m_guid,
            sw_db_entry.m_general_sw_info.m_lid);

    /*
     * Step 2.1 -
     *  Iterate through all lids and calculate group bitmask for each lid.
     */
    osm_switch_t *p_osm_sw = sw_db_entry.m_general_sw_info.m_p_osm_sw;
    uint16_t max_lid = p_osm_sw->max_lid_ho;
    uint8_t **hops = p_osm_sw->hops;
    //last plft will include only static route (no AR)
    GroupData group_data_array[MAX_KDOR_PLFT_NUMBER - 1];
    //ports to neighbors with invalid dim
    GroupData invalid_dim_group_data_array[MAX_KDOR_PLFT_NUMBER - 1];
    int group_data_array_size = MAX_KDOR_PLFT_NUMBER - 1;
    KdorGroupData kdor_group_data;
    KdorRouteInfo route_info;
    KdorRouteInfo ingress_route_info;
    KdorConnection ingress_connection;
    ingress_connection.m_dim_idx = 0xFF; //max index TODO: optimize to max dim
    ingress_connection.m_remote_switch = p_osm_sw;

    uint16_t sw_lid_num;
    for (uint16_t lid_num = 1; lid_num <= max_lid; ++lid_num) {

        if (lid_to_sw_lid_mapping[lid_num] == lid_num) {
            //this is sw lid
            sw_lid_num = lid_num;
        } else {
            sw_lid_num = lid_to_sw_lid_mapping[lid_num];
            if (sw_lid_num == 0) {
                continue; //lid_num is invalid lid
            }
        }
        uint8_t least_hops = osm_switch_get_least_hops(p_osm_sw, sw_lid_num);
        OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
                   "lid_num: %u sw_lid_num: %u least_hops: %u\n",
                   lid_num , sw_lid_num, least_hops);

        if (least_hops == OSM_NO_PATH || least_hops == 0) {
            if(!hops[sw_lid_num]) {
                OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG, "hops[lid:%u] == NULL \n", sw_lid_num);
            }
            continue;
        }
#if 0 //Debug group calculation on a specific sw for specific destination sw lid
        osm_log_level_t log_level = OSM_LOG_DEBUG;
        u_int16_t debug_current_lid = 0;
        u_int16_t debug_destination_sw_lid = 0;
        if ((sw_db_entry.m_general_sw_info.m_lid == debug_current_lid) &&
            (sw_lid_num == debug_destination_sw_lid)) {
            log_level = OSM_LOG_INFO;
        }
#endif
        LidToKdorGroupDataMapIter lid_to_kdor_group_iter =
            algorithm_data.m_lid_to_kdor_group_map.find(sw_lid_num);
        if (lid_to_kdor_group_iter != algorithm_data.m_lid_to_kdor_group_map.end()) {

            algorithm_data.AddLidToKdorGroupData(
                lid_num,
                sw_lid_num,
                lid_to_kdor_group_iter->second,
                false,
                m_p_osm_log);
            continue;
        }

        KdorSwData *p_dest_switch_data = m_sw_lid_to_kdor_data_[sw_lid_num];
        if (p_dest_switch_data == NULL) {
            OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
                       "Failed to find kDOR data for switch lid %u\n",
                       sw_lid_num);
            continue;
        }

        for (int i = 0; i < group_data_array_size; ++i) {
            group_data_array[i].m_group_bitmask.reset();
            group_data_array[i].m_group_size = 0;
            invalid_dim_group_data_array[i].m_group_bitmask.reset();
            invalid_dim_group_data_array[i].m_group_size = 0;
        }

        /*
         *  Calculate group bitmask.
         */
        SwToConnectionMap &connections = sw_db_entry.m_kdor_data->GetConnections();

        for (SwToConnectionMap::iterator connections_iter = connections.begin();
             connections_iter != connections.end(); ++connections_iter) {

            KdorConnection &connection = connections_iter->second;

            if (m_p_osm_log->level & OSM_LOG_DEBUG) {
                OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
                           "Calculate AR group on Switch GUID " U64H_FMT ", LID %u: "
                           "to destination switch LID: %u %s\n",
                           sw_db_entry.m_general_sw_info.m_guid,
                           sw_db_entry.m_general_sw_info.m_lid,
                           sw_lid_num,
                           connection.ToString().c_str());
            }

            if (connection.m_remote_switch == NULL) {
                continue; //CAs connection
            }
            //first port
            u_int8_t port_num = (*connection.m_ports.begin())->port_num;
            if (osm_switch_get_hop_count(p_osm_sw, sw_lid_num, port_num) !=
                least_hops) {
                continue; //not shortest path
            }

            OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
                       "shortest path to LID: %u using port: %u \n",
                       sw_lid_num, port_num);

            ARSWDataBaseEntry *p_remote_db_entry =
                connection.GetRemoteSwitchDbEntry();
            KdorSwData *p_remote_switch_data = p_remote_db_entry->m_kdor_data;

            KdorRouteInfo &remote_route_info =
                p_remote_switch_data->GetRouteInfo(p_dest_switch_data);

            //calculate route info from curr sw
            CalculateRouteInfo(connection,
                               remote_route_info,
                               route_info);

            OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
                       "Route from remote SW LID: %u to destination SW LID: %u "
                       "vl_inc: %u turn_type: %u\n",
                       p_remote_db_entry->m_general_sw_info.m_lid,
                       sw_lid_num,
                       route_info.m_vl_inc,
                       route_info.m_turn_type);

            //calculate worst case route info from ingress to destination
            CalculateRouteInfo(ingress_connection,
                               route_info,
                               ingress_route_info);

            OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
                       "Route from local SW LID: %u to destination SW LID: %u "
                       "vl_inc: %u turn_type: %u\n",
                       sw_db_entry.m_general_sw_info.m_lid,
                       sw_lid_num,
                       ingress_route_info.m_vl_inc,
                       ingress_route_info.m_turn_type);
            //
            //            -----------        ------------         -------------
            // ---------> | curr sw | -----> | neighbor | ------> |destination|
            //  ingress   |   (U)   |        |    (V)   |  static |    (D)    |
            //            -----------        ------------         -------------
            //
            //
            //
            //the rout out of this port is valid if:
            //ingress plane + (is invalid turn from ingress to neighbor) +
            //min static vl increase to destination +
            //is invalid turn from current sw to neighbor static egress <
            //m_max_planes_
            //there for:
            //plft + ingress_route_info.m_vl_inc < m_max_planes_

            //The following table show the final VL/plft for given
            //in VL/plft (row) and ingress_route vl increase (column)

            // vl inc     | 0 (T1) | 1 (T2) | 2 (T1) | 3 (T2) | 4 (T1) | 5 (T2)
            // -----------|        |        |        |        |        |
            // in plft/vl |        |        |        |        |        |
            // -----------+--------+--------+--------+--------+--------+--------
            //     0      |   0    |   1    |   2    |   3    |   4    |   5
            // -----------+--------+--------+--------+--------+--------+--------
            //     1      |   2    |   1    |   4    |   3    |   6    |   5
            // -----------+--------+--------+--------+--------+--------+--------
            //     2      |   2    |   3    |   4    |   5    |   6    |   7
            // -----------+--------+--------+--------+--------+--------+--------
            //     3      |   4    |   3    |   6    |   5    |   8    |   7
            // -----------+--------+--------+--------+--------+--------+--------
            //     4      |   4    |   5    |   6    |   7    |   8    |   -
            // -----------+--------+--------+--------+--------+--------+--------
            //     5      |   6    |   5    |   8    |   7    |   -    |   -
            // -----------+--------+--------+--------+--------+--------+--------
            //
            //From the table:
            //
            //For ingress_route_info.m_turn_type == T1:
            //final_vl = (in_vl + 1) / 2 * 2 + vl_inc <= planes_number - 1
            //max_in_vl is even -> in_vl + vl_inc <= planes_number - 1
            // -> max_in_vl <= planes_number - vl_inc - 1
            //max_in_vl is even - round doun to even number
            //max_in_vl <= (planes_number - vl_inc - 1) / 2 * 2
            //
            //For ingress_route_info.m_turn_type == T2:
            //final_vl = (in_vl / 2) * 2 + vl_inc <= planes_number - 1
            //max_in_vl is odd -> in_vl - 1 + vl_inc <= planes_number - 1
            //in_vl <= planes_number - vl_inc
            //max_in_vl is odd - convert down to odd number
            //max_in_vl <= (planes_number - vl_inc + 1) / 2 * 2 - 1;

            //The folowing teble provide max_plft for given vl_inc and max planes
            // vl inc     | 0 (T1) | 1 (T2) | 2 (T1) | 3 (T2) | 4 (T1) | 5 (T2)
            // -----------|        |        |        |        |        |
            // max planes |        |        |        |        |        |
            // -----------+--------+--------+--------+--------+--------+--------
            //     4      |   2    |   3    |    0   |   1    |   -    |   -
            // -----------+--------+--------+--------+--------+--------+--------
            //     5      |   4    |   3    |    2   |   1    |   0    |   -
            // -----------+--------+--------+--------+--------+--------+--------
            //     6      |   4    |   5    |    2   |   3    |   0    |  1
            // -----------+--------+--------+--------+--------+--------+--------

            if (ingress_route_info.m_vl_inc >= m_planes_number_) {
                continue; //round doun works only for positive number
            }

            GroupData *p_group_data_array = group_data_array;
            if (remote_route_info.m_connection != NULL && // not the destination
                (remote_route_info.m_connection->m_dim_idx == INVALID_DIM_INDEX ||
                 route_info.m_connection->m_dim_idx == INVALID_DIM_INDEX)) {
                p_group_data_array = invalid_dim_group_data_array;
            }

            int max_plft;
            if (ingress_route_info.m_turn_type == KDOR_TURN_TYPE_2) {
                max_plft = (m_planes_number_ - ingress_route_info.m_vl_inc + 1)
                    / 2 * 2 - 1;
            } else {
                max_plft = (m_planes_number_ - ingress_route_info.m_vl_inc - 1)
                    / 2 * 2;
            }

            max_plft = std::min(max_plft, group_data_array_size - 1);
            for (int plft = 0; plft <= max_plft; ++plft) {
                p_group_data_array[plft].m_group_bitmask |=
                    connection.m_ports_bitset;
                p_group_data_array[plft].m_group_size +=
                    connection.m_ports.size();

                OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
                           "Add ports %s to group_data of plft_id: %u "
                           "total size: %u is_valid_dim: %u\n",
                           connection.m_ports_bitset.to_string().c_str(),
                           plft,
                           p_group_data_array[plft].m_group_size,
                           p_group_data_array == group_data_array);

            }
        }

        //if no route through neighbors with valid coordinate
        //use route through neighbors without valid coordinate
        if (invalid_dim_group_data_array[0].m_group_size) {

            for (int plft = 0; plft < group_data_array_size; ++plft) {
                if ((group_data_array[plft].m_group_size == 0) &&
                    (invalid_dim_group_data_array[plft].m_group_size != 0)) {

                    OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
                               "AR group on Switch GUID " U64H_FMT ", LID %u: "
                               "to destination switch LID: %u plft_id:%u "
                               "uses neighbors without valid coordinate.\n",
                               sw_db_entry.m_general_sw_info.m_guid,
                               sw_db_entry.m_general_sw_info.m_lid,
                               sw_lid_num,
                               plft);

                    group_data_array[plft].m_group_bitmask =
                        invalid_dim_group_data_array[plft].m_group_bitmask;
                    group_data_array[plft].m_group_size =
                        invalid_dim_group_data_array[plft].m_group_size;
                }
            }
        }

        if (m_p_osm_log->level & OSM_LOG_DEBUG) {
            for (int plft = 0; plft < group_data_array_size; ++plft) {
                OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
                           "AR group on Switch GUID " U64H_FMT ", LID %u: "
                           "to destination switch LID: %u plft_id:%u "
                           "is %s size: %u\n",
                           sw_db_entry.m_general_sw_info.m_guid,
                           sw_db_entry.m_general_sw_info.m_lid,
                           sw_lid_num,
                           plft,
                           group_data_array[plft].m_group_bitmask.to_string().c_str(),
                           group_data_array[plft].m_group_size);
            }
        }

        //build kdor_group_data with discovered group_data for each plft
        //group for plf == 0 is the largest groupt
        if (group_data_array[0].m_group_size > 1) {

            for (int plft = 0; plft < group_data_array_size; ++plft) {
                if (group_data_array[plft].m_group_size > 1) {
                    if (m_p_osm_log->level & OSM_LOG_DEBUG) {
                        OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
                                   "Insert group bitmask: %s to "
                                   "algorithm_data groups_map\n",
                                   group_data_array[plft].m_group_bitmask.
                                   to_string().c_str());
                    }

                    GroupBitmaskToDataMapInsertRes res =
                        algorithm_data.m_groups_map.insert(
                            make_pair(group_data_array[plft].m_group_bitmask,
                                      group_data_array[plft]));
                    kdor_group_data.group_data_array[plft] = &(res.first->second);
                } else {
                    kdor_group_data.group_data_array[plft] = NULL;
                }
            }

            //Save discovered kdor_group_data for new destination
            algorithm_data.AddLidToKdorGroupData(
                lid_num,
                sw_lid_num,
                kdor_group_data,
                true,
                m_p_osm_log);
        }
    }

    /*
     * Step 2.2 -
     *  try to assign group number for all groups.
     */
    OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG, "Step 2.2 AssignPortGroups \n");

    if (!algorithm_data.m_groups_map.empty()) {
        int rc = m_ar_mgr_.AssignPortGroups(sw_db_entry, algorithm_data);
        if (rc) {
            //sanity check, should not reach here
            OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
                       "Failed to assign port groups.\n");
        }
    }

    /*
     * Step 3 -
     *  Update calculated_ar_group_table:
     *  At this point we have all sggigned groups in algorithm_data.m_assign_groups list
     */
    unsigned calculated_max_pg_id = 0;
    if (!algorithm_data.m_assign_groups.empty()) {

        GroupsListIter list_iter = algorithm_data.m_assign_groups.begin();
        for ( ;
              list_iter != algorithm_data.m_assign_groups.end();
              ++list_iter) {
              UpdateCalculatedArGroupTable(
                  calculated_ar_group_table,
                  (*list_iter)->m_group_number,
                  (*list_iter)->m_group_bitmask,
                  calculated_max_pg_id);
        }
    }

    /*
     * Step 4 -
     *  Now we will decide if the calculated settings are different from old,
     *  or if we need to ignore old settings and use the new ones
     */

    m_ar_mgr_.ARUpdateSWGroupTable(sw_db_entry,
            calculated_ar_group_table,
            calculated_max_pg_id+1);

    OSM_AR_LOG_RETURN_VOID(m_p_osm_log);
}

void ArKdorAlgorithm::CalculateArPlfts(
    ARSWDataBaseEntry &sw_db_entry,
    LidMapping *p_lid_mapping,
    KdorAlgorithmData &algorithm_data,
    SMP_AR_LFT calculated_ar_plft[])
{
    OSM_AR_LOG_ENTER(m_p_osm_log);

    OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG, "Step 4 configure AR PLFTs \n");
    LidPortMappingPtrSX     lft_entry_ptr_sx;

    KdorConnection *dst_sw_lid_to_conection[IB_LID_UCAST_END_HO + 1];
    memset(dst_sw_lid_to_conection, 0, sizeof(dst_sw_lid_to_conection));
    BuildDstSwLidToConnection(sw_db_entry, dst_sw_lid_to_conection);

    //save calculated port for base lid (in order to select same port to all vlids)
    uint8_t base_lid_to_port[IB_LID_UCAST_END_HO + 1];
    memset(base_lid_to_port, 0, sizeof(base_lid_to_port));

    unsigned calculated_max_lid = 0;

    for (unsigned dest_lid = 1;
         dest_lid <= m_ar_mgr_.m_p_osm_subn->max_lid;
         ++dest_lid) {

        u_int8_t ucast_lft_port = GetStaticUcastLftPort(sw_db_entry,
                                                        p_lid_mapping,
                                                        dst_sw_lid_to_conection,
                                                        base_lid_to_port,
                                                        dest_lid);

        if (ucast_lft_port == OSM_NO_PATH)  //no route to this lid_num from this switch
            continue;
        calculated_max_lid = dest_lid;

        u_int8_t lft_entry_num = 0;
        u_int16_t lft_block_num = 0;

        lft_entry_num = dest_lid % AR_LFT_TABLE_BLOCK_SIZE_SX;
        lft_block_num = dest_lid / AR_LFT_TABLE_BLOCK_SIZE_SX;

        LidToKdorGroupDataMapConstIter ar_kdor_groups_iter =
            algorithm_data.m_lid_to_kdor_group_map.find(dest_lid);

        for (int plft = 0; plft < m_planes_number_; ++plft) {
            lft_entry_ptr_sx.SetPtr(&calculated_ar_plft[plft].
                                    m_ar_lft_table_sx[lft_block_num].
                                    LidEntry[lft_entry_num]);

            CalculateArPlft(sw_db_entry,
                            algorithm_data,
                            ucast_lft_port,
                            dest_lid,
                            plft,
                            &lft_entry_ptr_sx,
                            ar_kdor_groups_iter);

            if (m_p_osm_log->level & OSM_LOG_DEBUG) {
                OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
                           "LID %u dest_lid:%u - "
                           "calculated_ar_lft_table[%u].LidEntry[%u] "
                           "GroupNumber:%u LidState:%u\n",
                           sw_db_entry.m_general_sw_info.m_lid,
                           dest_lid,
                           lft_block_num, lft_entry_num,
                           lft_entry_ptr_sx.GetGroupNumber(),
                           lft_entry_ptr_sx.GetLidState());
            }

        }
    }      /* End of loop destination lids */

    for (int plft = 0; plft < m_planes_number_; ++plft) {
        UpdateSwitchPlftTable(sw_db_entry,
                              calculated_ar_plft[plft].m_ar_lft_table_sx,
                              sw_db_entry.m_kdor_data->m_plft[plft],
                              calculated_max_lid,
                              plft);
    }

    OSM_AR_LOG_RETURN_VOID(m_p_osm_log);
}

void ArKdorAlgorithm::CalculateArPlft(
    ARSWDataBaseEntry &sw_db_entry,
    KdorAlgorithmData &algorithm_data,
    u_int8_t ucast_lft_port,
    unsigned dest_lid,
    int plft,
    LidPortMappingPtrSX *p_lft_entry_ptr,
    LidToKdorGroupDataMapConstIter &ar_kdor_groups_iter)
{
    OSM_AR_LOG_ENTER(m_p_osm_log);

    OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
               "CalculateArPlfts LID %u dest_lid:%u static_port: %u plft: %u\n",
               sw_db_entry.m_general_sw_info.m_lid,
               dest_lid,
               ucast_lft_port,
               plft);

    if (plft == (m_planes_number_ - 1) || //last plane is static only
        ar_kdor_groups_iter == algorithm_data.m_lid_to_kdor_group_map.end() ||
        ar_kdor_groups_iter->second.group_data_array[plft] == NULL) {

        OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
                   "No AR group - Set lid sate Static Group: 0\n");

        p_lft_entry_ptr->SetData(AR_IB_LID_STATE_STATIC, 0, 0, ucast_lft_port);

        OSM_AR_LOG_RETURN_VOID(m_p_osm_log);
    }

    GroupData* p_group_data = ar_kdor_groups_iter->second.group_data_array[plft];

    if (p_group_data->m_group_table_number == ARMGR_MAX_GROUP_TABLES) {

        p_lft_entry_ptr->SetData(AR_IB_LID_STATE_STATIC, 0, 0, ucast_lft_port);

        if (m_p_osm_log->level & OSM_LOG_DEBUG) {
            OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
                       "No AR group because group bitmask:%s was not mapped.\n",
                       p_group_data->m_group_bitmask.to_string().c_str());
        }

    } else {
        uint16_t  group_number = p_group_data->m_group_number;
#if 0 //RN Not implemented
        if (is_group_per_leaf_sw) {

            uint16_t sw_lid_num;
            if (p_lid_mapping->m_lid_to_sw_lid_mapping[dest_lid] != dest_lid) {
                //this is not sw lid. get the sw lid connected to this lid
                sw_lid_num =
                    p_lid_mapping->m_lid_to_sw_lid_mapping[dest_lid];
            } else {
                sw_lid_num = dest_lid;
            }

            LidToGroupNumberMapConstIter sw_grp_iter =
                p_group_data->m_sw_lid_to_group_number.find(sw_lid_num);

            if (sw_grp_iter ==
                p_group_data->m_sw_lid_to_group_number.end()) {
                group_number = ar_kdor_groups_iter->second->m_group_number;
                /*
                OSM_AR_LOG(m_p_osm_log, OSM_LOG_INFO,
                           "LID %u dest_lid:%u failed to find "
                           "sw_lid_num:%u in sw_lid_to_group_number\n",
                           sw_db_entry.m_general_sw_info.m_lid,
                           dest_lid, sw_lid_num);
                           */
            } else
                group_number = sw_grp_iter->second;
        }
        enum SMP_AR_LID_STATE ar_mode = this->m_master_db.m_ar_mode;
        if ((is_group_per_leaf_sw && group_number == AR_GROUP_UNASSIGNED) ||
            !this->m_master_db.m_ar_enable) //only FR/FRN enabled
            ar_mode = AR_IB_LID_STATE_STATIC;
#endif
        enum SMP_AR_LID_STATE ar_mode = m_ar_mgr_.m_master_db.m_ar_mode;
        OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG, "LID %u dest_lid:%u - "
                   "Set lid state %u Group:%u\n",
                   sw_db_entry.m_general_sw_info.m_lid, dest_lid,
                   ar_mode,
                   group_number);

        p_lft_entry_ptr->SetData(ar_mode,
                                 p_group_data->m_group_table_number,
                                 group_number,
                                 ucast_lft_port);
    }

    OSM_AR_LOG_RETURN_VOID(m_p_osm_log);
}

uint8_t ArKdorAlgorithm::GetStaticUcastLftPort(
    ARSWDataBaseEntry &sw_db_entry,
    LidMapping *p_lid_mapping,
    KdorConnection *dst_sw_lid_to_conection[],
    uint8_t base_lid_to_port[],
    uint16_t dest_lid)
{
    OSM_AR_LOG_ENTER(m_p_osm_log);
    uint8_t port_number;
    KdorConnection *connection;
    if (p_lid_mapping->m_lid_to_sw_lid_mapping[dest_lid] == 0) {
        OSM_AR_LOG_RETURN(m_p_osm_log, OSM_NO_PATH);
    }

    if (p_lid_mapping->m_lid_to_sw_lid_mapping[dest_lid] == dest_lid) {
        //sw lid
        connection = dst_sw_lid_to_conection[dest_lid];
        if (connection == NULL) {
            if (dest_lid == sw_db_entry.m_general_sw_info.m_lid) {
                OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
                           "GetStaticUcastLftPort for local switch lid: %u "
                           "return: %u\n",
                           dest_lid, 0);
                OSM_AR_LOG_RETURN(m_p_osm_log, 0);
            }
            OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
                       "Failed to find static connection to "
                       "destination lid: %u\n",
                       dest_lid);
            OSM_AR_LOG_RETURN(m_p_osm_log, OSM_NO_PATH);
        }
        //No load balancing to switch lid
        port_number = (*connection->m_next_port_for_routing)->port_num;
        OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
                   "Switch GUID: " U64H_FMT ", LID: %u - "
                   "Use static port %u for destination switch lid: %u\n",
                   sw_db_entry.m_general_sw_info.m_guid,
                   sw_db_entry.m_general_sw_info.m_lid,
                   port_number, dest_lid);

        OSM_AR_LOG_RETURN(m_p_osm_log, port_number);
    }
    //CA lid
    uint16_t base_lid = p_lid_mapping->m_lid_to_base_lid_mapping[dest_lid];
    if (base_lid_to_port[base_lid] != 0) {
        port_number = base_lid_to_port[base_lid];

        OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
                   "Switch GUID: " U64H_FMT ", LID: %u - "
                   "Use static port %u for dest_lid: %u base_lid: %u\n",
                   sw_db_entry.m_general_sw_info.m_guid,
                   sw_db_entry.m_general_sw_info.m_lid,
                   port_number, dest_lid, base_lid);

        OSM_AR_LOG_RETURN(m_p_osm_log, port_number);
    }

    uint16_t sw_lid = p_lid_mapping->m_lid_to_sw_lid_mapping[dest_lid];
    connection = dst_sw_lid_to_conection[sw_lid];
    if (connection == NULL) {
        if (sw_lid == sw_db_entry.m_general_sw_info.m_lid) {
            port_number = OsmSwitchGetPortByLidUseNewLft(
                sw_db_entry.m_general_sw_info.m_p_osm_sw, base_lid);

            OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
                       "Switch GUID: " U64H_FMT ", LID: %u - "
                       "Use static port %u for dest_lid: %u base_lid: %u"
                       "on local switch\n",
                       sw_db_entry.m_general_sw_info.m_guid,
                       sw_db_entry.m_general_sw_info.m_lid,
                       port_number, dest_lid, base_lid);

            OSM_AR_LOG_RETURN(m_p_osm_log, port_number);
        }
        OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
                   "Failed to find static connection to "
                   "switch lid:%u of destination lid: %u\n",
                   sw_lid, dest_lid);
        OSM_AR_LOG_RETURN(m_p_osm_log, OSM_NO_PATH);
    }
    port_number = (*connection->m_next_port_for_routing)->port_num;
    base_lid_to_port[base_lid] = port_number;
    OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
               "GetStaticUcastLftPort for base_lid: %u "
               "of destination lid: %u set and return: %u\n",
               base_lid, dest_lid, port_number);

    //increment next_port_for_routing
    ++connection->m_next_port_for_routing;
    if (connection->m_next_port_for_routing == connection->m_ports.end()) {
        connection->m_next_port_for_routing = connection->m_ports.begin();
    }
    OSM_AR_LOG_RETURN(m_p_osm_log, port_number);
}

void ArKdorAlgorithm::BuildDstSwLidToConnection(
    ARSWDataBaseEntry &sw_db_entry,
    KdorConnection *dst_sw_lid_to_conection[])
{
    OSM_AR_LOG_ENTER(m_p_osm_log);

    KdorConnection *connection;
    map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it = m_sw_map_.begin();
    for ( ; sw_node_it != m_sw_map_.end();  ++sw_node_it) {
        if (&sw_node_it->second == &sw_db_entry) {
            continue;
        }
        ARSWDataBaseEntry &dst_sw_db_entry = sw_node_it->second;
        KdorRouteInfo &route_info_to_dst =
            sw_db_entry.m_kdor_data->GetRouteInfo(dst_sw_db_entry.m_kdor_data);
        connection = route_info_to_dst.m_connection;
        if (connection == NULL) {
            OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
                       "-A- Failed to get RouteInfo from "
                       "Switch GUID: " U64H_FMT ", LID: %u "
                       "to switch GUID: " U64H_FMT ", LID: %u\n",
                       sw_db_entry.m_general_sw_info.m_guid,
                       sw_db_entry.m_general_sw_info.m_lid,
                       dst_sw_db_entry.m_general_sw_info.m_guid,
                       dst_sw_db_entry.m_general_sw_info.m_lid);
            continue;
        }
        dst_sw_lid_to_conection[dst_sw_db_entry.m_general_sw_info.m_lid] =
            connection;
        connection->m_next_port_for_routing = connection->m_ports.begin();
    }

    OSM_AR_LOG_RETURN_VOID(m_p_osm_log);
}

void ArKdorAlgorithm::BuildKdorVl2VlPerOpVl()
{
    OSM_AR_LOG_ENTER(m_p_osm_log);

    //for turn type 1 increase vl if vl%2 == 1
    //for turn type 2 increase vl if vl%2 == 0
    m_turn_type_to_vl2vl_per_op_vls_[KDOR_TURN_TYPE_0] = m_vl2vl_per_op_vls_;
    m_turn_type_to_vl2vl_per_op_vls_[KDOR_TURN_TYPE_1] = m_turn_1_vl2vl_per_op_vls_;
    m_turn_type_to_vl2vl_per_op_vls_[KDOR_TURN_TYPE_2] = m_turn_2_vl2vl_per_op_vls_;



    //currently support only one vl per plane, there for:
    uint8_t vls_per_stream = m_planes_number_;

    OSM_AR_LOG(m_p_osm_log, OSM_LOG_VERBOSE, "BuildVl2Vl t1,t2 PerOpVl.\n");
    u_int8_t t1_vl2vl_array[MAX_SL_NUMBER]; //turn_type_1
    u_int8_t t2_vl2vl_array[MAX_SL_NUMBER]; //turn_type_2
    u_int8_t num_vls;
    u_int8_t max_vl;
    u_int8_t qos_max_vl; //max vl for the current qos
    u_int8_t t1_out_vl; //turn_type_1
    u_int8_t t2_out_vl; //turn_type_2

    //turn type 1 is valid on even vls ((in_vl % 2) == 0)
    //turn type 2 is valid on odd vls ((in_vl % 2) != 0)

    memset(m_turn_1_vl2vl_per_op_vls_, 0, sizeof(m_turn_1_vl2vl_per_op_vls_));
    memset(m_turn_2_vl2vl_per_op_vls_, 0, sizeof(m_turn_2_vl2vl_per_op_vls_));

    for (uint8_t op_vls = 2; op_vls <= MAX_OP_VL_CODE; ++op_vls) {

        num_vls = ( 1 << (op_vls-1));
        max_vl = num_vls - 1;

        //On invalid turn that my lead to loop we set to qos_max_vl

        for (u_int8_t in_vl = 0, vl = 0; in_vl < MAX_SL_NUMBER; ++in_vl, ++vl) {
            //there are (max_vl / vls_per_stream) qos groups
            //(vl / vls_per_stream) + 1 ) is the qos groups number of given vl [1..last]
            //in each qos group there are vls_per_stream vls so for vl in specific
            //qos group the max vl is:
            qos_max_vl = ((vl / vls_per_stream) + 1 ) * vls_per_stream - 1;

            //when moving to new qos group, if qos_max_vl > max_vl round robin
            //back to out vl = 0
            if ((in_vl % vls_per_stream == 0) && (qos_max_vl > max_vl)) {
                vl = 0;
                //recalculate qos_max_vl with new vl==0
                qos_max_vl = vls_per_stream - 1;
            }
            qos_max_vl = min (qos_max_vl, max_vl);

            if((in_vl % 2) == 0) {
                t1_out_vl = min(qos_max_vl, vl);
                t2_out_vl  = min(qos_max_vl, (uint8_t)(vl + 1));
            } else {
                t1_out_vl = min(qos_max_vl, (uint8_t)(vl + 1));
                t2_out_vl  = min(qos_max_vl, vl);
            }

            t1_vl2vl_array[in_vl] = t1_out_vl;
            t2_vl2vl_array[in_vl] = t2_out_vl;
        }

        SetSLToVLMappingTable(&m_turn_1_vl2vl_per_op_vls_[op_vls], t1_vl2vl_array);
        SetSLToVLMappingTable(&m_turn_2_vl2vl_per_op_vls_[op_vls], t2_vl2vl_array);

    } //op_vls loop

    if (OSM_LOG_DEBUG & m_p_osm_log->level) {
        OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
                   "Calculated kdor turns vl to vl mapping, vls_per_stream: %u\n",
                   vls_per_stream);

        for (uint8_t op_vls = 1; op_vls <= MAX_OP_VL_CODE; ++op_vls) {
            OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
                       "op_vls: %u turn type 1 vl to vl mapping %s\n",
                       op_vls,
                       AdaptiveRoutingManager::ConvertSLToVLMappingToStr(
                           &m_turn_1_vl2vl_per_op_vls_[op_vls]).c_str());
        }
        for (uint8_t op_vls = 1; op_vls <= MAX_OP_VL_CODE; ++op_vls) {
            OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
                       "op_vls: %u turn type 2 vl to vl mapping %s\n",
                       op_vls,
                       AdaptiveRoutingManager::ConvertSLToVLMappingToStr(
                           &m_turn_2_vl2vl_per_op_vls_[op_vls]).c_str());
        }
    }
    OSM_AR_LOG_RETURN_VOID(m_p_osm_log);
}

void ArKdorAlgorithm::BuildKdorPlftMap()
{
    OSM_AR_LOG_ENTER(m_p_osm_log);

    OSM_AR_LOG(m_p_osm_log, OSM_LOG_VERBOSE, "BuildKdorPlftMap\n");
    PortSLToPLFT vl_to_plft_map;
    u_int8_t vl_to_plft_array[MAX_SL_NUMBER];
    u_int8_t plft_id;

    memset(vl_to_plft_array, 0, sizeof(vl_to_plft_array));

    plft_id = 0;
    for (u_int8_t in_vl = 0; in_vl < MAX_SL_NUMBER; ++in_vl) {
        if ((in_vl % m_planes_number_) == 0) {
            plft_id = 0;
        }

        vl_to_plft_array[in_vl] = plft_id;
        plft_id++;
    }

    SetSlToPlftMap(&vl_to_plft_map, vl_to_plft_array);

    for (int i = 0; i < PLFT_MAP_PORT_BLOCK_SIZE; ++i) {
        memcpy(&m_plft_map_.PortSLToPLFT[i], &vl_to_plft_map,
               sizeof(vl_to_plft_map));
    }

    if (OSM_LOG_DEBUG & m_p_osm_log->level) {
        OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
                   "Calculated VL to plft mapping, planes_number: %u\n",
                   m_planes_number_);

        for (u_int8_t in_vl = 0; in_vl < MAX_SL_NUMBER; ++in_vl) {
            OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
                       "map VL: %u to plft id:%u\n",
                       in_vl, vl_to_plft_array[in_vl]);
        }
    }

    OSM_AR_LOG_RETURN_VOID(m_p_osm_log);
}


//Might be need to optimize, might use thread pool
void ArKdorAlgorithm::CalculateVl2VlMappingnOnSwitches()
{
    OSM_AR_LOG_ENTER(m_p_osm_log);
    map<u_int64_t, ARSWDataBaseEntry>::iterator sw_node_it = m_sw_map_.begin();
    for ( ; sw_node_it != m_sw_map_.end();  ++sw_node_it) {

        CalculateVl2VlMappingnOnSwitch(sw_node_it->second);
    }
    OSM_AR_LOG_RETURN_VOID(m_p_osm_log);
}

//on first cycle set to m_sl2vl_per_op_vls_
void ArKdorAlgorithm::SetSl2VlMappingnOnSwitchEnhancedPort0(
    ARSWDataBaseEntry &sw_db_entry)
{
    OSM_AR_LOG_ENTER(m_p_osm_log);

    osm_node_t *p_osm_node = sw_db_entry.m_general_sw_info.m_p_osm_sw->p_node;
    osm_physp_t *p_physp = osm_node_get_physp_ptr(p_osm_node, 0);
    if (!p_physp) {
        OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
                "Failed to get phys enhanced port 0 of "
                "switch GUID: " U64H_FMT ", LID: %u \n",
                sw_db_entry.m_general_sw_info.m_guid,
                sw_db_entry.m_general_sw_info.m_lid);

        OSM_AR_LOG_RETURN_VOID(m_p_osm_log);
    }

    uint8_t op_vls = ib_port_info_get_op_vls(&p_physp->port_info);

    for (u_int8_t port_num = 1;
         port_num <= p_osm_node->node_info.num_ports;
         port_num++) {

        SetVl2VlMappingn(sw_db_entry, 0, port_num, &m_sl2vl_per_op_vls_[op_vls]);
    }

    OSM_AR_LOG_RETURN_VOID(m_p_osm_log);
}

//Might be need to optimize, might use thread pool
void ArKdorAlgorithm::CalculateVl2VlMappingnOnSwitch(
    ARSWDataBaseEntry &sw_db_entry)
{
    osm_node_t *p_osm_node = sw_db_entry.m_general_sw_info.m_p_osm_sw->p_node;
    if(ib_switch_info_is_enhanced_port0(&p_osm_node->sw->switch_info)) {
        //set port0 only if it is enhanced_port
        //On first cycle m_vl2vl[0][1] == NULL
        if (m_update_sl2vl_ || sw_db_entry.m_osm_update_needed ||
            sw_db_entry.m_kdor_data->m_vl2vl[0][1] == NULL) {
            SetSl2VlMappingnOnSwitchEnhancedPort0(sw_db_entry);
        }

    }

    SwToConnectionMap &connections = sw_db_entry.m_kdor_data->GetConnections();

    for (SwToConnectionMap::iterator from_iter = connections.begin();
         from_iter != connections.end(); ++from_iter) {

        KdorConnection &from_connection = from_iter->second;
        for (SwToConnectionMap::iterator to_iter = connections.begin();
             to_iter != connections.end(); ++to_iter) {

            KdorConnection &to_connection = to_iter->second;
            CalculateVl2VlMappingnOnSwitch(
                sw_db_entry, from_connection, to_connection);

        }
    }
}

void ArKdorAlgorithm::CalculateVl2VlMappingnOnSwitch(
    ARSWDataBaseEntry &sw_db_entry,
    KdorConnection &from_connection,
    KdorConnection &to_connection)
{
    kdor_turn_t turn_type = GetTurnType(from_connection, to_connection);
    uint8_t in_port;
    uint8_t out_port;
    uint8_t op_vls;

    for (PhysPortsList::iterator from_iter = from_connection.m_ports.begin();
         from_iter != from_connection.m_ports.end(); ++from_iter) {

        in_port = (*from_iter)->port_num;

        for (PhysPortsList::iterator to_iter = to_connection.m_ports.begin();
             to_iter != to_connection.m_ports.end(); ++to_iter) {

            out_port = (*to_iter)->port_num;
            if (in_port == out_port) {
                sw_db_entry.m_kdor_data->
                    m_is_to_set_vl2vl[in_port][out_port] = false;
                continue;
            }
            op_vls = ib_port_info_get_op_vls(&(*to_iter)->port_info);

            if (m_p_osm_log->level & OSM_LOG_DEBUG) {
                OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
                        "Calculate Vl2Vl Mappingn on "
                        "switch GUID: " U64H_FMT ", LID: %u "
                        "from port %u to port: %u turn_type: %u\n",
                        sw_db_entry.m_general_sw_info.m_guid,
                        sw_db_entry.m_general_sw_info.m_lid,
                        in_port, out_port, turn_type);

                OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
                           "op_vls: %u turn type %u vl to vl mapping %s\n",
                           op_vls, turn_type,
                           AdaptiveRoutingManager::ConvertSLToVLMappingToStr(
                               &m_turn_type_to_vl2vl_per_op_vls_[turn_type][op_vls]).c_str());

            }
            SetVl2VlMappingn(sw_db_entry, in_port, out_port,
                             &(m_turn_type_to_vl2vl_per_op_vls_[turn_type][op_vls]));
        }
    }
}

void ArKdorAlgorithm::SetVl2VlMappingn(
    ARSWDataBaseEntry &sw_db_entry,
    uint8_t in_port, uint8_t out_port,
    SMP_SLToVLMappingTable *vl2vlmapping)
{
    if (sw_db_entry.m_osm_update_needed ||
        sw_db_entry.m_kdor_data->m_vl2vl[in_port][out_port] != vl2vlmapping) {

        sw_db_entry.m_kdor_data->m_vl2vl[in_port][out_port] = vl2vlmapping;
        sw_db_entry.m_kdor_data->m_is_to_set_vl2vl[in_port][out_port] = true;
    }
}

void ArKdorAlgorithm::SetPlftMap(ARSWDataBaseEntry &sw_db_entry)
{
    uint8_t max_block = sw_db_entry.m_general_sw_info.m_p_osm_sw->p_node->
        node_info.num_ports / PLFT_MAP_PORT_BLOCK_SIZE;

    for (uint8_t port_block = 0; port_block <= max_block; ++port_block) {
        sw_db_entry.m_kdor_data->m_plft_map[port_block] = &m_plft_map_;
    }
    sw_db_entry.m_kdor_data->m_set_plft_map = true;
}


////////////////////////////////////////////////////////////////////////////////

void KdorAlgorithmData::AddLidToKdorGroupData(uint16_t lid_num,
                                              uint16_t sw_lid_num,
                                              KdorGroupData &kdor_group_data,
                                              bool is_new_group,
                                              osm_log_t *p_osm_log)
{
    OSM_AR_LOG(p_osm_log, OSM_LOG_DEBUG,
               "Add lid_num:%u sw_lid_num: %u to KdorGroupData is_new_group: %u\n",
               lid_num, sw_lid_num, is_new_group);

    if (is_new_group) {
        AddLidToKdorGroupData(sw_lid_num, kdor_group_data);
    }

    if (lid_num != sw_lid_num) {
        AddLidToKdorGroupData(lid_num, kdor_group_data);
    }
}

void KdorAlgorithmData::AddLidToKdorGroupData(uint16_t lid_num,
                                              KdorGroupData &kdor_group_data)
{
    static int num_groups = MAX_KDOR_PLFT_NUMBER - 1;

    for (int plft = 0; plft < num_groups; ++plft) {
        //might be NULL incase of group size <= 1
        if (kdor_group_data.group_data_array[plft] != NULL) {
            kdor_group_data.group_data_array[plft]->
                m_lids_list.push_back(lid_num);
        }
    }
    m_lid_to_kdor_group_map.insert(
            make_pair(lid_num, kdor_group_data));
}

////////////////////////////////////////////////////////////////////////////////

// Update connection::direction dimensions index and sign for given
// src and dst switches
// Valid connection direction is connection cange of one and only one dimension.
// If no chane found (same connection for src and dst)
// or more than one dimension changed, return non zero value
int ArHcAlgorithm::SetDirection(
    osm_switch_t* src, osm_switch_t* dst, KdorConnection &connection)
{
    OSM_AR_LOG_ENTER(m_p_osm_log);

    if (src->coord == OSM_SW_NO_COORD ||
        dst->coord == OSM_SW_NO_COORD) {

        connection.m_dim_idx = INVALID_DIM_INDEX;
        connection.m_dim_sign = (dim_sign_t)(1);

        if ( m_p_osm_log->level & OSM_LOG_DEBUG) {
            u_int64_t src_node_guid = ADB2C_BE64_TO_CPU(osm_node_get_node_guid(
                src->p_node));
            u_int16_t src_node_lid = ADB2C_BE16_TO_CPU(osm_node_get_base_lid(
                src->p_node, 0));
            u_int64_t dst_node_guid = ADB2C_BE64_TO_CPU(osm_node_get_node_guid(
                src->p_node));
            u_int16_t dst_node_lid = ADB2C_BE16_TO_CPU(osm_node_get_base_lid(
                src->p_node, 0));

            OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
                       "Invalid coordinate on set direction from "
                       "switch GUID " U64H_FMT ", LID %u coord: 0x%x "
                       "to switch GUID " U64H_FMT ", LID %u coord: 0x%x "
                       "dim_idx: %u  dim_sign: %d\n",
                       src_node_guid, src_node_lid, src->coord,
                       dst_node_guid, dst_node_lid, dst->coord,
                       connection.m_dim_idx, connection.m_dim_sign);
        }
        OSM_AR_LOG_RETURN(m_p_osm_log, 0);
    }

    bool diff_found = false;
    dim_index_t dim_idx = 0;
    uint16_t coord_diff = src->coord ^ dst->coord;

    while (coord_diff) {
        if (coord_diff & 0x1) {
            if (coord_diff == 0x1) {
                diff_found = true;
            }
            break;
        }
        coord_diff = coord_diff >> 1;
        dim_idx++;
    }

    if (!diff_found) {
        u_int64_t src_node_guid = ADB2C_BE64_TO_CPU(osm_node_get_node_guid(
            src->p_node));
        u_int16_t src_node_lid = ADB2C_BE16_TO_CPU(osm_node_get_base_lid(
            src->p_node, 0));
        u_int64_t dst_node_guid = ADB2C_BE64_TO_CPU(osm_node_get_node_guid(
            src->p_node));
        u_int16_t dst_node_lid = ADB2C_BE16_TO_CPU(osm_node_get_base_lid(
            src->p_node, 0));

        OSM_AR_LOG(m_p_osm_log, OSM_LOG_ERROR,
                   "Switch GUID: " U64H_FMT ", LID: %u coord: 0x%x "
                   "is not valid HC member of "
                   "switch GUID: " U64H_FMT ", LID: %u coord: 0x%x\n",
                   src_node_guid, src_node_lid, src->coord,
                   dst_node_guid, dst_node_lid, dst->coord);

        OSM_AR_LOG_RETURN(m_p_osm_log, 1);
    }

    connection.m_dim_idx = dim_idx;
    if (src->coord >> dim_idx & 0x1) {
        connection.m_dim_sign = (dim_sign_t)(-1);
    } else {
        connection.m_dim_sign = (dim_sign_t)(1);
    }

    if ( m_p_osm_log->level & OSM_LOG_DEBUG) {
        u_int64_t src_node_guid = ADB2C_BE64_TO_CPU(osm_node_get_node_guid(
            src->p_node));
        u_int16_t src_node_lid = ADB2C_BE16_TO_CPU(osm_node_get_base_lid(
            src->p_node, 0));
        u_int64_t dst_node_guid = ADB2C_BE64_TO_CPU(osm_node_get_node_guid(
            src->p_node));
        u_int16_t dst_node_lid = ADB2C_BE16_TO_CPU(osm_node_get_base_lid(
            src->p_node, 0));

        OSM_AR_LOG(m_p_osm_log, OSM_LOG_DEBUG,
                   "Direction from switch GUID " U64H_FMT ", LID %u coord: 0x%x"
                   "to switch GUID " U64H_FMT ", LID %u coord: 0x%x "
                   "dim_idx: %u  dim_sign: %d\n",
                   src_node_guid, src_node_lid, src->coord,
                   dst_node_guid, dst_node_lid, dst->coord,
                   connection.m_dim_idx, connection.m_dim_sign);
    }

    OSM_AR_LOG_RETURN(m_p_osm_log, 0);
}

////////////////////////////////////////////////////////////////////////////////

int KdorSwData::Init(ArKdorAlgorithm &algorithm,
                     ARSWDataBaseEntry &db_entry,
                     uint32_t sw_idx,
                     uint32_t switches_number)
{
    OSM_AR_LOG_ENTER(algorithm.GetLog());

    OSM_AR_LOG(algorithm.GetLog(), OSM_LOG_DEBUG,
            "Init KdorSwData of "
            "switch GUID: " U64H_FMT ", LID: %u sw_idx: %u\n",
            db_entry.m_general_sw_info.m_guid,
            db_entry.m_general_sw_info.m_lid,
            sw_idx);

    osm_physp_t* p_physp;
    osm_node_t *p_remote_node;
    osm_switch_t  *p_osm_sw = db_entry.m_general_sw_info.m_p_osm_sw;
    osm_node_t *p_osm_node = p_osm_sw->p_node;
    KdorConnection init_connection;

    m_calculate_port_groups_task_.Init(&db_entry, &algorithm);
    m_sw_idx_ = sw_idx;
    if (switches_number > (uint32_t)m_route_info_.size()) {
        m_route_info_.resize((uint32_t)(switches_number * 1.05));
    }

    //set priv to point to current ARSWDataBaseEntry
    m_saved_priv_ = p_osm_sw->priv;
    p_osm_sw->priv = &(db_entry);

    for (KdorRouteInfoVec::iterator iter = m_route_info_.begin();
         iter != m_route_info_.end(); ++iter) {
        iter->Init();
    }

    //TODO: Optimization reset connection only if one sw connection changed
    m_connections_.clear();
    m_port_num_to_connection_tbl_.assign(m_port_num_to_connection_tbl_.size(),
                                         NULL);
    osm_switch_t  *p_remote_sw;
    //Build connections
    for(u_int8_t port_num = 1 ; port_num <= p_osm_node->node_info.num_ports ;
        port_num++) {
        p_physp = osm_node_get_physp_ptr(p_osm_node, port_num);
        if (!p_physp) {
            continue;
        }
        osm_physp_t *p_remote_physp = osm_physp_get_remote(p_physp);
        if (!p_remote_physp) {
            OSM_AR_LOG(algorithm.GetLog(), OSM_LOG_DEBUG,
                       "Unconnected port: %d\n", port_num);
            continue;
        }
        p_remote_node = osm_physp_get_node_ptr(p_remote_physp);
        if (!p_remote_node) {
            OSM_AR_LOG(algorithm.GetLog(), OSM_LOG_ERROR,
                       "-A- Faile to get remote node connected to "
                       "switch GUID: " U64H_FMT ", LID: %u port: %u\n",
                       db_entry.m_general_sw_info.m_guid,
                       db_entry.m_general_sw_info.m_lid,
                       port_num);
            continue;
        }

        u_int8_t remote_node_type = osm_node_get_type(p_remote_node);
        if (remote_node_type != IB_NODE_TYPE_SWITCH) {
            //Create one connection to all hosts with remote_sw key == NULL
            p_remote_sw = NULL;
        } else {
            //remote is switch
            p_remote_sw = p_remote_node->sw;
        }

        SwToConnectionMapInsertRes insert_res = m_connections_.insert(
            make_pair(p_remote_sw, init_connection));
        KdorConnection &connection = insert_res.first->second;

        if(insert_res.second) { //new connection was inserted

            connection.m_remote_switch = p_remote_sw;
            if (p_remote_sw != NULL &&
                algorithm.SetDirection(p_osm_sw, p_remote_sw, connection)) {
                OSM_AR_LOG_RETURN(algorithm.GetLog(), 1);
            }
        }

        m_port_num_to_connection_tbl_[port_num] = &connection;
        connection.m_ports.push_back(p_physp);
        connection.m_ports_bitset.set(port_num);

        OSM_AR_LOG(algorithm.GetLog(), OSM_LOG_DEBUG,
                "Add port:%u to connection to %p "
                "Total connections number:%u \n",
                port_num, p_remote_sw,
                (int)m_connections_.size());

    }
    OSM_AR_LOG_RETURN(algorithm.GetLog(), 0);
}

void KdorSwData::Resize(uint32_t ports_num) {
    m_port_num_to_connection_tbl_.resize(ports_num, NULL);

    m_vl2vl.resize(ports_num);
    m_is_to_set_vl2vl.resize(ports_num);

    for (uint32_t port = 0; port < ports_num; ++port) {
        m_vl2vl[port].resize(ports_num, NULL);
        m_is_to_set_vl2vl[port].resize(ports_num, false);
    }
}

string KdorConnection::ToString() const
{
    std::stringstream stream;

    if (m_remote_switch == NULL) {

        stream << "Connection to " << (int) m_ports.size() <<
            " locat CAs";
    } else {

        uint16_t next_hop_lid = ADB2C_BE16_TO_CPU(osm_node_get_base_lid(
                m_remote_switch->p_node, 0));

        stream << "Connection to " << (int)m_ports.size() <<
            " on dim_index " << (int)m_dim_idx <<
            " dim_sign " << (int)m_dim_sign <<
            " links to remote switch LID: " << next_hop_lid;
    }

    return stream.str();
}

//Select turn type using the following priority: T0, T2, T1
uint8_t KdorRouteInfo::turn_type_wait[KDOR_TURN_TYPE_LAST] = {
    0, //T0
    2, //T1
    1};//T2

bool KdorRouteInfo::operator<(const KdorRouteInfo& __rhs) const
{
    if ((__rhs.m_connection->m_dim_idx == INVALID_DIM_INDEX ||
         m_connection->m_dim_idx == INVALID_DIM_INDEX) &&
        m_connection->m_dim_idx != __rhs.m_connection->m_dim_idx) {
        //prefer destination with valid coordinate
        return m_connection->m_dim_idx != INVALID_DIM_INDEX;
    }
    if (m_vl_inc != __rhs.m_vl_inc) {
        return m_vl_inc < __rhs.m_vl_inc;
    }
    if (m_turn_type != __rhs.m_turn_type) {
        //Select by order T0, T2, T1
        return turn_type_wait[m_turn_type] < turn_type_wait[__rhs.m_turn_type];
    }
    // Among top candidate choose one laying in the maximum dimension
    if (m_connection->m_dim_idx != __rhs.m_connection->m_dim_idx) {
        return m_connection->m_dim_idx > __rhs.m_connection->m_dim_idx;
    }
    if (m_connection->m_dim_sign != __rhs.m_connection->m_dim_sign) {
        if (m_turn_type == KDOR_TURN_TYPE_2) {
            //This will guarantee that prev hop will not have KDOR_TURN_TYPE_1
            //and there for will not increase vl
            return m_connection->m_dim_sign < __rhs.m_connection->m_dim_sign;
        } else {
            //This will guarantee that prev hop will not have KDOR_TURN_TYPE_2
            //and there for will not increase vl
            return m_connection->m_dim_sign > __rhs.m_connection->m_dim_sign;
        }
    }
    return false;
}

