/*                  - Mellanox Confidential and Proprietary -
 *
 *  Copyright (C) 2010-2011, Mellanox Technologies Ltd.  ALL RIGHTS RESERVED.
 *
 *  Except as specifically permitted herein, no portion of the information,
 *  including but not limited to object code and source code, may be reproduced,
 *  modified, distributed, republished or otherwise exploited in any form or by
 *  any means for any purpose without the prior written permission of Mellanox
 *  Technologies Ltd. Use of software subject to the terms and conditions
 *  detailed in the file "COPYING".
 *
 */


#ifndef AR_MGR_H_
#define AR_MGR_H_

/* osm_config.h needs to be before all other includes */
#include <opensm/osm_config.h>
#include <typeinfo>
#include <string>
#include <stack>
using namespace std;

#include <opensm/osm_opensm.h>
#include <opensm/osm_helper.h>
#include <opensm/osm_log.h>
#include <opensm/osm_event_plugin.h>
#include <vendor/osm_vendor_api.h>

#include "ar_mgr_types.h"
#include "ar_clbck.h"
#include "parallel_port_groups_calculator.h"
#include "osm_plugins_common.h"

typedef vector <bool> BoolVec;
typedef map <u_int64_t, u_int16_t> GuidToGroupMap;
typedef GuidToGroupMap::iterator GuidToGroupMapIter;

class AdaptiveRoutingAlgorithm;
class PlftBasedArAlgorithm;
class ArKdorAlgorithm;

/*#include "fit_api.h"
#include "osm_api.h"*/

struct AnalizeDFSetupData {
    stack <ARSWDataBaseEntry*> m_group_stack;
    SwDbEntryPrtList m_spines;
    SwDbEntryPrtList m_leafs;

    u_int16_t hca_to_sw_lid_mapping[IB_LID_UCAST_END_HO + 1];
    DfSwData* sw_lid_to_df_data_mapping[IB_LID_UCAST_END_HO + 1];

    AnalizeDFSetupData() {
        memset(hca_to_sw_lid_mapping, 0, sizeof(hca_to_sw_lid_mapping));
        memset(sw_lid_to_df_data_mapping, 0, sizeof(sw_lid_to_df_data_mapping));
    }
};



//============================================================================//
class AdaptiveRoutingManager {
public:
    ////////////////////
    //methods
    ////////////////////

    /**
     * Constructor AdaptiveRoutingManager
     *
     * @param p_osm pointer to OpenSM structure
     */
    AdaptiveRoutingManager(osm_opensm_t *p_osm);

    /**
     *  Destructor AdaptiveRoutingManager
     */
    ~AdaptiveRoutingManager();

    /**
     * Set osm_event_id
     *
     * @param osm_event_id
     */
    void SetOSMEventId(osm_epi_event_id_t osm_event_id);

    /**
     * Update fabric nodes. Update node info for all nodes
     *
     * @return int 0 if succeeded, otherwise 1
     */
    void UpdateFabricSwitches();

    /**
     * Configure all nodes the support adaptive routing in the
     * fabric with adaptove routing parameters with user defined
     * parameters or default if none where given.
     *
     * @return int return 1 is succeeded, otherwise 0
     */
    int Run();

    /**
     * Calculate adaptive routing settings according to fat tree algorithm 
     * for the given sw_db_entry 
     */
    void ARCalculateSwitchPortGroupsTree(
        ARSWDataBaseEntry &sw_db_entry,
        uint8_t max_rank,
        u_int16_t *hca_to_sw_lid_mapping,
        uint8_t *sw_lid_to_rank_mapping);

    /**
     * Dump the switch adaptive routing settings
     *
     * @param sw_db_entry the entry of the switch in the db
     */
    void ARDumpSWSettings(ARSWDataBaseEntry &sw_db_entry);

    /**
     * Print Exception
     *
     * @param e exception
     * @param p_osm_log log file
     */
    static void printException(exception& e, osm_log_t * p_osm_log) {
        string e_what(e.what( ));
        string e_type(typeid(e).name( ));
        OSM_LOG(p_osm_log, OSM_LOG_ERROR,
                "AR_MGR - caught an exception: %s. Type: %s\n",
                e_what.c_str(), e_type.c_str());
    }

    /**
     * Decide if two adaptive_routing_info structs are equal or not
     *
     * @param p_ar_info1 struct one to compare
     * @param p_ar_info2 struct two to compare
     * @param ignoreE if true, do not compare E attribute
     *
     * @return true if equal or false otherwise
     */
    static bool IsEqualSMPARInfo(adaptive_routing_info *p_ar_info1,
            adaptive_routing_info *p_ar_info2, bool ignoreE, bool ignoreTop);

    /**
     * Convert SL to VL mapping table into string
     *
     * @param p_slvl_mapping
     *
     * @return the result string
     */
    static string ConvertSLToVLMappingToStr(SMP_SLToVLMappingTable *p_slvl_mapping);

private:
    ////////////////////
    //members
    /////////////////////
    Ibis                            m_ibis_obj;             /** IBIS object, used to send mads */
    u_int64_t                       m_port_guid;            /** Port guid to bind for */
    osm_epi_event_id_t              m_osm_event_id;         /** OSM event id */
    osm_opensm_t                    *m_p_osm;               /** Pointer to osm opensm struct */
    osm_vendor_t                    *m_p_osm_vendor;        /** Pointer to osm vendor struct */
    osm_subn_t                      *m_p_osm_subn;          /** Pointer to osm subnet struct */
    osm_log_t                       *m_p_osm_log;           /** Pointer to osm log */

    ThreadPool                      m_thread_pool;
    ParallelPortGroupsCalculator    m_port_groups_calculator;

    struct ARSWDataBase             m_sw_db;                /** Switchs database */
    struct ARCADataBase             m_ca_db;                /** CAs database */
    struct MasterDataBase           m_master_db;            /** Master settings database */
    struct AdaptiveRoutingClbck     m_ar_clbck;             /** Ibis clbck methods manager */

    string                          m_conf_file_name;       /** Name of user input file */

    int                             m_oldest_error;         /** index of oldest error in error window array */
    unsigned int                    m_num_errors;           /** Number of errors */
    struct timeval                  *m_p_error_window_arr;  /** Error window array */

    bool                            m_is_permanent_error;   /** Indicates permanent error occurred
                                                                                 * (will not be solved after heavy sweep) */
    bool                            m_is_temporary_error;   /** Indicates temporary error occurred
                                                                                 * (will be solved after heavy sweep) */

    bool                            m_df_configured;        /** If true should remove df configuration
                                                             * if protocol is not DF PLUS*/
    AdaptiveRoutingAlgorithm       *m_ar_algorithm;         //!!! TAMIR add to OSM should replace m_df_configured

    bool                            m_sw_info_configured;  /** If true should remove sw info configuration
                                                             * if protocol is not DF PLUS*/

    uint16_t                        m_max_df_group_number;
    GuidToGroupMap                  m_guid_to_dfp_group;
    bool                            m_group_discovered;

    uint32_t                        m_options_file_crc;

    OsmMKeyManager                  m_mkey_manager;

    ////////////////////
    //methods
    ////////////////////
    /**
     * Reset error window mechanism. Remove old errors
     */
    void ResetErrorWindow();

    /**
     * Set default configuration values
     */
    void SetDefaultConfParams();

    /**
     * Take the configuration values that have been parsed
     */
    void TakeParsedConfParams();

    /**
     * Check if file exists and have permissions
     *
     * @return true if exists, otherwise false
     */
    bool IsFileExists(const char *file_name);

    /**
     * Merge between default settings and user option settings
     */
    void UpdateUserOptions();

    /**
     * Check if more than m_max_errors occurred during the
     *  last m_error_window. If so, throw an exception
     *
     * @throw Throws int when condition holds
     *
     * @param rc return value that indicates if an error has occurred
     */
    void CheckRC(int& rc);

    /**
     * Parse OSM parameters for cc_mgr conf file
     *
     * @param osm_params
     */
    void ParseConfFileName(char* osm_plugin_options);

    /**
     * Initialize objects.
     * Bind port to AR mads
     *
     * @return int 0 if succeeded, otherwise throw exception with int 1
     */
    int Init();

    /**
     * Check if device id is supported.
     *
     * @param node_info node info holds device id
     *
     * @return bool true if supported, otherwise false
     */
    bool IsDeviceIDSupported(const ARGeneralSWInfo& general_sw_info);

    /**
     * Update Switch from sm database to local database
     *
     * @param node_info node info
     *
     * @return int 0 if succeeded, otherwise 1
     */
    void UpdateSW(const ARGeneralSWInfo& general_sw_info);

    /**
     * Add switches that don't exist ar_mgr from osm database
     *  and update all switches in ar_mgr

     */
    void AddNewAndUpdateExistSwitches();

    /**
     * Remove switches that don't exist in m_p_osm_subn
     *  from ar_mgr database
     */
    void RemoveAbsentSwitches();

    /**
     * Mark switch as not supports for feature for all future cycles
     *
     * @param sw_db_entry the entry of the switch in the db
     * @param feature [AR|DF|RN]
     * @param type the reason that this switch is marked as not supports
     */
    //void MarkSWNotSupport(ARSWDataBaseEntry &sw_db_entry,
    //                      supported_feature_t feature,
    //                      support_errs_t type);
    /**
     * Check if switch supports for feature
     *
     * @param sw_db_entry the entry of the switch in the db
     * @param feature [AR|DF|RN]
     * @return true if supported or false otherwise
     */
    //bool IsNotSupported(ARSWDataBaseEntry &sw_db_entry,
    //                    supported_feature_t feature);

    /**
     * Mark switch as not supports for adaptive routing for all future cycles
     *
     * @param sw_db_entry the entry of the switch in the db
     * @param type the reason that this switch is marked as not supports
     */
    void MarkSWNotSupportAR(ARSWDataBaseEntry &sw_db_entry,
            support_errs_t type);

    /**
     * Check if switch supports for adaptive routing
     *
     * @param sw_db_entry the entry of the switch in the db
     *
     * @return true if supported or false otherwise
     */
    bool IsARNotSupported(ARSWDataBaseEntry &sw_db_entry);

    /**
     * Mark switch as supports for adaptive routing
     *
     * @param sw_db_entry the entry of the switch in the db
     */
    void MarkSWSupportAR(ARSWDataBaseEntry &sw_db_entry);

    /**
     * Check if adaptive routing is active in a switch (both support and enable)
     *
     * @param sw_db_entry the entry of the switch in the db
     *
     * @return true if active or false otherwise
     */
    bool IsARActive(ARSWDataBaseEntry &sw_db_entry);

    /**
     * Check if Dragonfly is active in a switch (both support and AR
     * enable)
     *
     * @param sw_db_entry the entry of the switch in the db
     *
     * @return true if active or false otherwise
     */
    bool IsDFActive(ARSWDataBaseEntry &sw_db_entry);

    /**
     * Mark all switch as supports / not supports adaptive routing
     *
     * @return number of unsupportes switches
     *
     */
    int ARInfoGetProcess();

    /**
     * Set Required ARInfo R/W values for all supportes switches
     *
     */
    void SetRequiredARInfo(ARSWDataBaseEntry &db_entry);
    /**
     * Get calculated GroupCap from all supportes switches
     *
     * @return number of unsupportes switches
     *
     */
    int  ARInfoGetGroupCapProcess();

    /**
     *  Configure ARInfo on needed ones
     *
     * @return number of unsupportes switches
     *
     */
    int ARInfoSetProcess();

    /**
     *  Configure RoutingNotification tree settings on needed
     *  switches
     *
     *
     */
    void TreeRoutingNotificationProcess();

    /**
     * Update DB used to copy a single group to moltiple
     * groups
     *
     * @param sw_db_entry the entry of the switch in the db
     * @param assign_groups the calculated GroupsList object
     */
    void ARUpdateSWGroupToCopy(
        ARSWDataBaseEntry &sw_db_entry,
        GroupsList &assign_groups);

    /**
     * Update update a single GroupElement in the switch group table
     *
     * @param sw_db_entry the entry of the switch in the db
     * @param calculated_element the calculated group element
     * @param group_number the calculated group number
     */
    void ARUpdateSWGroupElement(
        ARSWDataBaseEntry &sw_db_entry,
        PortGroup_Block_Element &calculated_element,
        u_int16_t group_number);

    /**
     * Update the switch group table with the calculated one
     *
     * @param sw_db_entry the entry of the switch in the db
     * @param p_ar_calculated_group_table the calculated group table
     * @param calculated_groups_number total groups number in the calculated table
     */
    void ARUpdateSWGroupTable(ARSWDataBaseEntry &sw_db_entry,
            struct SMP_ARGroupTable p_ar_calculated_group_table[],
            u_int16_t calculated_groups_number);

    /**
     * Update the switch lft table with the calculated one
     *
     * @param sw_db_entry the entry of the switch in the db
     * @param p_ar_calculated_lft_table the calculated lft table
     * @param calculated_max_lid the last lid that is valid in the calculated table
     */
    void ARUpdateSWLFTTable(ARSWDataBaseEntry &sw_db_entry,
            struct SMP_ARLinearForwardingTable p_ar_calculated_lft_table[],
            u_int16_t calculated_max_lid);

    /**
     * Update the SX switch lft table with the calculated one
     *
     * @param sw_db_entry the entry of the switch in the db
     * @param p_ar_calculated_lft_table the calculated lft table
     * @param calculated_max_lid the last lid that is valid in the calculated table
     */
    void ARUpdateSWLFTTable(ARSWDataBaseEntry &sw_db_entry,
            struct SMP_ARLinearForwardingTable_SX p_ar_calculated_lft_table[],
            u_int16_t calculated_max_lid);

    void UpdateRNRcvString(ARSWDataBaseEntry &sw_db_entry,
                           uint8_t max_rank,
                           uint8_t sw_rank,
                           uint8_t max_consume_rank);

    void UpdateRNXmitPortMask(
            ARSWDataBaseEntry &sw_db_entry,
            PortsBitset &ca_ports,
            PortsBitset &sw_ports,
            bool is_down_sw);

    /**
     * Calculate adaptive routing settings according to parallel links algorithm
     */
    void ARCalculatePortGroupsParallelLinks();

    /**
     * Calculate adaptive routing settings according to fat tree algorithm
     */
    void ARCalculatePortGroupsTree();

    /**
     * Configure Vlids Lfts on all switches
     * Used in case ucast cache is valid and only vlid changed
     */
    void UpdateVlidsLfts();

    /**
     * Calculate Vlids Lfts on given switch
     * Used in case ucast cache is valid and only vlid changed
     */
    void CalculateVlidsLft(LidsToLidsVecVec &sw_lid_to_vlids,
                           osm_switch_t  *p_osm_sw,
                           SMP_AR_LFT &ar_lft,
                           bool to_set_lft_table[]);

    /**
     * Map sw lid to all its connected vlids
     */
    void BuildSwToVlidsMap(LidsToLidsVecVec &sw_lid_to_vlids);

    /**
     * Update data structures on adding new lid to AR groups
     */
    void AddLidToARGroup(uint16_t lid_num, uint16_t sw_lid_num,
                         GroupData* p_group_data,
                         TreeAlgorithmData &algorithm_data,
                         bool is_new_group,
                         bool is_group_per_leaf_sw);

    /**
     * Print the given group_data object to log file
     */
    void PrintGroupData(const char * str, GroupData &group_data);

    /**
     * map all lids in the given lids_list to p_group in the given lid_to_group_map map
     */
    //int MapLidsToGroup(LidToGroupDataMap  &lid_to_group_map, 
    //                   list<uint16_t> &lids_list,
    //                   GroupData* p_group);

    /**
     * Assign group number and table number to discovered groups
     */
    int AssignPortGroups(ARSWDataBaseEntry &sw_db_entry, TreeAlgorithmData &algorithm_data);

    int AssignPerLeafSwitchGroups(ARSWDataBaseEntry &sw_db_entry,
                                  TreeAlgorithmData &algorithm_data);

    int AssignToContainedGroup(ARSWDataBaseEntry &sw_db_entry, TreeAlgorithmData &algorithm_data, GroupData* group);
    int GetContainedGroupList(GroupsList& group_list, PortsBitset &group_bitmask, GroupsList& contained_group_list, PortsBitset &total_bitmask);

    /**
     * Get GroupList ordered by group size ad lids number
     */
    int GetOrderedGroupList(TreeAlgorithmData &algorithm_data, GroupsList& group_list);

    /**
     * Update all needed switches with the new calculated group table
     */
    void ARGroupTableProcess();

    void ARGroupTableProcess(ARSWDataBaseEntry &sw_entry,
                             u_int16_t group_top,
                             bool reset_direction,
                             SMP_ARGroupTable ar_group_table_array[AR_GROUP_TABLE_NUM_BLOCKS]);

    void ARCopyGroupTableProcess();

    /**
     * Update all needed switches with the new calculated lft table
     */
    void ARLFTTableProcess();

    void ARLFTTableProcess(ARSWDataBaseEntry &sw_entry,
                           u_int16_t max_lid,
                           u_int8_t  pLFTID,
                           bool to_set_lft_table[AR_LFT_TABLE_NUM_BLOCKS_SX],
                           SMP_AR_LFT &ar_lft);

    /**
     * Run an adaptive routing cycle on all switches
     */
    int ARCycle();

    /**
     * Mark all switch flags for osm
     *  Some switches were configured with adaptive routing and some need to
     *  be configured with osm with ucast lft table
     */
    void AROSMIntegrationProcess();

    /**
     * Send ARGroupTable MAD SET()/GET() by lid
     *
     * @param lid destination lid
     * @param method method
     * @param group_block group block [0,127] reduce by the bumber of group supported by the device
     * @param group_table group table [0,15] reduce by the bumber of group supported by the device
     * @param p_ar_group_table the MAD info to be sent
     *
     * @return mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
     *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
     */
    int ARGroupTableMadGetSetByLid(u_int16_t lid,
            u_int8_t  method,
            u_int16_t group_block,
            u_int8_t  group_table,
            u_int8_t  pLFTID,
            u_int8_t  reset_direction_attr,
            struct SMP_ARGroupTable *p_ar_group_table,
            const clbck_data_t *p_clbck_data = NULL);

    /**
     * Send ARGroupTable MAD SET()/GET() by direct
     *
     * @param p_direct_route direct route
     * @param method method
     * @param group_block group block [0,4095] reduce by the bumber of group supported by the device
     * @param group_table group table [0,15] reduce by the bumber of group supported by the device
     * @param pLFTID private LFTID [0,255] reduce by the bumber of private LFTs supported by the device
     * @param p_ar_group_table the MAD info to be sent
     *
     * @return mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
     *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
     */
    int ARGroupTableMadGetSetByDirect(direct_route_t *p_direct_route,
            u_int8_t  method,
            u_int16_t group_block,
            u_int8_t  group_table,
            u_int8_t  pLFTID,
            u_int8_t  reset_direction_attr,
            struct SMP_ARGroupTable *p_ar_group_table,
            const clbck_data_t *p_clbck_data = NULL);

    /**
     * Send ARLinearForwardingTable MAD SET()/GET() by lid
     *
     * @param lid destination lid
     * @param method method
     * @param lid_block lid block [0,1535] reduce by the size of linear FDB table of the SW
     * @param pLFTID private LFTID [0,255] reduce by the bumber of private LFTs supported by the device
     * @param p_ar_group_table the MAD info to be sent
     *
     * @return mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
     *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
     */
    int ARLinearForwardingTableMadGetSetByLid(u_int16_t lid,
            u_int8_t method,
            u_int32_t lid_block,
            u_int8_t pLFTID,
            struct SMP_ARLinearForwardingTable *p_ar_linear_forwarding_table,
            const clbck_data_t *p_clbck_data = NULL);

    /**
     * Send ARLinearForwardingTable MAD SET()/GET() by lid (for SX device)
     *
     * @param lid destination lid
     * @param method method
     * @param lid_block lid block [0,1535] reduce by the size of linear FDB table of the SW
     * @param pLFTID private LFTID [0,255] reduce by the bumber of private LFTs supported by the device
     * @param p_ar_group_table the MAD info to be sent
     *
     * @return mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
     *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
     */
    int ARLinearForwardingTableMadGetSetByLid(u_int16_t lid,
            u_int8_t method,
            u_int32_t lid_block,
            u_int8_t pLFTID,
            struct SMP_ARLinearForwardingTable_SX *p_ar_linear_forwarding_table,
            const clbck_data_t *p_clbck_data = NULL);

    /**
     * Send ARLinearForwardingTable MAD SET()/GET() by direct
     *
     * @param p_direct_route direct route
     * @param method method
     * @param lid_block lid block [0,1535] reduce by the size of linear FDB table of the SW
     * @param pLFTID private LFTID [0,255] reduce by the bumber of private LFTs supported by the device
     * @param p_ar_group_table the MAD info to be sent
     *
     * @return mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
     *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
     */
    int ARLinearForwardingTableMadGetSetByDirect(direct_route_t *p_direct_route,
            u_int8_t method,
            u_int32_t lid_block,
            u_int8_t pLFTID,
            struct SMP_ARLinearForwardingTable *p_ar_linear_forwarding_table,
            const clbck_data_t *p_clbck_data = NULL);

    /**
     * Send ARLinearForwardingTable MAD SET()/GET() by direct (for SX device)
     *
     * @param p_direct_route direct route
     * @param method method
     * @param lid_block lid block [0,1535] reduce by the size of linear FDB table of the SW
     * @param pLFTID private LFTID [0,255] reduce by the bumber of private LFTs supported by the device
     * @param p_ar_group_table the MAD info to be sent
     *
     * @return mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
     *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
     */
    int ARLinearForwardingTableMadGetSetByDirect(direct_route_t *p_direct_route,
            u_int8_t method,
            u_int32_t lid_block,
            u_int8_t pLFTID,
            struct SMP_ARLinearForwardingTable_SX *p_ar_linear_forwarding_table,
            const clbck_data_t *p_clbck_data = NULL);

    /**
     * Send PLFTInfoMadGetSetByDirect MAD SET()/GET() by direct
     *
     * @param p_direct_route direct route
     * @param method method
     * @param p_plft_info the MAD info to be sent
     * @param p_clbck_data calback function to handle reply
     *
     * @return mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
     *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
     */
    int PLFTInfoMadGetSetByDirect(direct_route_t *p_direct_route,
            u_int8_t method,
            struct SMP_PrivateLFTInfo *p_plft_info,
            const clbck_data_t *p_clbck_data = NULL);

    /**
     * Send PLFTDefMadGetSetByDirect MAD SET()/GET() by direct
     *
     * @param p_direct_route direct route
     * @param method method
     * @param plft block of 16 PLFTs
     * @param p_plft_def the MAD info to be sent
     * @param p_clbck_data calback function to handle reply
     *
     * @return mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
     *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
     */
    int PLFTDefMadGetSetByDirect(direct_route_t *p_direct_route,
            u_int8_t method,
            u_int8_t plft_block,
            struct SMP_PrivateLFTDef *p_plft_def,
            const clbck_data_t *p_clbck_data = NULL);

    /**
     * Send PLFTMapMadGetSetByDirect MAD SET()/GET() by direct
     *
     * @param p_direct_route direct route
     * @param method method
     * @param pLFTID private LFTID [0,255] reduce by the bumber of private LFTs supported by the device
     * @param p_plft_map the MAD info to be sent
     * @param p_clbck_data calback function to handle reply
     *
     * @return mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
     *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
     */
    int PLFTMapMadGetSetByDirect(direct_route_t *p_direct_route,
            u_int8_t method,
            u_int8_t pLFTID,
            struct SMP_PrivateLFTMap *p_plft_map,
            const clbck_data_t *p_clbck_data);

    /**
     * Send PLFTMapMadGetSetByDirect MAD SET()/GET() by direct
     *
     * @param p_direct_route direct route
     * @param method method
     * @param port_block [0-63] determine a PortBlock of size 4 ports
     * @param p_plft_map the MAD info to be sent
     * @param p_clbck_data calback function to handle reply
     *
     * @return mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
     *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
     */
    int PortSLToPrivateLFTMapGetSetByDirect(direct_route_t *p_direct_route,
            u_int8_t method,
            u_int8_t port_block,
            struct SMP_PortSLToPrivateLFTMap *p_plft_map,
            const clbck_data_t *p_clbck_data = NULL);

    /**
     * Send ExtendedSwitchInfoMadGetSetByDirect MAD SET()/GET() by direct
     *
     * @param p_direct_route direct route
     * @param method method
     * @param p_ext_sw_info the MAD info to be sent
     * @param p_clbck_data calback function to handle reply
     *
     * @return mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
     *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
     */
    int ExtendedSwitchInfoMadGetSetByDirect(direct_route_t *p_direct_route,
            u_int8_t method,
            struct SMP_ExtSWInfo *p_ext_sw_info,
            const clbck_data_t *p_clbck_data = NULL);

    /**
     * Send SMPSLToVLMappingTableGetSetByDirect MAD SET()/GET() by direct
     *
     * @param p_direct_route direct route
     * @param method method
     * @param p_ext_sw_info the MAD info to be sent
     * @param p_clbck_data calback function to handle reply
     *
     * @return mad status[bits: 0-7] / IBIS_MAD_STATUS_SEND_FAILED / IBIS_MAD_STATUS_RECV_FAILED /
     *  IBIS_MAD_STATUS_TIMEOUT / IBIS_MAD_STATUS_GENERAL_ERR
     */
    int SMPSLToVLMappingTableGetSetByDirect(direct_route_t *p_direct_route,
        u_int8_t method,
        phys_port_t out_port_number,
        phys_port_t in_port_number,
        struct SMP_SLToVLMappingTable *p_slvl_mapping,
        const clbck_data_t *p_clbck_data);

    void ConvertARInfoToStr(ARGeneralSWInfo &general_sw_info,
                            adaptive_routing_info &ar_info,
                            char *buff);

    /**
     * Convert group table into string
     *
     * @param p_ar_group_table block of group table
     * @param block_num block number
     *
     * @return the result string
     */
    string ConvertARGroupTableBlockToStr(
        struct SMP_ARGroupTable *p_ar_group_table,
        adaptive_routing_info &ar_info,
        unsigned block_num);

    /**
     * Convert ldt table into string
     *
     * @param p_ar_lft_table block of lft table
     * @param block_num block number
     *
     * @return the result string
     */
    string ConvertARLFTTableBlockToStr(struct SMP_ARLinearForwardingTable *p_ar_lft_table,
            unsigned block_num);

    /**
     * Convert ldt table into string (for SX device)
     *
     * @param p_ar_lft_table block of lft table (for SX device)
     * @param block_num block number
     *
     * @return the result string
     */
    string ConvertARLFTTableBlockToStr(struct SMP_ARLinearForwardingTable_SX *p_ar_lft_table,
            unsigned block_num);

    /**
     * Decide if two SMP_ARGroupTable structs are equal or not
     *
     * @param p_ar_group_table_block1 struct one to compare
     * @param p_ar_group_table_block2 struct two to compare
     *
     * @return true if equal or false otherwise
     */
    bool IsEqualSMPARGroupTableBlock(SMP_ARGroupTable *p_ar_group_table_block1,
            SMP_ARGroupTable *p_ar_group_table_block2);

    /**
     * Decide if two SMP_ARLinearForwardingTable structs are equal or not
     *
     * @param p_ar_lft_table_block1 struct one to compare
     * @param p_ar_lft_table_block2 struct two to compare
     *
     * @return true if equal or false otherwise
     */
    bool IsEqualSMPARLftTableBlock(SMP_ARLinearForwardingTable *p_ar_lft_table_block1,
            SMP_ARLinearForwardingTable *p_ar_lft_table_block2);

    /**
     * Decide if two SMP_ARLinearForwardingTable_SX structs are equal or not
     *
     * @param p_ar_lft_table_block1 struct one to compare
     * @param p_ar_lft_table_block2 struct two to compare
     *
     * @return true if equal or false otherwise
     */
    bool IsEqualSMPARLftTableBlock(SMP_ARLinearForwardingTable_SX *p_ar_lft_table_block1,
            SMP_ARLinearForwardingTable_SX *p_ar_lft_table_block2);

    /**
     * Fill lft block with default values
     *  OSM_NO_PATH in port and static in state
     *
     * @param p_block the lft block to be set
     */
    void ARLftTableBlockSetDefault(SMP_ARLinearForwardingTable *p_block);
    void ARLftTableBlockSetDefault(SMP_ARLinearForwardingTable_SX *p_block);

    /**
     * Fill lft table with default values (all blocks)
     *
     * @param p_lft the lft table to be set
     */
    void ARLftTableSetDefault(SMP_ARLinearForwardingTable p_lft[]);
    void ARLftTableSetDefault(SMP_ARLinearForwardingTable_SX p_lft[]);

    /**
     * Fill group table with default values (all blocks)
     *  All is 0 ==> no groups
     *
     * @param p_group the group table to be set
     */
    void ARGroupTableSetDefault(SMP_ARGroupTable p_group[]);

    /**
     * Map CAs lids to connected SW lid
     *
     * @param hca_to_sw_lid_mapping
     */
     int SetHcaToSwLidMapping(osm_physp_t *p_hca_physp,
                             osm_node_t *p_remote_sw_node,
                             u_int16_t *hca_to_sw_lid_mapping );

     /**
      * Get vlid list on the given hca_physp
      *
      * @param OUT vlids_list
      */
     void GetVlidsList(osm_physp_t *p_hca_physp, LidsList &vlids_list);

    //void AllocateArGroupPerSw(u_int16_t *hca_to_sw_lid_mapping);

    //The allocation of group number to leaf sw is saved
    //after the first allocation
    u_int16_t AllocateSwArGroup(uint16_t sw_lid, uint16_t group_cap);

    /** ----------  DRAGON FLY PLUS ---------- */

    /**
     * Run DragonFly routing cycle on all switches
     *
     * @return non zero on failure
     */
    int ARDragonFlyCycle();

    void ARDragonFlyCycleEnd(int rc);

    void ClearAllDragonflyConfiguration();

    /**
     * Init DragonFlyPlus Data structures on cycle begining
     *
     * @return non zero on failure
     */
    int InitDragonFlyPlus(AnalizeDFSetupData &setup_data);

    /**
     * Set SUPPORT_DF bit for all switches on fabric that are
     * DragonFlyPlusCapable
     *
     * @return true if the fabric is DragonFlyPlusCapable
     */
    bool SetDragonFlyPlusCapable();

    /**
     * Helper function for AnalizeDragonFlyPlusSetup
     *
     * @return non zero on failure
     */
    int MarkLeafsByCasNumber(AnalizeDFSetupData &setup_data,
                             SwDbEntryPrtList &leafs_list);

    int MarkLeafsByGroupsNumber(AnalizeDFSetupData &setup_data,
                                SwDbEntryPrtList &leafs_list);

    void SetGroupNumber(ARSWDataBaseEntry* p_sw_entry, u_int16_t group_number);

    int SetPrevGroupNumber(ARSWDataBaseEntry* p_sw_entry,
                           BoolVec &used_group_numbers);

    int DiscoverGroups(AnalizeDFSetupData &setup_data,
                       SwDbEntryPrtList &leafs_list,
                       BoolVec &used_group_numbers,
                       int cycle = 1);


    /**
     * Helper function for AnalizeDragonFlyPlusSetup update
     * m_up_ports and m_down_ports data structures
     *
     * @return non zero on failure
     */
    int SetPortsDirection();

    /**
     * Update SM osm_switch_t struct rank and coord values
     *
     */
    void UpdateSmDbSwInfo();

    /**
     * Helper function for AnalizeDragonFlyPlusSetup
     *
     * @return non zero on failure
     */
    int DiscoverLeaf(
        AnalizeDFSetupData &setup_data,
        ARSWDataBaseEntry* p_db_entry);

    /**
     * Helper function for AnalizeDragonFlyPlusSetup
     *
     * @return non zero on failure
     */
    int DiscoverSpine(
        AnalizeDFSetupData &setup_data,
        ARSWDataBaseEntry* p_db_entry);

    /**
     * Helper function for AnalizeDragonFlyPlusSetup
     *
     * @return non zero on failure
     */
    int SetLeaf(
        AnalizeDFSetupData &setup_data,
        SwDbEntryPrtList &leafs_list,
        osm_node_t *p_nod);

    /*
    int SetLeaf(
        AnalizeDFSetupData &setup_data,
        osm_node_t *p_node,
        osm_physp_t *p_physp,
        bool from_down);
        */

    /**
     * Helper function for AnalizeDragonFlyPlusSetup
     *
     * @return non zero on failure
     */
    int SetSpine(
        AnalizeDFSetupData &setup_data,
        osm_node_t *p_nod);

    /*
    int SetSpine(
        AnalizeDFSetupData &setup_data,
        osm_node_t *p_node,
        osm_physp_t *p_physp,
        bool from_down);
        */

    /**
     * Helper function for AnalizeDragonFlyPlusSetup
     *
     * @return non zero on failure
     */
    //int HandleLeafsWithoutHosts(AnalizeDFSetupData &setup_data,
    //                            int &max_group_number);

    /**
     * Analize Dragon Fly Plus Setup
     * Set SW type to LEAF or SPINE
     * Set ports as up or down ports
     * @param [out] setup_data
     *
     * @return non zero on failure
     */
    int AnalizeDragonFlyPlusSetup(AnalizeDFSetupData &setup_data);

    /** Helper method for ARCalculatePortGroupsDF*/
    void SavePortGroupsAndDump();

    void ARCalculatePortGroupsDFCleanup();

    int ARCalculatePLFTPortGroups(
        PathDescription paths_description[DLID_LOCATION_NUMBER][PATHS_NUMBER],
        bool enforce_complete_route,
        list <ARSWDataBaseEntry*> &switches,
        int plft_id,
        AnalizeDFSetupData &setup_data);

    void PrintPSGroupData(const char * str,PSGroupData &group_data);

    bool IsRemoteSupportsDF(ARSWDataBaseEntry &sw_db_entry, u_int8_t port_num);

    bool IsRouteOnRemote(ARSWDataBaseEntry &sw_db_entry,
            int plft_id, u_int8_t port_num, uint16_t lid_num);

    bool IsTrueHopsOnRemote(ARSWDataBaseEntry &sw_db_entry,
            u_int8_t port_num, uint16_t lid_num, int hops_num);

    void ClearNoValidDFRouteToLid(ARSWDataBaseEntry &sw_db_entry,
            int plft_id)
    {
        sw_db_entry.m_p_df_data->m_plft[plft_id].m_no_df_valid_route.clear();
    }

    void SetNoValidDFRouteToLid(ARSWDataBaseEntry &sw_db_entry,
            int plft_id, uint16_t lid_num)
    {
        sw_db_entry.m_p_df_data->m_plft[plft_id].m_no_df_valid_route.insert(lid_num);
    }

    /*
    void ARUpdateDFGroupTable(ARSWDataBaseEntry &sw_db_entry,
            int plft_id);

    void ARUpdateDFGroupTable(ARSWDataBaseEntry &sw_db_entry,
            struct SMP_ARGroupTable p_ar_calculated_group_table[],
            u_int16_t calculated_max_group_id,
            int plft_id);
            */

    void ARUpdateDFLFTTable(ARSWDataBaseEntry &sw_db_entry,
            struct SMP_ARLinearForwardingTable_SX p_ar_calculated_lft_table[],
            u_int16_t calculated_max_lid,
            int plft_id);

    void ARDumpDFAnalizedSetup();

    void ARDumpDFSettings(ARSWDataBaseEntry &sw_db_entry);

    //num_ports - num of phys ports, ports number not including port 0
    u_int8_t GetNextStaticPort(
            uint16_t *ports_load,
            PSPortsBitset &ps_group_bitmask,
            bool isHCA,
            u_int8_t num_ports);

    /**
     * Calculate adaptive routing settings according to Dragon Flay algorithm
     */
    int ARCalculatePortGroupsDF(AnalizeDFSetupData &setup_data);

    /**
     * Update all needed switches with the new calculated DF group table
     */
    int  ARGroupTableProcessDF();

    /**
     * Update all needed switches with the new calculated DF lft table
     */
    int ARLFTTableProcessDF();

    /**
     * Configure Vlids DF pLfts on all switches in case ucast cache
     * is valid and only vlid changed
     */
    void UpdateVlidsLftsDF();

    void SetLftTop(ARSWDataBaseEntry &sw_db_entry,
                   PLFTMads *p_plft_mads, //!!! TAMIR add to OSM
                   u_int8_t  pLFTID);

    /**Set Private LFTs and VL2VL Mapping on the switch*/
    int ARMapPLFTsAndVL2VLs();

    /**Set SL2VL definition on the hosts CAs */
    void ARMapSL2VLOnHosts();

    /**Set VL2VL definition on the switch*/
    void ARMapVL2VL(ARSWDataBaseEntry &sw_db_entry,
                    u_int8_t port_x);

    /** 
     * Helper for ARMapVL2VL get op_vl 
     * validate that op_vl >1 or that out port is not switch 
     * else return non zero to indicate not to set SL2VL 
     */
    int GetOpVlForVL2VL(ARSWDataBaseEntry &sw_db_entry,                                            
                        u_int8_t out_port,
                        u_int8_t &op_vls);

    /**Set Private LFTs definition on the switch*/
    int ARDefinePLFTs();

    /**Map in port/VL to PLFT*/
    void ARMapPLFTs(ARSWDataBaseEntry &sw_db_entry,
                    u_int8_t port_block);

    friend class AdaptiveRoutingAlgorithm;
    friend class PlftBasedArAlgorithm;
    friend class ArKdorAlgorithm;
};


#endif          /* AR_MGR_H_ */

